#
# @file: Y3PUtilPrettyPartition.awk
#
# @desc: prettyfy IsoFun output
#
# inputs : the IsoFun partition output
#
# the IGNORE optional parameter indicates columns to ignore in statistics.
# columns number should be separated by "|"  e.g. -v IGNORE="1|3"
#
# the SORT optional parameter indicates how to sort the classes
# if should be one of : "ByNodes" or "ByElements" 
# (default is ByNodes)
#

# ----------------------------
# library functions
# ----------------------------

# get array size
#
function ArraySize(array, _local_, e, size) {
  size = 0;
  for (e in array)
    size++;
  return size;
}

# execute command with arrayOut as input
#
function ExecuteOut(cmd, arrayOut, _local_, size) {
  size = ArraySize(arrayOut);
  for (i = 1 ; i<= size ; i++)
    print arrayOut[i] | cmd;
  close(cmd);
}

# execute command and get result into array ; return array size
#
function ExecuteIn(cmd, arrayIn, _local_, size) {
  size = 0;
  while (cmd | getline) 
    arrayIn[++size] = $0;
  close(cmd);
  return size;
}

# create a temporary file (with prefix)
#
function TempFile(prefix, _local_, arrayTmp) {
  ExecuteIn("echo " prefix "_$$", arrayTmp);
  return arrayTmp[1];
}

# delete a file
#
function DeleteFile(filename, _local_, arrayTmp) {
  ExecuteIn("rm -f " filename, arrayTmp);
}

# sort arrayIn into arrayOut
#
function Sort(arrayIn, arrayOut, options, _local_, size, tempo, arrayTmp) {

  tempo = TempFile("TSort");
  
  ExecuteOut("tr -d '+' > " tempo, arrayIn);

  size = ExecuteIn("sort " options " " tempo, arrayOut); 

  DeleteFile(tempo);
  
  return size;
}

# ----------------------------
# rules
# ----------------------------

BEGIN {
  if ((IGNORE != 0) && (IGNORE != "")) {
    na = split(IGNORE, a, "|");
    for (i = 1 ; i <= na ; i++)
      ignore[a[i]]++;
    IgnoreInfo = "(ignoring columns " IGNORE ") "
  }
  else {
    IgnoreInfo = ""
  }
    
  if ((SORT == 0) || (SORT != "ByElements"))
    SORT = "ByNodes";
}

# ------------------------------------ reading Partition file

/^q / {  # --------------------------- q : class definition line

  nClass++; 

  # ------------------------------------------
  # get nbColors if not already defined
  
  if (nColors == 0) {
    nColors = NF - 2;
  }

  nTuple = $2;
  
  # ------------------------------------------
  # update statistics

  # tuple stats
  
  histoTuple[nTuple]++;
  if ((minTuple == 0) || (nTuple < minTuple))
    minTuple = nTuple;
  if (nTuple > maxTuple)
    maxTuple = nTuple;

  # elements stats

  for (i = 1 ; i <= nColors ; i++) {
    count[i] = $(i+2);
  }

  min = nTuple;
  for (i = 1 ; i <= nColors ; i++) {
    if (ignore[i])
      continue;
    if (count[i] < min)
      min = count[i];
  }

  histoElement[min]++;
  if ((minElement == 0) || (min < minElement))
    minElement = min;
  if (min > maxElement)
    maxElement = min;

  # ------------------------------------------
  # keep class label for later printing

  score = (SORT == "ByNodes") ? nTuple : min;
  classString = score " " nTuple " " min " " nClass

  for (i = 1 ; i <= nColors ; i++) {
    if (ignore[i])
      classString = classString " (" count[i] ")";
    else
      classString = classString " " count[i];
  }
  
  Class[nClass] = classString;
  
  next;
}

/^v / {  # --------------------------- v : vertex line

  line = substr($0, 3);
  
  nVertex++;
  
  Vertex[nVertex] = line;

  Class[nClass] = Class[nClass] " " nVertex;
  
  next;    
}

/^g / {  # --------------------------- g : gap info line

  NInfo[nClass]++;
  Info[nClass, NInfo[nClass]] = "gaps pattern : " substr($0, 2);
  
  next;
}

END {

  # ------------------------------------------
  # sort classes by size

  if (nClass > 0)
    Sort(Class, ClassSort, "-nr");

  # ------------------------------------------
  # printout header  

  print "XX"
  print "XX CCC statistics :"
  print "XX N = " nClass + 0
  print "XX min = " minTuple
  print "XX max = " maxTuple
  print "XX histogram:"
  for (i = minTuple ; i <= maxTuple ; i++) {
    if (histoTuple[i] == 0)
      continue;
    print "XX t " i, histoTuple[i];
  }
  print "XX"
  print "XX Elements statistics " IgnoreInfo ":"
  print "XX N = " nClass + 0
  print "XX min = " minElement
  print "XX max = " maxElement
  print "XX histogram:"
  for (i = minElement ; i <= maxElement ; i++) {
    if (histoElement[i] == 0)
      continue;
    print "XX e " i, histoElement[i];
  }
  print "XX"
  if (SORT == "ByNodes") 
    print "XX Classes sorted by nodeSize :"
  else
    print "XX Classes sorted by min[elementSize] " IgnoreInfo ":"
  print "XX"

  # ------------------------------------------
  # printout classes  

  for (i = 1 ; i <= nClass ; i++) {
  
    na = split(ClassSort[i], a);
    
    score = a[1];
    nTuple = a[2];
    esize = a[3];
    iClass = a[4];

    printf("CL score=%d nodeSize=%d min[eltSize]=%d eltSize : ", score, nTuple, esize);
    
    for (j = 1 ; j <= nColors ; j++)
      printf("%s ", a[j+4]);
    printf("\n");
    
    for (j = nColors+5 ; j <= na ; j++) {
      print "VE " Vertex[a[j]];
    }    

    for (j = 1 ; j <= NInfo[iClass] ; j++) {
      print "IF " Info[iClass, j];
    }


    print "XX"
  }

}
