package helix.graph.algo;

import helix.graph.model.Vertex;
import helix.graph.model.Edge;
import helix.graph.model.Graph;

/**
 * a basic skeleton for a Visitor following the connected components (CC) of a graph
 * 
 * note: you basically need to implement your own enter(), exit() and visit(Vertex v) methods
 * 
 * enter()  is called when entering a new CC
 * exit()   is called when finishing the current CC
 * visit(v) is called when adding v into the current CC
 * 
 * in addition you may override the accept(Edge edge) method to select
 * proper edges (by default all existing edges are acceptable)
 *
 */
public abstract class AbstractCCVisitor implements Visitor {

	static protected final String VisitedFlag = "_abstract_visited_";

	protected Graph graph_;

	/*
	 * internal stuff
	 */
	
	/**
	 * remove VisitedFlag
	 */
	protected void removeVisitedFlags_() {
		for (Vertex v : graph_.vertices()) {
			v.remove(VisitedFlag);
		}
	}

	/**
	 * traverse all vertices in connected component
	 * containing v
	 */
	protected void traverse_(Vertex v) {
		if (v.has(VisitedFlag))
			return;
		this.visit(v);
		v.set(VisitedFlag);
		for (Edge e : v.edges()) {
			if (! this.accept(e))
				continue;
			this.traverse_(e.getVertexTo(v));
		}
	}
	
	/*
	 * constructor
	 */
	
	public AbstractCCVisitor(Graph g) {
		graph_ = g;
	}
	
	/*
	 * Visitor API
	 */

	/**
	 * default accept : accept any edge
	 */
	public boolean accept(Edge e) {
		return true;
	}

    /**
     * components traversal
     * return this for piping
     */	

	public Visitor traverse() {
		
		for (Vertex v : graph_.vertices()) {
			if (! v.has(VisitedFlag)) {
				this.enter();
				this.traverse_(v);
				this.exit();
			}
		}
		removeVisitedFlags_();
		return this;
	}
}
