package helix.graph.algo;

/**
 * a Visitor to compute connected components (CC) of a graph.
 * you can retrieve them by the getComponents() method.
 * <pre><code>
 *	CCVisitor visitor = new CCVisitor(graph);
 *  for (GSet<Vertex cc : visitor.traverse().getComponents()) {
 *   ... do something
 *  }
 * </code></pre>
 * note: you may override the accept(Edge edge) method to select
 * proper edges (by default all edges are acceptable)
 * <pre><code>
 *	CCVisitor visitor = new CCVisitor(graph) {
 *    public boolean accept(Edge e) {
 *      ... your condition here ...
 *    }
 *  };
 *  for (GSet<Vertex cc : visitor.traverse().getComponents()) {
 *   ... do something
 *  }
 * </code></pre>
 */


import java.util.Set;

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

public class CCVisitor extends AbstractCCVisitor {

	protected Set<GSet<Vertex>> cc_;
	protected GSet<Vertex> current_;
	
	public CCVisitor(Graph graph) {
		super(graph);
		cc_ = graph.vertexSetFactory().makeSuper();
		current_ = graph.vertexSetFactory().make();
	}

	public void enter() {
		if (current_.size() != 0) // test is for subclasses
			current_ = graph_.vertexSetFactory().make();
	}

	public void exit() {
		if (current_.size() != 0) // test is for subclasses
			cc_.add(current_);
	}

	public void visit(Vertex v) {
		current_.add(v);
	}

	@Override
	/**
	 * override for piping
	 */ 
	public CCVisitor traverse() {
		super.traverse();
		return this;
	}


	/*
	 * added api
	 */

	public Iterable<GSet<Vertex>> getComponents() {
		return cc_;
	}

}
