package helix.graph.algo;

import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

import helix.graph.model.impl.RVertex;

/**
 * a class to duplicate graphs or parts of graph
 */
public class Cloner {
	
	static private final String COPY_VERTEX = "_clonerCopyVertex_";
	
	/**
	 *  protected constructor : this class is static
	 */
	protected Cloner() {
	}
	
	/**
	 * create new vertices in dst graph from set of vertices (in other or same graph)
	 * note: vertices iterates on a Set (there is no duplicated vertex)
	 * @return set of new vertices
	 */ 
	protected static final GSet<Vertex> copyVertices_(Iterable<Vertex> vertices, Graph dst) {
		GSet<Vertex> copySet = dst.vertexSetFactory().make();
		for (Vertex v : vertices) {
			Vertex nv = dst.addVertex();
			nv.copy(v);
			v.set(COPY_VERTEX, new RVertex(nv));
			copySet.add(nv);
		}
		return copySet;
	}

	/**
	 * get the copy of vertex v in copySet if it exists
	 * returns nv = (hasCopy(v) ? copy(v) : v)
	 */

	protected static final Vertex getCopy_(Vertex v) {
		if (v.has(COPY_VERTEX)) {
			return ((RVertex) v.get(COPY_VERTEX)).get();
		}
		return v;
	}

	/**
	 * create new edges in dst graph from set of vertices (in other or same graph)
	 * foreach vertex v in itv
	 *   foreach edge (v, u)
	 *      createEdge(nv, nu)
	 * with  
	 *   nv = (hasCopy(v) ? copy(v) : v)
	 *   nu = (hasCopy(u) ? copy(u) : u)
	 * note: vertices iterates on a Set (there is no duplicated vertex)
	 * 
	 * if 'inside' flag is true : only create edges inside dst
	 * i.e. createEdge(nv, nu) iff (nv != v) && (nu != u)
	 */
	
	protected static final void copyEdges_(Iterable<Vertex> src, Graph dst, boolean inside) {
		for (Vertex v : src) {
			Vertex nv = getCopy_(v);
			if (inside && (nv == v))
				continue;
			for (Edge e : v.edges()) {
				Vertex u = e.getVertexTo(v);
				Vertex nu = getCopy_(u);
				if (inside && (nu == u))
					continue;
				if (! dst.hasEdge(nv, nu)) {
					Edge ne = dst.addEdge(nv, nu);
					ne.copy(e);
				}
			}
		}
	}
	
	/**
	 * remove all temporary registered data
	 */
	protected static final void removeFlags_(Iterable<Vertex> vertices) {
		for (Vertex v : vertices) {
			v.remove(COPY_VERTEX);
		}
	}
	

	/**
	 * make a subgraph from graph restricted to given vertices 
	 */
	public static Graph subGraph(Graph src, Iterable<Vertex> vertices) {
		
		Graph dst = src.factory().make();

		dst.copy(src); // copy graph Registry
		
		copyVertices_(vertices, dst);

		copyEdges_(vertices, dst, true);

		removeFlags_(vertices);
		
		return dst;
	}

	/**
	 * clone a graph completely
	 * same as subGraph(g, g.vertices())
	 */
	public static Graph cloneGraph(Graph src) {
		
		return subGraph(src, src.vertices());
	}

	/**
	 * make union of two disjoints graphs 
	 */
	public static Graph union(Graph g1, Graph g2) {
		
		Graph union = cloneGraph(g1);

		copyVertices_(g2.vertices(), union);

		copyEdges_(g2.vertices(), union, true);

		removeFlags_(g2.vertices());
		
		return union;
	}
	
	/**
	 * duplicate a set of vertices within the same graph<br>
	 * <code>
	 * -1- duplicate each vertex v in vertices to copy(v)
	 * -2- foreach vertex v in vertices
	 *       foreach edge (v, u)
	 *          createEdge(copy(v), u')
	 *     with u' = (hasCopy(u) ? copy(u) : u)
	 * </code>
	 *     
	 * @param graph Graph to modify
	 * @param vertices subset of vertices in graph
	 * @return set of newly created vertices 
	 */
	
	public static GSet<Vertex> duplicateVertices(Graph graph, GSet<Vertex> vertices) {
		
		GSet<Vertex> copy = copyVertices_(vertices, graph);

		copyEdges_(vertices, graph, false);

		removeFlags_(vertices);
		
		return copy;
	}

	/**
	 * duplicate a single vertex in a graph <br>
	 * note: this is quicker than calling duplicateVertices
	 * with a singleton<br>
	 * <code>
	 * -1- duplicate vertex v to copy(v)
	 * -2- foreach edge (v, u)
	 *       createEdge(copy(v), u)
	 * </code>
	 *     
	 * @param graph Graph to modify
	 * @param v Vertex to duplicate
	 * @return newly created vertex (copy(v)) 
	 */

	public static Vertex duplicateVertex(Graph graph, Vertex v) {
		Vertex nv = graph.addVertex();
		nv.copy(v);
		for (Edge e : v.edges()) {
			Vertex u = e.getVertexTo(v);
			Edge ne = graph.addEdge(nv, u);
			ne.copy(e);
		}
		return nv;
	}

}