package helix.graph.algo;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

/**
 * Dijkstra's algorithm
 * solves the single-source shortest path problem
 * 
 * note1 : this implementation is O(V2 + E)
 *         i.e does not use a priority queue
 * 
 * note2: Dijkstra's algorithm is only useful if you want the shortest path
 * from one source. In case you want all pairwise distances, you may
 * consider the Floyd-Warshall algorithm instead.
 * 
 */
public class Dijkstra {

	private static final int Infinity = Integer.MAX_VALUE - 1;

	protected static final String Index    = "_dijkstraIndex_";
	protected static final String Previous = "_dijkstraPrevious_";

	public static final String Distance = "_distance_";
	
	protected Graph graph_;
	protected Vertex source_;
	
	/**
	 * extract min element from *non empty* queue
	 * may improve this implementation with a priority queue
	 * (but take care of external modifications of distance)
	 * 
	 * ppty-1: by construction there is at least one element in queue
	 * with Distance < Infinity, so we shall pass at least once in
	 * the inner test.
	 */
	protected static Vertex extractMin_(GSet<Vertex> queue) {
		int minDist = Infinity;
		Vertex found = null;
		for (Vertex v : queue) {
			int dist = v.get(Distance).toInt();
			if (dist < minDist) {
				minDist = dist;
				found = v;
			}
		}
		queue.remove(found); // by ppty-1 : found != null
		return found;
	}
	
	/*
	 * remove all internal falgs
	 */
	protected void removeFlags_() {
		for (Vertex v : graph_.vertices()) {
			v.remove(Distance);
			v.remove(Index);
			v.remove(Previous);
		}
	}

	/*
	 * API
	 */
	
	/**
	 * constructor
	 */
	
	public Dijkstra(Graph graph) {
		graph_ = graph;
	}

	/**
	 * main Dijkstra algorithm
	 * get all distances from source to all vertices in connected component of source (CC(source))
	 * distance are stored into vertices registry with label Dijkstra.Distance
	 * note: vertices not in CC(source) do not have the Dijkstra.Distance flag
	 */
	public void run(Vertex source) {
		
		// forget all previously set flags
		removeFlags_();
		
		// keep source
		source_ = source;
		
		// get connected component of v
		ComponentsFinder finder = new ComponentsFinder(graph_);
		GSet<Vertex> comp = finder.component(source);

		// initializations
		int index = 1;
		for (Vertex v : comp) {
			v.set(Distance, Infinity);       // Unknown distance from source to v
			v.set(Index, index++);           // Temporary index
			v.set(Previous, 0);              // Unknown previous index on path from source to v
		}
		source.set(Distance, 0);             // Distance from s to s is 0
		
		// queue of unvisited vertices
		GSet<Vertex> queue = graph_.vertexSetFactory().make();
		queue.addAll(comp);

		// the main loop
		while (! queue.isEmpty()) {
			Vertex u = extractMin_(queue);   // Remove best vertex from priority queue
			int alt = u.get(Distance).toInt() + 1;
			for (Vertex v : u.neighbors()) {
				int dist = v.get(Distance).toInt();
				if (alt < dist) {            // Relax (u,v)
					v.set(Distance, alt);
					v.set(Previous, u.get(Index).toInt());
				}
			}
		}
	}

	/**
	 * get current source
	 */
	public Vertex getSource() {
		return source_;
	}
	
	/**
	 * get distance from source to target
	 * return -1 if target is not connected to source
	 */
	public int distance(Vertex target) {
		return (target.has(Distance) ? target.get(Distance).toInt() : -1);
	}

	/**
	 * get shortest path from source to target
	 * return a list [source, ..., target] or empty list if there is no path
	 */
	public List<Vertex> shortestPath(Vertex target) {
		Stack<Vertex> stack = new Stack<Vertex>();
		if (! target.has(Distance))
			return stack; // aka empty list
		Vertex p = target;
		while (p != null) {
			stack.push(p);
			int pindx = p.get(Previous).toInt();
			p = graph_.findVertex(Index, pindx);
		}
		// reverse list order
		ArrayList<Vertex> path = new ArrayList<Vertex>();
		while (! stack.isEmpty()) {
			Vertex v = stack.pop();
			path.add(v);
		}
		return path;
	}

}

// AV : TODO TestDijkstra

