package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph
 * provides methods for all C3Part algorithms.
 * 
 * you basically have to provide a the <code>run_()</code> method
 * implementing the actual algorithm.
 * 
 * In addition you may override some methods of C3PartAlgo inteface
 * in order to change the default behavior.
 */

//
// for readability, we use uppercase characters for variables
// holding nodes of the multigraph (e.g. V, U) and lowercase characters for
// nodes of the datagraph (also called polygraph) (e.g. v, u)
//

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import helix.graph.algo.c3part.aggregator.CCAggregator;
import helix.graph.algo.c3part.aggregator.CenteredAggregator;
import helix.graph.algo.c3part.aggregator.CliqueAggregator;
import helix.graph.algo.c3part.aggregator.DenseCCAggregator;
import helix.graph.algo.c3part.aggregator.NoSuchAggregatorException;
import helix.graph.algo.c3part.aggregator.NodeAggregator;
import helix.graph.algo.c3part.optimizer.ColorOptimizer;
import helix.graph.algo.c3part.optimizer.FixColorOptimizer;
import helix.graph.algo.c3part.optimizer.GlobalColorOptimizer;
import helix.graph.algo.c3part.optimizer.IdentityColorOptimizer;
import helix.graph.algo.c3part.optimizer.NoSuchOptimizerException;
import helix.graph.algo.c3part.optimizer.TerminalsColorOptimizer;
import helix.graph.model.Edge;
import helix.graph.model.GSetFactory;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;
import helix.graph.util.CommandLineException;
import helix.graph.util.IPermutation;

public abstract class C3PartAbstract implements C3PartAlgo {

	// the datagraph beeing processed
	//
	protected Graph datagraph_;
	
	// a graph factory
	//
	protected GraphFactory graphFactory_;
	
	// a Vertex set factory
	//
	protected GSetFactory<Vertex> setFactory_;
	
	// common C3Part options
	//
	protected C3PartOptions options_;

	// the node aggregator to be used
	//
	protected NodeAggregator nodeAggregator_ = null;
	
	/*
	 * utilities
	 */
	
	/**
	 * utility : copy tuple into list (because java.util.List as no addAll(Iterable)
	 */ 
	private static final void copyTuple_(List<Vertex> lst, Iterable<Vertex> tuple) {
		lst.clear();
		for (Vertex v : tuple) {
			lst.add(v);
		}
	}
	
	
	/**
	 * utility : get the size of tuples in multigraph vertices
	 * this function returns the actual tuple size
	 * of nodes in the graph if graph is not empty
	 * else it relies on the number of colors set
	 * in the multigraph.
	 */
	protected static final int tupleSize_(Graph multi) {
		for (Vertex V : multi.vertices()) { // pseudo loop
			return C3PartUtil.getNodes(V).size();
		}
		return RColorSet.nbColors(multi);
	}
	
	/**
	 * utility: reorder multi graph tuples according to permutation
	 */
	protected void reorderTuples_(Graph multi) {
		
		ColorOptimizer optimizer = C3PartUtil.getOptimizer(multi);
		if (optimizer == null) // no optimizer => no color reordering
			return;
		
		List<Integer> colors = optimizer.getUsedColors();
		IPermutation permutation = new IPermutation(colors).shift().inverse();
		if (permutation.isIdentity())
			return;
		
		//;;; System.err.println("@@ reordering with " + optimizer);
		
		ArrayList<Vertex> tmp = new ArrayList<Vertex>();
		
		for (Vertex V : options_.compressor.getVertices(multi)) {
			RVertexList tuple = C3PartUtil.getNodes(V);
			copyTuple_(tmp, tuple);
			List<Vertex> ptmp = permutation.apply(tmp);
			RVertexList ptuple = new RVertexList(ptmp);
			V.set(C3PartUtil.Nodes, ptuple);
		}
	}
	
	/**
	 * utility: shift multigraph colors (vertices and edges) by specified amount.
	 * amount can be positive or negative
	 */
	protected void shiftColors_(Graph multi, int amount) {

		// simple case
		if (amount == 0)
			return;
			
		// shift multigraph colors
		RColorSet.getColors(multi).shift(amount);

		// shift multi edges colors
		Set<Edge> edges = multi.edgeSetFactory().make();
		for (Vertex V : multi.vertices()) {
			RColorSet.getColors(V).shift(amount);
			for (Edge E : V.edges()) {
				if (! edges.contains(E)) { // don't shift edges twice
					RColorSet.getColors(E).shift(amount);
					edges.add(E);
				}
			}
		}
	}
	
	/**
	 * make default options using the super defaults defined in C3PartOptions
	 */
	protected C3PartOptions makeDefaultOptions_() {
		C3PartOptions options = new C3PartOptions();
		options.colors = new IPermutation(1, RColorSet.nbColors(datagraph_) - 1);
		return options;
	}
	
	/*
	 * C3PartAlgo API
	 */

	/**
	 * default pseudo constructor
	 * note: defining the constructor in the abstract class
	 * will force its requirement in subclasses : just
	 * call super() if you don't need to override.
	 */
	public C3PartAbstract(Graph dataGraph) {
		datagraph_ = dataGraph;
		graphFactory_ = dataGraph.factory();
		setFactory_ = dataGraph.vertexSetFactory();
		options_ = this.makeDefaultOptions_();
	}

	/**
	 * get current options
	 * return a hard copy of options, so modifying them will not modify
	 * internal algorithm options. to do so you should set options back
	 * with <code>setOptions</code>
	 */
	public C3PartOptions getOptions() {
		return new C3PartOptions(options_);
	}
	
	/**
	 * get current options
	 * this method returns the internal algo options. Take care
	 * not modifying them (use <code>getOptions</code> instead)
	 */
	public C3PartOptions getPhysicalOptions() {
		return options_;
	}
	
	/**
	 * set options
	 * 
	 * time to check for option incompatibilities
	 */
	public void setOptions(C3PartOptions options) {
		//
		// record options
		//
		options_ = options;

		//
		// check the center aggregator
		// [AV] improve this : center aggregator just implies
		//      the first color not all of them ;-)
		//
		if (        options_.nodeAggregatorName.equalsIgnoreCase(C3PartOptions.CENTER_AGGREGATOR)
			 && (! (options_.optimizerName.equalsIgnoreCase(C3PartOptions.USER_OPTIMIZER)))) {
			options_.optimizerName = C3PartOptions.USER_OPTIMIZER;
			options_.colors = (options_.userColors != null) 
            				? new IPermutation(1, options_.userColors.size()) 
            				: new IPermutation(1, RColorSet.nbColors(datagraph_)-1);
			options_.logger.severe(  "--aggregator 'center' requires "
					               + "--optimizer 'user colors' "
					               + "(will be forced to " + options.colors + ")");
			}

		//
		// check user colors
		//
		if (options_.userColors != null) {
			IPermutation colors = new IPermutation(1, RColorSet.nbColors(datagraph_)-1);
			if (! colors.containsAll(options_.userColors)) {
				throw new CommandLineException("invalid user colors");
			}
		}
		
		//
		// check user optimizer
		//
		if (options_.optimizerName.equalsIgnoreCase(C3PartOptions.USER_OPTIMIZER)) {
			IPermutation colors = (options_.userColors != null) 
			                      ? new IPermutation(1, options_.userColors.size()) 
			                      : new IPermutation(1, RColorSet.nbColors(datagraph_)-1);
			if (! (   colors.containsAll(options_.colors)
				   && options_.colors.containsAll(colors))) {
				throw new CommandLineException("invalid colorset " + options_.colors
						      + " for optimizer, valid range is " + colors);
			}
		}
		
	}
	
	/**
	 * run algo with current options.
	 * this calls in turn :
	 * <pre>
	 *   boolean this.beforeRun_()
	 *       run() stops if result is false, else continue
	 *   Set<Graph> this.run_()
	 *       actually compute the partition
	 *   Set<Graph> this.afterRun_(Set<Graph> partition)
	 *       perform postprocessing
	 * </pre>
	 * this Abstract class proposes some default behavior for
	 * <code>beforeRun_</code> and <code>afterRun_</code>
	 */
	public Set<Graph> run() {

		// before run
		//
		if (! this.beforeRun_())
			return null;

		// run our internal algorithm (see below)
		//
		Set<Graph> partition = this.run_();

		// after run
		//
		return this.afterRun_(partition);
	}
	
	/**
	 * get aggregator by name
	 * this abstract class knows :
	 * <pre>
	 * Name    Aggregator
	 * clique  CliqueAggregator
	 * star    CenteredAggregator
	 * </pre>
	 */
	public NodeAggregator getAggregatorFor() {
		
		// avoid multiple instanciation
		if (nodeAggregator_ == null) {

			String name = options_.nodeAggregatorName;

			if (C3PartOptions.CLIQUE_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new CliqueAggregator(datagraph_, options_);
			else if (C3PartOptions.CENTER_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new CenteredAggregator(datagraph_, options_);
			else if (C3PartOptions.CC_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new CCAggregator(datagraph_, options_, null);
			else if (C3PartOptions.DENSE_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new DenseCCAggregator(datagraph_, options_, null);
			else
				throw new NoSuchAggregatorException("aggregator " + name 
						          + " is not available for algo " + this.getClass().getSimpleName());
		}
	
		return nodeAggregator_;
	}
	
	/**
	 * get optimizer for Clas
	 * this abstract class knows :
	 * <pre>
	 * Name    Optimizer
	 * off     IdentityColorOptimizer
	 * user    FixColorOptimizer
	 * global  GlobalColorOptimizer
	 * local   TerminalsColorOptimizer
	 * </pre>
	 */
	public ColorOptimizer getOptimizerFor(Graph Clas) {
		String name = options_.optimizerName;
		
		if (C3PartOptions.NO_OPTIMIZER.compareToIgnoreCase(name) == 0) {
			return new IdentityColorOptimizer(RColorSet.nbColors(datagraph_) - 1);
		}
		if (C3PartOptions.USER_OPTIMIZER.compareToIgnoreCase(name) == 0) {
			return new FixColorOptimizer(options_.colors);
		}
		if (C3PartOptions.LOCAL_OPTIMIZER.compareToIgnoreCase(name) == 0) {
			return new TerminalsColorOptimizer(this, datagraph_, Clas);
		}
		if (C3PartOptions.GLOBAL_OPTIMIZER.compareToIgnoreCase(name) == 0) {
			return new GlobalColorOptimizer(this, datagraph_, Clas);
		}
		throw new NoSuchOptimizerException("optimizer " + name 
				        + " is not available for algo " + this.getClass().getSimpleName());
	}
	
	/*
	 * added API
	 */
	
	/**
	 * added API
	 * default beforeRun_ : setup internal fields, please
	 * don't forget to call super.
	 */
	protected boolean beforeRun_() {
		
		// setup aggregator
		//
		nodeAggregator_ = getAggregatorFor();
		
		// notification
		//
		this.logAlgoInfo_();
		
		return true; // always continue
	}

	/**
	 * added API
	 * internal run of the algorithm once all options have been set.
	 * you should provide this method to implement the actual algorithm.
	 */
	protected abstract Set<Graph> run_();
	
	/**
	 * added API
	 * default afterRun_ : don't do anything, just return partition as is
	 */
	protected Set<Graph> afterRun_(Set<Graph> partition) {
		return partition;
	}
	
	/**
	 * added API
	 * log information about this algorithm
	 */
	protected void logAlgoInfo_() {
		
		ColorOptimizer optimizer = this.getOptimizerFor(null);
		
		options_.logger.info("--------------------------------- Algo parameters");
		options_.logger.info("algo_class:       " + this.getClass().getSimpleName());
		options_.logger.info("aggregator_class: " + this.nodeAggregator_.getClass().getSimpleName() + " " + nodeAggregator_.getParamInfo());
		options_.logger.info("optimizer_class:  " + optimizer.getClass().getSimpleName() + " " + optimizer.getParamInfo() + " " + optimizer.preview());
		options_.logger.info("compressor_class: " + options_.compressor.getClass().getSimpleName());
		options_.logger.info("minsize:          " + options_.minsize);
		options_.logger.info("mineltsize:       " + options_.mineltsize);
	}
}
