package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph
 * provides methods for OTF algorithms (i.e OTF and OTFx thru subclasses).
 * 
 * Note: this version contains some extra code that
 * is not really necessary for strict OTF. This code
 * is intended to give a more general behavior in order
 * to facilitate subclassing (especially for the OTFQuorum
 * family). This is related to keeping colors
 * in multigraph vertices. These colors indicate which colors
 * of the datagraph nodes are kept in the multigraph
 * vertex. In strict OTF all multigraph vertices have the
 * same colors. This may be not true for some subclasses. E.g in OTFQuorum
 * the star is colorless and therefore, the colors of vertices
 * are actually subsets of the multigraph colors.
 * This extra code is indicated below with the @NonStrict keyword.
 * 
 * experimental section : multiprocessors edition is marked as @concurrent
 * experimental section : memory optimization is marked as @memoryOptimize
 * experimental section : graph compression is marked as @compress
 * 
 */

//
// for readability, we use uppercase characters for variables
// holding nodes of the multigraph (e.g. V, U) and lowercase characters for
// nodes of the datagraph (also called polygraph) (e.g. v, u)
//

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.compressor.MultiGraphCompressor;
import helix.graph.algo.c3part.optimizer.ColorOptimizer;
import helix.graph.algo.c3part.preprocessor.Chromatizer;
import helix.graph.algo.c3part.preprocessor.MultiGraphCloser;
import helix.graph.algo.c3part.visitor.CCByColorVisitor;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.util.ConcurrentService;

import java.util.HashSet;
import java.util.Set;

public abstract class C3PartAbstractOTF extends C3PartAbstract {

	ConcurrentService concurrent_; // service for optional concurrency
	
	/**
	 * internal concurrent worker class
	 *
	 */
	protected class DFSWorker implements Runnable {
		
		protected Graph clas_;
		protected Set<Graph> ccc_;
		
		protected DFSWorker(Graph Clas, Set<Graph> ccc) {
			clas_ = Clas;
			ccc_ = ccc;
		}
		
		public void run() {
			splitGraphDFS__(clas_, ccc_);
		}
	}
	
	/*
	 * internal utilities
	 */

	/*
	 * Core methods
	 */
	
	/**
	 * split multigraph Clas into its connected components multisubgraphs on given color
	 */
	protected Set<Graph> splitGraphColor_(Graph Clas, int color) {

		// run the connected components visitor
		//
		CCVisitor visitor = new CCByColorVisitor(Clas, color);
		Iterable<GSet<Vertex>> cc = visitor.traverse().getComponents();
		
		// physically split graph
		// @memoryOptimize note: this destroys Clas
		//
		return Clas.physicalSplit(cc);
	}

	/**
	 * Get datagraph nodes suitable to expand multigraph vertex V
	 * 
	 */
	protected GSet<Vertex> terminalNodes_(Vertex V, int color) {
		return nodeAggregator_.getTerminals(V, color);
	}
	
	/**
	 * expand multinode V with new terminal t
	 */
	protected Vertex expandNode_(Vertex V, Vertex t, Graph multi) {
		
		RColorSet tColors = RColorSet.getColors(t);  // @NonStrict - see below -
		
		// make new multinode
		//
		Vertex NewV = multi.addVertex();
		
		// copy V nodes and colors
		//
		NewV.copy(V);				 				 // copy V Nodes (and colors)
		RColorSet.getColors(NewV).orColors(tColors); // @NonStrict
		
		// add terminal
		//
		C3PartUtil.addNode(NewV, t);

		return NewV;
	}

	/**
	 * Expand multigraph 'Clas' by adding 'color' nodes
	 * 
	 * Note: this version contains some extra code that
	 * is not really necessary for strict OTF. This code
	 * is intended to give a more general behavior in order
	 * to facilitate subclassing (especially for the OTFQuorum
	 * family). This is related to keeping colors
	 * in multigraph vertices. These colors indicate which colors
	 * of the datagraph nodes are kept in the multigraph
	 * vertex. In strict OTF all vertices have therefore the
	 * same colors (i.e. all the colors of the multigraph).
	 * This may be not true for some subclasses. E.g in OTFQuorum
	 * the star is colorless and therefore, the colors of vertices
	 * are subsets of the multigraph colors.
	 * This extra code is indicated below with the @NonStrict
	 * keyword.
	 * Also some parts or section are related to compressing
	 * multigraphs. Again, these parts are not vital for the
	 * basic algorithm. They are indicated by the @Compress
	 * keyword.
	 * 
	 */
	protected Graph expandMultiGraph_(Graph Clas, int color) {

		MultiGraphCompressor compress = options_.compressor;

		//;;; options_.debug.log("[expandMultiGraph_] Entering with " + Clas.nbVertices() + " vertices (actual=" + compress.nbVertices(Clas) + ") " + Clas.nbEdges() + " edges " + RColorSet.nbColors(Clas) + " colors on color " + color);
		
		Graph newClas = graphFactory_.make();
		newClas.copy(Clas);                           // copy colors
		RColorSet.getColors(newClas).setColor(color); // add new color

		GSet<Vertex> latest = setFactory_.make();
		GSet<Vertex> sonsOfNeighbors = setFactory_.make();
		

		// @Compress : without compression the loop simply reads as
		//
		// 		for (Vertex V : Clas.vertices()) 
		//
		// note that the compress.getVertices() should return
		// vertices in an ordered way : any root vertex should be immediately 
		// followed by it aliases
		//

		for (Vertex V : compress.getVertices(Clas)) {

			// @NonStrict : in OTF, vertex colors are the same
			// as Clas colors. this may differ in OTFQuorum
			//
			RColorSet VColors = RColorSet.getColors(V);

			// get datagraph vertices to add == terminals
			//
			GSet<Vertex> terminals = terminalNodes_(V, color);
			
					
			latest.addAll(terminals); 

			// ********************
			// @Compress section (without compression just skip section)
			//
			// we try to avoid full expansion of aliases edges
			// by keeping aliases with same terminals as their root
			// bound together.
			//
			Vertex Root = null;
			
			if (compress.isRoot(V)) {  // keep track of latest root found
				compress.storeSet(terminals);
			}

			
			if (compress.isAlias(V)) {   // V is an alias node

				Root = compress.getRoot(V);
				// retrieve last stored set of terminals
				// i.e. terminals of Root
				//
				GSet<Vertex> interTerminals = compress.retrieveSet();
				interTerminals.retainAll(terminals);
				
				// just make new aliases vertices with common terminals
				//
				for (Vertex v : interTerminals) {
					Vertex NewV = expandNode_(V, v, newClas);
					// and compress them immediately to their new root
					//
					Vertex SonOfRoot = C3PartUtil.getSon(Root, v);

					compress.compressVertex(NewV, SonOfRoot, newClas);   
				}
				// deal with remaining terminals
				//
				terminals.removeAll(interTerminals);
				if (terminals.isEmpty())
					continue;

				compress.expandVertex(V, Clas);
			}
			
			//
			// @Compress : end of compress section
			// *******************

			//
			// terminals loop
			//
			
			for (Vertex v : terminals) {
				
				RColorSet vColors = RColorSet.getColors(v); // @NonStrict
				
				// -1- make a new multigraph vertex with this terminal
				//
				
				Vertex NewV = expandNode_(V, v, newClas);	 // new multigraph vertex
				
				// -2- restore initial Clas edges 
				//     with their initial colors
				//
				// @Compress : original code (for uncompressed graph) is simply :
				//
				// for (Edge E : V.edges()) {
				//   Vertex U = E.getVertexTo(V);
				//   //
				//   // in the following loop some U may not have sons yet.
				//   // this does not hurt since the edge will be created
				//   // later on.
				//   //
				//   for (Vertex NewU : C3PartUtil.getSons(U)) {
			    //     Edge NewE = newClas.addEdge(NewV, NewU);
				//     NewE.copy(E); // copy colors
				//   }
			    // }

				for (Vertex U : compress.getNeighbors(V)) {  // @Compress : includes aliases of neighbors
					
					RColorSet eColors = compress.getColors(V, U); // @Compress : pseudo colors
					
					if (eColors.isEmpty()) // @Compress : no actual edge
						continue;
					
					for (Vertex NewU : C3PartUtil.getSons(U)) {
						
						if (compress.isAlias(NewU)) // @Compress :  NewU is an alias, no edge needed 
							continue;

						RColorSet neColors;
						// just a speedup
						if (compress.isNormal(NewV) && compress.isNormal(NewU))
							neColors = eColors;
						else
							neColors = compress.getColors(NewV, NewU); // @Compress : pseudo colors
						
						if (neColors.isEmpty()) // @Compress : no edge
							continue;
						
						Edge NewE = newClas.addEdge(NewV, NewU);
						
						RColorSet.getColors(NewE).setColors(neColors);
					}
				}

				// -3- add edges between brothers (i.e. sons of the same father)
				// these edges have all colors of father V since the initial
				// nodes are the same. @NonStrict : in strict OTF they all have
				// the colors of Clas.
				//
				// in the following loop some of brothers may not have
				// been born yet. this does not hurt since the edges will
				// be created later on.
				//
				for (Vertex NewU : C3PartUtil.getSons(V)) { // already born brothers
					Edge NewE = newClas.addEdge(NewV, NewU);
					RColorSet.getColors(NewE).setColors(VColors); // @NonStrict
				}
				
				// -4- add new edges corresponding to the terminal node v
				//     we should connect NewV to all sons of v and to all
				//     sons of v neighbors. note that only neighbors having
				//     already sons are eligible.
				//
				// @NonStrict: see -5- below : terminal nodes without color
				//             have no sons.
				//

				// -4.1- collect multigraph vertices to connect
				sonsOfNeighbors.clear();
				sonsOfNeighbors.addIterable(C3PartUtil.getSons(v));
				for (Vertex u : v.neighbors()) {
					if (C3PartUtil.hasSon(u)) {  // see -5- below 
						sonsOfNeighbors.addIterable(C3PartUtil.getSons(u));
					}
				}
				// -4.2- connect to them
				for (Vertex NewU : sonsOfNeighbors) {
					Edge E = newClas.assertEdge(NewV, NewU);
					RColorSet.getColors(E).setColor(color);
				}

				// -5- Finally register NewV as a son of V and v
				//
				if (! VColors.isEmpty()) {				 // @NonStrict : in OTF this should be always done
					C3PartUtil.addSon(V, NewV);			 // NewV is a son of V
				}
				
				if (vColors.hasColor(color)) {			// @NonStrict : in OTF this should be always done		
					C3PartUtil.addSon(v, NewV);			// NewV is son of v too
				}
				latest.add(v);
				
			} // end of v loop

			//;;; options_.debug.log("[expandMultiGraph_] current with " + newClas.nbVertices() + " vertices " + newClas.nbEdges() + " edges ");

			// @Compress section (without compression just skip following test)
			//
			// recompress previously expanded Alias to save memory
			//
			if (Root != null) {
				compress.compressVertex(V, Root, Clas);
			}

		} // end of V loop
		
		//
		// cleanup : reset terminals in datagraph for next time
		//
		C3PartUtil.removeSons(latest);

		//
		// Note: the vertices in Clas have been modified (Sons added)
		//       and should not be reused as-is
		//       [AV] we may cleanup Sons here too (see below)
		//       but this is not needed since they will be garbage collected
		//
		//C3PartUtil.removeSons(Clas.vertices());
		
		//
		// experimental section :  keep local ordering (deltashuf)
		//
		if (options_.deltaShuf != null) {
			int maxColor = RColorSet.getColors(datagraph_).maxColor();
			MultiGraphCloser closer = new MultiGraphCloser(options_.deltaShuf, maxColor);
			closer.run(newClas);
			RColorSet multiColors = RColorSet.getColors(newClas);
			Chromatizer chromo = new Chromatizer(multiColors, maxColor, options_.deltaShuf);
			chromo.run(newClas);
		}

		// @Compress section (without compression just skip next line)
		//
// AV experimental
//		float nv = (float) newClas.nbVertices();
//		float ne = (float) newClas.nbEdges();
//		float fract = ne * 2F / (nv * nv);
//		if ((ne > 10000) && (fract > 0.5)) // TODO parameter
		compress.compress(newClas, options_.debug.getLogger());
		
		//;;; options_.debug.log("[expandMultiGraph_] Ending with " + newClas.nbVertices() + " vertices (actual=" + compress.nbVertices(newClas) + ") " + newClas.nbEdges() + " edges ");

		//;;; if (options_.debug.getVerbose()) {
		//;;;	MultiGraphDebugger.checkColors(newClas, "End of expandMultiGraph", true);
		//;;; }
		
		return newClas;
	}
	
	/**
	 * filter multigraph when entering splitGraphDFS
	 * return true : split has to be continued
	 *       false : don't split this class
	 *       
	 * this filter is used here to check the mineltsize
	 * condition : if one genome (i.e. color) has
	 * less than mineltsize elements then we return
	 * false
	 */
	protected boolean beforeSplitGraphDFS_(Graph Clas) {

		GSet<Vertex> terminals = setFactory_.make();

		// check if for each color, the number of non-star nodes
		// is at least mineltsize
		
		int nbColors = tupleSize_(Clas);
		for (int indx = 0 ; indx < nbColors ; indx++) {
			terminals.clear();
			for (Vertex V : options_.compressor.getVertices(Clas)) { // @compress : Clas.vertices()
				Vertex v = C3PartUtil.getNodes(V).getVertex(indx);
				if (RColorSet.hasColors(v)) // @NonStrict
					terminals.add(v); 
				if (terminals.size() >= options_.mineltsize)
					break;
			}
			if (terminals.size() < options_.mineltsize)
				return false;  // condition not satisfied
		}
		
		return true; // condition satisfied for all colors
	}

	/**
	 * store class in CCC when leaving splitGraphDFS with a proper CCC
	 * return true : CCC has been stored
	 *       false : CCC should be discarded
	 * 
	 * here we only check the minsize condition
	 */
	protected boolean afterSplitGraphDFS_(Graph Clas, Set<Graph> ccc) {
		
		// @compress : Clas.nbVertices()
		if (options_.compressor.nbVertices(Clas) < options_.minsize) 
			return false;

		// TODO: choix a faire : renvoyer une partition de vertices
		//                       ou un graphe sans arete
		options_.compressor.flatten(Clas);
		
		ccc.add(Clas);
		
		//;;; options_.debug.log("@ After Stock OTF : " + ccc.size() + " " + Clas.nbVertices());
		
		return true;
	}

	/**
	 * @concurrent: experimental section : switcher for concurrent version of splitGraphDFS_
	 * heuristics : we don't start a new process if the number of Clas vertices is too low
	 * since this process is likely to terminate quickly, it seems better to continue
	 * within the current thread.
	 */

	protected final void splitGraphDFS_(Graph Clas, Set<Graph> ccc) {
		if (concurrent_.isEnabled() && (Clas.nbVertices() > options_.concthresh)) {
			concurrent_.addProcess(new DFSWorker(Clas, ccc));
		}
		else {
			splitGraphDFS__(Clas, ccc);
		}
	}

	/**
	 *  recursively split Clas on all colors (DFS approach).
	 *  store common connected components (as multigraphs) into ccc
	 *  if it is not done yet, Clas is extended to the colorth genome
	 *  
	 *  note: the @memoryOptimize comments indicate part of code that is not
	 *  fundamental in the algorithm but only to optimize memory usage
	 *  (i.e. to let the garbage collector work)
	 *  
	 */

	protected void splitGraphDFS__(Graph Clas, Set<Graph> ccc) {


		//;;; options_.debug.indentUp();
		//;;; options_.debug.log("[splitGraphDFS_] Entering with " + Clas.nbVertices() + " vertices " + Clas.nbEdges() + " edges " + RColorSet.nbColors(Clas) + " colors");
		//;;; options_.debug.log(Clas, false, "entering");

		// @memoryOptimize 
		// the current colors optimizer will never be used again so we can
		// tell it to free its internal components
		//
		C3PartUtil.getOptimizer(Clas).giveup(); // @memoryOptimize

		// precheck : experimental minEltSize option
		//
		if (! beforeSplitGraphDFS_(Clas)) {
			//;;; options_.debug.log("pruning");
			//;;; options_.debug.indentDown();
			Clas.clear(); // @memoryOptimize
			return;
		}

		// debug
		//;;; options_.debug.log("[splitGraphDFS_] optimizer " + Clas.get(C3PartUtil.Optimizer));
		
		// make a new color optimizer for this Clas
		// 
		ColorOptimizer colorsOptimizer = this.getOptimizerFor(Clas);
		C3PartUtil.setOptimizer(Clas, colorsOptimizer);
		
		// go ahead on remaining colors
		//
		while (colorsOptimizer.hasNext()) {
			
			int color = colorsOptimizer.next();
			
			if (RColorSet.hasColor(Clas, color)) {          // The multigraph is already computed 

				// try to split it on color just as in C3PartDFS
				// @memoryOptimize note: Clas will be destroyed 
				Set<Graph> split = splitGraphColor_(Clas, color);

				//;;; options_.debug.log("[splitGraphDFS_] Splitted on color " + color + " split size= " + split.size());
				
				if (split.size() == 1) {					// Clas unsplitted
					Clas = split.iterator().next();			//   continue to next color
				}
				else {										// Clas has been splitted
					for (Graph sub : split) {					// for all splits
						splitGraphDFS_(sub, ccc);				//   recurse to split subclasses on *all* colors
					}											//   (since split may disconnect previous colors)
					//;;; options_.debug.indentDown();
					return;
				}
			}
			else {                                          // The multigraph needs expansion

				// add color in multigraph
				//
				Clas = expandMultiGraph_(Clas, color);
	
				//;;; options_.debug.log("[splitGraphDFS_] Expanded on color " + color + " Clas size= " + Clas.nbVertices());
				//;;; options_.debug.log(Clas, false, "entering");
				
				// recurse to split expanded multigraph on *all* colors
				// (since expand can actually remove some vertices connecting previous colors)
				//
				// normal code is :
				// splitGraphDFS_(Clas, ccc);
				//
				// the following is a better variant :
				// we first try to split on latest color since it is likely to explode newClas
				// in smaller manageable pieces
				//
				Set<Graph> split = splitGraphColor_(Clas, color);
				//;;; options_.debug.log("[splitGraphDFS_] split on (latest) color " + color + " yields "+ split.size() + " classes");
				
				for (Graph sub : split) {
					splitGraphDFS_(sub, ccc);
				}
				
				//;;; options_.debug.indentDown();
				return;
			}
		} // end of colors loop

		// if we come here this means that 
		// Clas remains unsplitted after all colors => it's a CCC 
		
		//;;; options_.debug.log("[splitGraphDFS_] CCC Found ");
		//;;; options_.debug.log(Clas, false, "CCC");
		//;;; options_.debug.indentDown();
		
		concurrent_.lock();		// @concurrent:
		try {
			afterSplitGraphDFS_(Clas, ccc);
		} 
		finally {
			concurrent_.unlock();	// @concurrent:
		}
		
		colorsOptimizer.giveup(); // @memoryOptimize : giveup with latest optimizer
	}
	
	/**
	 * algorithm : multigraph initialization
	 */
	protected Graph initMultiGraph_() {

		Graph multi = graphFactory_.make();

		// initialize multigraph registry
		// with the first color
		ColorOptimizer optimizer = this.getOptimizerFor(null); // first clas
		C3PartUtil.setOptimizer(multi, optimizer);

		int color = optimizer.next(); // optimizer will guess first color to use
		
		RColorSet.getColors(multi).setColor(color);

		// initialize partition with 1 class of all vertices
		// of the first color
		
		// creating vertices
		
		for (Vertex v : datagraph_.vertices()) {
			if (RColorSet.hasColor(v, color)) { 
				Vertex V = multi.addVertex();
				RColorSet.getColors(V).setColor(color);
				C3PartUtil.addNode(V, v);
				C3PartUtil.setMultinode(v, V); // we cheat a bit here ; see -1- below
			}
		}
		
		// creating edges

		for (Vertex V : multi.vertices()) {
			Vertex v = C3PartUtil.getLastNode(V);
			for (Edge e : v.edges()) {
				Vertex u = e.getVertexTo(v);
				if (! RColorSet.hasColor(u, color))
					continue;
				Vertex U = C3PartUtil.getMultinode(u); // -1- this is the reason why
				if (! multi.hasEdge(V, U)) {
					Edge E = multi.addEdge(V, U);
					RColorSet.getColors(E).setColor(color);
				}
			}
		}
		
		// cleanup temporary father flag
		for (Vertex V : multi.vertices()) {
			Vertex v = C3PartUtil.getLastNode(V);
			C3PartUtil.removeMultinode(v);
		}
		
		return multi;
	}
	
	/*
	 * required C3PartAlgo API
	 */

	public C3PartAbstractOTF(Graph dataGraph) {
		super(dataGraph);
	}

	@Override
	/**
	 * our beforeRun_ : experimental implement concurrency 
	 */
	protected boolean beforeRun_() {
		
		concurrent_ = new ConcurrentService(options_.nbproc);
		
		return super.beforeRun_();
	}
	
	
	/**
	 * algorithm main entry point (with datagraph)
	 */
	protected Set<Graph> run_() {

		// init multigraph
		Graph multi = initMultiGraph_();

		// run recursive split
		// concurrent version
		
		Set<Graph> result = new HashSet<Graph>();

		if (concurrent_.isEnabled()) {
			concurrent_.addProcess(new DFSWorker(multi, result));
			concurrent_.waitCompletion();
			concurrent_.shutdown();
		}
		else {
			splitGraphDFS_(multi, result);
		}

		
		
		return result;
	}
	
	@Override
	/**
	 * our afterRun_ : reorder tuples in user's order and
	 * shift colors to fit with multigraph convention
	 * 
	 */
	protected Set<Graph> afterRun_(Set<Graph> partition) {
	
		// reorder tuples and shift multigraphs colors properly
		//
		for (Graph cc : partition) {
			reorderTuples_(cc);
			shiftColors_(cc, -1);
		}
		
		// uncompress connectons
		// TODO : the reporter should know about compression
		// TODO : or we change labels now and destroy aliases
		//
//		for (Graph cc : partition) {
//			compress.expand(cc);
//		}
		
		
		return partition;
	}
	
	
	@Override
	protected void logAlgoInfo_() {
		super.logAlgoInfo_();
		String deltashuf = options_.deltaShuf == null ? "[-1]" : options_.deltaShuf.toString();
		deltashuf = deltashuf.replaceAll("-1", "infinity");
		options_.logger.info("deltashuf:        " + deltashuf);
		options_.logger.info("nbproc:           " + concurrent_.getPoolSize());
		options_.logger.info("conc-threshold:   " + options_.concthresh);
	}
	
}
