package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph
 * provides methods for OTF algorithms (i.e OTF and OTFx thru subclasses)
 * 
 * This is the basic version of OTF algorithm, without any fancy code
 * related to concurrency or compression. it has been kept for
 * historical and pedagogical reasons.
 * 
 * experimental section : memory optimization is marked as @memoryOptimize
 * 
 */

//
// for readability, we use uppercase characters for variables
// holding nodes of the multigraph (e.g. V, U) and lowercase characters for
// nodes of the datagraph (also called polygraph) (e.g. v, u)
//

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.optimizer.ColorOptimizer;
import helix.graph.algo.c3part.preprocessor.Chromatizer;
import helix.graph.algo.c3part.preprocessor.MultiGraphCloser;
import helix.graph.algo.c3part.visitor.CCByColorVisitor;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RObject;

import java.util.HashSet;
import java.util.Set;

public abstract class C3PartAbstractOTFBasic extends C3PartAbstract {

	/*
	 * internal utilities
	 */

	/*
	 * Core methods
	 */
	
	/**
	 * split multigraph Clas into its connected components multisubgraphs on given color
	 */
	protected Set<Graph> splitGraphColor_(Graph Clas, int color) {

		// run the connected components visitor
		//
		CCVisitor visitor = new CCByColorVisitor(Clas, color);
		Iterable<GSet<Vertex>> cc = visitor.traverse().getComponents();
		
		// physically split graph
		// @memoryOptimize note: this destroys Clas
		//
		return Clas.physicalSplit(cc);
	}

	/**
	 * Get datagraph nodes suitable to expand multigraph vertex V
	 * 
	 */
	protected GSet<Vertex> terminalNodes_(Vertex V, int color) {
		return nodeAggregator_.getTerminals(V, color);
	}
	
	/**
	 * expand multinode V with new terminal t
	 */
	protected Vertex expandNode_(Vertex V, Vertex t, Graph multi) {
		
		RColorSet tColors = RColorSet.getColors(t);  // @NonStrict - see below -
		
		// make new multinode
		//
		Vertex NewV = multi.addVertex();
		
		// copy V nodes and colors
		//
		NewV.copy(V);				 				 // copy V Nodes (and colors)
		RColorSet.getColors(NewV).orColors(tColors); // @NonStrict
		
		// add terminal
		//
		C3PartUtil.addNode(NewV, t);

		return NewV;
	}

	/**
	 * Expand multigraph 'Clas' by adding 'color' nodes
	 * 
	 * Note: this version contains some extra code that
	 * is not really necessary for strict OTF. This code
	 * is intended to give a more general behavior in order
	 * to facilitate subclassing (especially for the OTFQuorum
	 * family). This is related to keeping colors
	 * in multigraph vertices. These colors indicate which colors
	 * of the datagraph nodes are kept in the multigraph
	 * vertex. In strict OTF all vertices have therefore the
	 * same colors (i.e. all the colors of the multigraph).
	 * This may be not true for some subclasses. E.g in OTFQuorum
	 * the star is colorless and therefore, the colors of vertices
	 * are subsets of the multigraph colors.
	 * This extra code is indicated below with the @NonStrict
	 * keyword.
	 */
	protected Graph expandMultiGraph_(Graph Clas, int color) {

		//;;; options_.debug.log("[expandMultiGraph_] Entering with " + Clas.nbVertices() + " vertices (actual=" + compress.nbVertices(Clas) + ") " + Clas.nbEdges() + " edges " + RColorSet.nbColors(Clas) + " colors on color " + color);
		
		Graph newClas = graphFactory_.make();
		newClas.copy(Clas);                           // copy colors
		RColorSet.getColors(newClas).setColor(color); // add new color

		GSet<Vertex> latest = setFactory_.make();
		GSet<Vertex> sonsOfNeighbors = setFactory_.make();
		
		for (Vertex V : Clas.vertices()) {

			// @NonStrict : in OTF, vertex colors are the same
			// as Clas colors. this may differ in OTFQuorum
			//
			RColorSet VColors = RColorSet.getColors(V);

			// get datagraph vertices to add == terminals
			//
			GSet<Vertex> terminals = terminalNodes_(V, color);
			latest.addAll(terminals); 

			//
			// terminals loop
			//
			
			for (Vertex v : terminals) {
				
				RColorSet vColors = RColorSet.getColors(v); // @NonStrict
				
				// -1- make a new multigraph vertex with this terminal
				//
				
				Vertex NewV = expandNode_(V, v, newClas);	 // new multigraph vertex
				
				// -2- restore initial Clas edges 
				//     with their initial colors
				//
				
				for (Edge E : V.edges()) {
					Vertex U = E.getVertexTo(V);
					//
					// in the following loop some U may not have sons yet.
					// this does not hurt since the edge will be created
					// later on.
					//
					for (Vertex NewU : C3PartUtil.getSons(U)) {
						Edge NewE = newClas.addEdge(NewV, NewU);
						NewE.copy(E); // copy colors
					}
				}
				
				// -3- add edges between brothers (i.e. sons of the same father)
				// these edges have all colors of father V since the initial
				// nodes are the same. @NonStrict : in strict OTF they all have
				// the colors of Clas.
				//
				// in the following loop some of brothers may not have
				// been born yet. this does not hurt since the edges will
				// be created later on.
				//
				for (Vertex NewU : C3PartUtil.getSons(V)) { // already born brothers
					Edge NewE = newClas.addEdge(NewV, NewU);
					RColorSet.getColors(NewE).setColors(VColors); // @NonStrict
				}
				
				// -4- add new edges corresponding to the terminal node v
				//     we should connect NewV to all sons of v and to all
				//     sons of v neighbors. note that only neighbors having
				//     already sons are eligible.
				//
				// @NonStrict: see -5- below : terminal nodes without color
				//             have no sons.
				//

				// -4.1- collect multigraph vertices to connect
				sonsOfNeighbors.clear();
				sonsOfNeighbors.addIterable(C3PartUtil.getSons(v));
				for (Vertex u : v.neighbors()) {
					if (C3PartUtil.hasSon(u)) {  // see -5- below 
						sonsOfNeighbors.addIterable(C3PartUtil.getSons(u));
					}
				}
				// -4.2- connect to them
				for (Vertex NewU : sonsOfNeighbors) {
					Edge E = newClas.assertEdge(NewV, NewU);
					RColorSet.getColors(E).setColor(color);
				}

				// -5- Finally register NewV as a son of V and v
				//
				if (! VColors.isEmpty()) {				 // @NonStrict : in OTF this should be always done
					C3PartUtil.addSon(V, NewV);			 // NewV is a son of V
				}
				
				if (vColors.hasColor(color)) {			// @NonStrict : in OTF this should be always done		
					C3PartUtil.addSon(v, NewV);			// NewV is son of v too
				}
				latest.add(v);
				
			} // end of v loop

			//;;; options_.debug.log("[expandMultiGraph_] current with " + newClas.nbVertices() + " vertices " + newClas.nbEdges() + " edges ");

		} // end of V loop
		
		//
		// cleanup : reset terminals in datagraph for next time
		//
		C3PartUtil.removeSons(latest);

		//
		// Note: the vertices in Clas have been modified (Sons added)
		//       and should not be reused as-is
		//       [AV] we may cleanup Sons here too (see below)
		//       but this is not needed since they will be garbage collected
		//
		//C3PartUtil.removeSons(Clas.vertices());
		
		//
		// experimental section :  keep local ordering (deltashuf)
		//
		if (options_.deltaShuf != null) {
			RColorSet multiColors = RColorSet.getColors(newClas);
			int maxColor = RColorSet.getColors(datagraph_).maxColor();
			MultiGraphCloser closer = new MultiGraphCloser(options_.deltaShuf, maxColor);
			closer.run(newClas);
			Chromatizer chromo = new Chromatizer(multiColors, maxColor, options_.deltaShuf);
			chromo.run(newClas);
		}

		//;;; options_.debug.log("[expandMultiGraph_] Ending with " + newClas.nbVertices() + " vertices (actual=" + compress.nbVertices(newClas) + ") " + newClas.nbEdges() + " edges ");

		return newClas;
	}
	
	/**
	 * filter multigraph when entering splitGraphDFS
	 * return true : split has to be continued
	 *       false : don't split this class
	 *       
	 * this filter is used here to check the mineltsize
	 * condition : if one genome (i.e. color) has
	 * less than mineltsize elements then we return
	 * false
	 */
	protected boolean beforeSplitGraphDFS_(Graph Clas) {

		GSet<Vertex> terminals = setFactory_.make();

		// check if for each color, the number of non-star nodes
		// is at least mineltsize
		
		int nbColors = tupleSize_(Clas);
		for (int indx = 0 ; indx < nbColors ; indx++) {
			terminals.clear();
			for (Vertex V : options_.compressor.getVertices(Clas)) { // @compress : Clas.vertices()
				Vertex v = C3PartUtil.getNodes(V).getVertex(indx);
				if (RColorSet.hasColors(v)) // @NonStrict
					terminals.add(v); 
				if (terminals.size() >= options_.mineltsize)
					break;
			}
			if (terminals.size() < options_.mineltsize)
				return false;  // condition not satisfied
		}
		
		return true; // condition satisfied for all colors
	}

	/**
	 * store class in CCC when leaving splitGraphDFS with a proper CCC
	 * return true : CCC has been stored
	 *       false : CCC should be discarded
	 * 
	 * here we only check the minsize condition
	 */
	protected boolean afterSplitGraphDFS_(Graph Clas, Set<Graph> ccc) {
		
		// @compress : Clas.nbVertices()
		if (options_.compressor.nbVertices(Clas) < options_.minsize) 
			return false;
		
		ccc.add(Clas);
		
		;;; options_.debug.log("@ After Stock OTF : " + ccc.size() + " " + Clas.nbVertices());
		
		return true;
	}

	/**
	 *  recursively split Clas on all colors (DFS approach).
	 *  store common connected components (as multigraphs) into ccc
	 *  if it is not done yet, Clas is extended to the colorth genome
	 *  
	 *  note: the @memoryOptimize comments indicate part of code that is not
	 *  fundamental in the algorithm but only to optimize memory usage
	 *  (i.e. to let the garbage collector work)
	 *  
	 */

	protected void splitGraphDFS_(Graph Clas, Set<Graph> ccc) {


		//;;; options_.debug.indentUp();
		//;;; options_.debug.log("[splitGraphDFS_] Entering with " + Clas.nbVertices() + " vertices " + Clas.nbEdges() + " edges " + RColorSet.nbColors(Clas) + " colors");
		//;;; options_.debug.log(Clas, false, "entering");

		// @memoryOptimize 
		// the current colors optimizer will never be used again so we can
		// tell it to free its internal components
		//
		C3PartUtil.getOptimizer(Clas).giveup(); // @memoryOptimize

		// precheck : experimental minEltSize option
		//
		if (! beforeSplitGraphDFS_(Clas)) {
			//;;; options_.debug.log("pruning");
			//;;; options_.debug.indentDown();
			Clas.clear(); // @memoryOptimize
			return;
		}

		// debug
		//;;; options_.debug.log("[splitGraphDFS_] optimizer " + Clas.get(C3PartUtil.Optimizer));
		
		// make a new color optimizer for this Clas
		// 
		ColorOptimizer colorsOptimizer = this.getOptimizerFor(Clas);
		C3PartUtil.setOptimizer(Clas, colorsOptimizer);
		
		// go ahead on remaining colors
		//
		while (colorsOptimizer.hasNext()) {
			
			int color = colorsOptimizer.next();
			
			if (RColorSet.hasColor(Clas, color)) {          // The multigraph is already computed 

				// try to split it on color just as in C3PartDFS
				// @memoryOptimize note: Clas will be destroyed 
				Set<Graph> split = splitGraphColor_(Clas, color);

				//;;; options_.debug.log("[splitGraphDFS_] Splitted on color " + color + " split size= " + split.size());
				
				if (split.size() == 1) {					// Clas unsplitted
					Clas = split.iterator().next();			//   continue to next color
				}
				else {										// Clas has been splitted
					for (Graph sub : split) {					// for all splits
						splitGraphDFS_(sub, ccc);				//   recurse to split subclasses on *all* colors
					}											//   (since split may disconnect previous colors)
					//;;; options_.debug.indentDown();
					return;
				}
			}
			else {                                          // The multigraph needs expansion

				// add color in multigraph
				//
				Clas = expandMultiGraph_(Clas, color);
	
				//;;; options_.debug.log("[splitGraphDFS_] Expanded on color " + color + " Clas size= " + Clas.nbVertices());
				//;;; options_.debug.log(Clas, false, "entering");
				
				// recurse to split expanded multigraph on *all* colors
				// (since expand can actually remove some vertices connecting previous colors)
				//
				// normal code is :
				// splitGraphDFS_(Clas, ccc);
				//
				// the following is a better variant :
				// we first try to split on latest color since it is likely to explode newClas
				// in smaller manageable pieces
				//
				Set<Graph> split = splitGraphColor_(Clas, color);
				//;;; options_.debug.log("[splitGraphDFS_] split on (latest) color " + color + " yields "+ split.size() + " classes");
				
				for (Graph sub : split) {
					splitGraphDFS_(sub, ccc);
				}
				
				//;;; options_.debug.indentDown();
				return;
			}
		} // end of colors loop

		// if we come here this means that 
		// Clas remains unsplitted after all colors => it's a CCC 
		
		//;;; options_.debug.log("[splitGraphDFS_] CCC Found ");
		//;;; options_.debug.log(Clas, false, "CCC");
		//;;; options_.debug.indentDown();
		
		afterSplitGraphDFS_(Clas, ccc);
		
		colorsOptimizer.giveup(); // @memoryOptimize : giveup with latest optimizer
	}
	
	/**
	 * algorithm : multigraph initialization
	 */
	protected Graph initMultiGraph_() {

		Graph multi = graphFactory_.make();

		// initialize multigraph registry
		// with the first color
		ColorOptimizer colors = this.getOptimizerFor(null); // first clas
		multi.set(C3PartUtil.Optimizer, new RObject(colors));

		int color = colors.next(); // optimizer will guess first color to use
		
		RColorSet.getColors(multi).setColor(color);

		// initialize partition with 1 class of all vertices
		// of the first color
		
		// creating vertices
		
		for (Vertex v : datagraph_.vertices()) {
			if (RColorSet.hasColor(v, color)) { 
				Vertex V = multi.addVertex();
				RColorSet.getColors(V).setColor(color);
				C3PartUtil.addNode(V, v);
				C3PartUtil.setMultinode(v, V); // we cheat a bit here ; see -1- below
			}
		}
		
		// creating edges

		for (Vertex V : multi.vertices()) {
			Vertex v = C3PartUtil.getLastNode(V);
			for (Edge e : v.edges()) {
				Vertex u = e.getVertexTo(v);
				if (! RColorSet.hasColor(u, color))
					continue;
				Vertex U = C3PartUtil.getMultinode(u); // -1- this is the reason why
				if (! multi.hasEdge(V, U)) {
					Edge E = multi.addEdge(V, U);
					E.copy(e);
				}
			}
		}
		
		// cleanup temporary father flag
		for (Vertex V : multi.vertices()) {
			Vertex v = C3PartUtil.getLastNode(V);
			C3PartUtil.removeMultinode(v);
		}
		
		return multi;
	}
	
	/*
	 * required C3PartAlgo API
	 */

	public C3PartAbstractOTFBasic(Graph dataGraph) {
		super(dataGraph);
	}

	/**
	 * algorithm main entry point (with datagraph)
	 */
	protected Set<Graph> run_() {

		// init multigraph
		Graph multi = initMultiGraph_();

		// run recursive split
		// concurrent version
		
		Set<Graph> result = new HashSet<Graph>();

		splitGraphDFS_(multi, result);
		
		return result;
	}
	
	@Override
	/**
	 * our afterRun_ : reorder tuples in user's order and
	 * shift colors to fit with multigraph convention
	 * 
	 */
	protected Set<Graph> afterRun_(Set<Graph> partition) {
	
		// reorder tuples and shift multigraphs colors properly
		//
		for (Graph cc : partition) {
			reorderTuples_(cc);
			shiftColors_(cc, -1);
		}
	
		return partition;
	}
	
	
	@Override
	protected void logAlgoInfo_() {
		super.logAlgoInfo_();
		String deltashuf = options_.deltaShuf == null ? "[-1]" : options_.deltaShuf.toString();
		deltashuf = deltashuf.replaceAll("-1", "infinity");
		options_.logger.info("deltashuf:        " + deltashuf);
	}
	
}
