package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph
 * provides base methods for OTFQLocal algorithm
 */

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.aggregator.NoSuchAggregatorException;
import helix.graph.algo.c3part.lookahead.AllPathConnector;
import helix.graph.algo.c3part.lookahead.ColorPathConnector;
import helix.graph.algo.c3part.lookahead.FullColorConnector;
import helix.graph.algo.c3part.lookahead.LookaheadConnector;
import helix.graph.algo.c3part.lookahead.OneWayPathConnector;
import helix.graph.algo.c3part.lookahead.TwoWayPathConnector;
import helix.graph.algo.c3part.visitor.CCStarByColorVisitor;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;

import java.util.Set;

public abstract class C3PartAbstractOTFQLocal extends C3PartAbstractOTFQuorum {

	protected LookaheadConnector lookahead_;

	
	/**
	 * get LookaheadConnector by name
	 */
	protected LookaheadConnector getLookaheadFor() {

		// mincore > 0 => don't care about user's value
		if (this.options_.mincore > 0)
			return ColorPathConnector.getInstance();
		
		String name = options_.lookaheadName;
		if (C3PartOptions.FULLCOLOR_LOOKAHEAD.compareToIgnoreCase(name) == 0)
			return FullColorConnector.getInstance();
		if (C3PartOptions.ALLPATH_LOOKAHEAD.compareToIgnoreCase(name) == 0)
			return AllPathConnector.getInstance();
		if (C3PartOptions.ONEWAYPATH_LOOKAHEAD.compareToIgnoreCase(name) == 0)
			return OneWayPathConnector.getInstance();
		if (C3PartOptions.TWOWAYPATH_LOOKAHEAD.compareToIgnoreCase(name) == 0)
			return TwoWayPathConnector.getInstance();
		throw new NoSuchAggregatorException("lookahead " + name + " is not available for algo " + this.getClass().getSimpleName());
	}
	
	/*
	 * Core methods
	 * AbstractC3PartOTF overriden methods
	 */
	
	@Override
	/**
	 * override splitGraphColor_ to add partial stars in usual CC's
	 */ 
	protected Set<Graph> splitGraphColor_(Graph Clas, int color) {
		
		// run the connected components visitor
		// TODO a resetable visitor 

		int colorIndex = C3PartUtil.getOptimizer(Clas).getColorIndex(color);

		CCVisitor visitor = new CCStarByColorVisitor(Clas, color, colorIndex, starVertex_);
		
		Iterable<GSet<Vertex>> cc = visitor.traverse().getComponents();
		
		// physically split graph
		//
		return Clas.physicalSplit(cc);
	}

	@Override
	/**
	 * override expandMultiGraph_ to add the lookahead 
	 * stars in expansion
	 */ 
	protected Graph expandMultiGraph_(Graph Clas, int color) {
		
		// call 'normal' expansion using our super
		//
		// note: it's not really normal expansion since
		// our terminalNodes_ will add end stars to normal nodes
		// and the whole color-th genome to the black star 
		//
		Graph multi = super.expandMultiGraph_(Clas, color);

		// if super expansion yields an empty graph then there is
		// no need to continue, this is a dead end
		//
		if (multi.nbVertices() == 0)
			return multi;
		
		// check if black star is present in the initial class
		//
		// note: in this case, the initial class is actually a singleton
		//       and the multi graph has been expanded on all vertices of
		//       the color-th genome so we don't need to expand further
		//
		for (Vertex V : Clas.vertices()) { // pseudo loop
			if (C3PartUtil.isFullStar(V))
				return multi; // TODO -> balck star has no colors !
			else
				break;  // full star is always a singleton
		}

		// check if we can add lookahead stars
		// i.e. if maxstar quorum is not over
		//
		if (tupleSize_(Clas) > options_.maxstar)
			return multi;
		
		// add lookahead stars : 
		//
		// At this point we should add the lookahead stars i.e. multigraph vertices
		// of the form (*, *...*, vi) where vi are the 'pseudoterminals'.
		// the main difficulty is to collect the pseudoterminals.
		// In order to experiment with various algorithms and implementations
		// these pseudoterminals are collected by using the current LookaheadConnector
		// that may be set thru command line (--lookahead option)
		//
		// case mincore > 0 : the CCC should guided by at least one core node. 
		//          then just add lookahead stars in front of each
		//			vertex in color-connected-components of current non star terminals.
		//			LookaheadConnector = ColorPathConnector
		//
		// case mincore == 0 : the CCC cannot be guided only by current terminals.
		//          we actually need to add lookahead stars in front of the whole color-th genome
		//			-> FullColorConnector
		//			Since this is very time consuming, several other algorithms are provided :
		//			AllPathConnector, TwoWayPathConnector, OneWayPathConnector
		//			they may be set from command line by usisng --lookahead
		//			allpath | oneway | twoway respectively
		//
		// ----------------------------------------------------------------------
		// important note: 
		//       all the three versions work with clique and center aggregators but may fail
		//       with other one : so... take care !
		// ----------------------------------------------------------------------
		//

		// -0- record latest non-star terminals and mark them as fathers
		//
		GSet<Vertex> latest = setFactory_.make();

		for (Vertex V : multi.vertices()) {
			Vertex v = C3PartUtil.getLastNode(V);
			if (v != starVertex_) {
				latest.add(v);
				C3PartUtil.addSon(v, V);
			}
		}
		
		// -1- collect nodes of color-th genome that need
		//     lookahead stars (i.e `pseudoterminals')
		//
		
		GSet<Vertex> pseudo;
		
		if (options_.mincore > 0) {
			pseudo = ColorPathConnector.getInstance().getLookahead(datagraph_, multi, color);
		}
		else {
			pseudo = lookahead_.getLookahead(datagraph_, multi, color);
		}
		
		//;;; System.err.println("@ latest.size=" + latest.size() + " pseudo.size=" + pseudo.size());

		// -2- make new multigraph vertices : (* ... * v) 
		//
		GSet<Vertex> newVertices = setFactory_.make();
		RVertexList StarTuple = makeStarTuple_(Clas);

		for (Vertex v : pseudo) {
			Vertex V = multi.addVertex();
			V.set(C3PartUtil.Nodes, StarTuple.klone());
			C3PartUtil.getNodes(V).addVertex(v);
			RColorSet.getColors(V).orColors(RColorSet.getColors(v));
			newVertices.add(V);
			C3PartUtil.addSon(v, V);
		}

		// -3- set edges between new vertices
		//
		GSet<Vertex> sonsOfNeighbors = setFactory_.make();
		for (Vertex v : pseudo) {
			sonsOfNeighbors.clear();
			sonsOfNeighbors.addIterable(C3PartUtil.getSons(v));
			for (Vertex u : v.neighbors()) {
				if (C3PartUtil.hasSon(u))
					sonsOfNeighbors.addIterable(C3PartUtil.getSons(u));
			}
			for (Vertex V : C3PartUtil.getSons(v)) {
				for (Vertex U : sonsOfNeighbors) {
					if (U == V) continue;
					Edge E = multi.assertEdge(U, V);
					RColorSet.getColors(E).setColor(color);
				}
			}
		}			

		// remove Son flag for later round
		//
		C3PartUtil.removeSons(latest); // huh latest should be included in pseudo ?
		C3PartUtil.removeSons(pseudo);
		
		return multi;
	}

	@Override
	/**
	 * override beforeSplitDFS_ to add the mincore condition
	 */ 
	protected boolean beforeSplitGraphDFS_(Graph Clas) {
		
		// check our super (for the mineltsize condition)
		//
		if (! super.beforeSplitGraphDFS_(Clas))
			return false;

		// add the mincore condition
		//
		if (options_.mincore > 0) {
			int count = 0;
			for (Vertex V : options_.compressor.getVertices(Clas)) {
				if (C3PartUtil.isCoreNode(V))
					count++;
				if (count >= options_.mincore)
					break;
			}
			if (count < options_.mincore)
				return false;
		}

		// all tests passed
		return true;
	}
	
	/*
	 * required C3PartAlgo API
	 */

	public C3PartAbstractOTFQLocal(Graph dataGraph) {
		super(dataGraph);
	}
	
	@Override
	/**
	 * our beforeRun_ : setup proper lookahead
	 */
	protected boolean beforeRun_() {

		// setup proper lookahead connector

		lookahead_ = getLookaheadFor();
		
		return super.beforeRun_();
	}
	
	@Override
	/**
	 * 
	 */
	protected void logAlgoInfo_() {
		
		super.logAlgoInfo_();
		options_.logger.info("mincore:          " + options_.mincore);
		options_.logger.info("lookahead_class:  " + this.lookahead_.getClass().getSimpleName());
	}
	
}
