package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph
 * provides base methods for OTF-Quorum algorithm (i.e OTF with ghost star)
 * such as OTFQLocal and OTFQGlobal
 * 
 */

import helix.graph.algo.c3part.aggregator.CCAggregator;
import helix.graph.algo.c3part.aggregator.CenteredAggregator;
import helix.graph.algo.c3part.aggregator.CliqueAggregator;
import helix.graph.algo.c3part.aggregator.DenseCCAggregator;
import helix.graph.algo.c3part.aggregator.NoSuchAggregatorException;
import helix.graph.algo.c3part.aggregator.NodeAggregator;
import helix.graph.algo.c3part.optimizer.ColorOptimizer;
import helix.graph.algo.c3part.optimizer.TerminalsStarColorOptimizer;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;

import java.util.HashSet;
import java.util.Set;

public abstract class C3PartAbstractOTFQuorum extends C3PartAbstractOTF {

	/**
	 * a fake ghost star that should be added to current datagraph
	 * by the subclass.
	 * TODO fixme - we should use the transparent color whenever we could.
	 */
	protected Vertex starVertex_;
	
	/**
	 * a fake ghost tuple
	 */
	protected RVertexList starTuple_ = new RVertexList();
	
	/**
	 * a flag for graph containing a star free Node
	 * such a Node is called a 'core Node'
	 */
	protected static String HAS_CORE_NODE = "C3P_CoreNode";
	
	/**
	 * some aggregators (i.e CC) also need the
	 * current multigraph colors
	 */
	RColorSet multicolors_ = null;
	
	/*
	 * internal utilities
	 */
	
	/**
	 * tell if multigraph has at least one core node
	 */
	protected boolean hasCoreNode_(Graph multi) {
		for (Vertex V : multi.vertices()) {
			if (C3PartUtil.isCoreNode(V))
				return true;
		}
		return false;
	}
	
	/**
	 * make a black star tuple of suitable size
	 */
	protected RVertexList makeStarTuple_(Graph multi) {
		int size = tupleSize_(multi);
		if (starTuple_.size() != size) {
			starTuple_.clear();
			for (int i = 0 ; i < size ; i++) {
				starTuple_.addVertex(starVertex_);
			}
		}
		return starTuple_;
	}

	/*
	 * tell if a multigraph vertex V1 is included in multigraph vertex V2
	 * V1 included V2 <=> foreach i : v1i == v2i or v1i = *
	 *                    and V1 != V2 (physically)
	 * TODO [AV] what if V1 logically equals V2 ? => both will disappear
	 */
	protected boolean isIncluded_(Vertex V1, Vertex V2) {
		if (V1 == V2) return false;

		RVertexList t1 = C3PartUtil.getNodes(V1);
		RVertexList t2 = C3PartUtil.getNodes(V2);
		int size = t1.size();
		for (int i = 0 ; i < size ; i++) {
			Vertex v1 = t1.getVertex(i);
			Vertex v2 = t2.getVertex(i);
			if (v1 == v2) continue;
			if (v1 == starVertex_) continue;
			return false;
		}
		// we passed all tests
		return true;
	}
	
	/*
	 * tell if a multigraph vertex V is included in multigraph G
	 * <=> V is included in at least one vertex of G
	 */
	protected boolean isIncluded_(Vertex V, Graph G) {
		for (Vertex U : G.vertices()) {
			if (isIncluded_(V, U))
				return true;
		}
		// we passed all tests
		return false;
	}

	/*
	 * tell if a multigraph G1 is included in multigraph G2
	 * TODO [AV] what if G1 logically equals G2 ? => both will disappear
	 */
	protected boolean isIncluded_(Graph G1, Graph G2) {
		if (G1 == G2) return false;
		for (Vertex V1 : G1.vertices()) {
			if (! isIncluded_(V1, G2))
				return false;
		}
		// we passed all tests
		return true;
	}
	
	/*
	 * clean up degenerated vertices from CCC
	 * TODO [AV] what if V1 logically equals V2 ? => both will disappear
	 */
	protected void cleanupIntra_(Graph ccc) {
		GSet<Vertex> removed = setFactory_.make();
		for (Vertex V : ccc.vertices()) {
			for (Vertex U : ccc.vertices()) {
				if (isIncluded_(V, U))
					removed.add(V);
			}
		}
		for (Vertex V : removed) {
			ccc.removeVertex(V);
		}
	}
	
	
	/*
	 * Core methods
	 * AbstractC3PartOTF overriden methods
	 */

	@Override
	/**
	 * override terminalNodes_ to add a ghost star
	 * as a terminal and to produce terminals of black star
	 */ 
	protected GSet<Vertex> terminalNodes_(Vertex V, int color) {

		// get terminals the usual way
		//
		GSet<Vertex> terminals = super.terminalNodes_(V, color);
		
		// black star case (note: in that case previous terminals should be empty)
		//
		if (C3PartUtil.isFullStar(V)) {
			for (Vertex v : datagraph_.vertices()) {
				if (RColorSet.hasColor(v, color))
					terminals.add(v);
			}
		}

		// add ghost star as a terminal
		//
		int nbStar = C3PartUtil.countStars(V);

		if (nbStar < options_.maxstar) {
			if (nodeAggregator_.isAggregable(V,starVertex_)) terminals.add(starVertex_);
		}
		
		return terminals;
	}
	
	@Override
	/**
	 * override expandMultiGraph_ to keep track of multigraph colors.
	 * this is needed by some aggregators
	 */ 
	protected Graph expandMultiGraph_(Graph Clas, int color) {

		if (multicolors_ != null)
			multicolors_.setColors(RColorSet.getColors(Clas));
		
		// call 'normal' expansion using our super
		return super.expandMultiGraph_(Clas, color);
	}

	@Override
	/**
	 * override afterSplitGraphDFS_ to add CCC intra and inter cleanup
	 */ 
	protected boolean afterSplitGraphDFS_(Graph Clas, Set<Graph> ccc) {

		// TODO: choix a faire : renvoyer une partition de vertices
		//                       ou un graphe sans arete
		options_.compressor.flatten(Clas);
		
		// cleanup intra vertices
		cleanupIntra_(Clas);

		// test minsize condition
		if (Clas.nbVertices() < options_.minsize)
			return false;

		// reorder tuples to compare with already stored CCC
		reorderTuples_(Clas);

		// check if this one has a core node
		if (hasCoreNode_(Clas))
			Clas.set(HAS_CORE_NODE);

		// if Clas does not have a core node
		// first check inclusion of Clas in one already defined ccc
		if (! Clas.has(HAS_CORE_NODE)) {
			for (Graph g : ccc) {
				if (isIncluded_(Clas, g))
					return false; // discard Clas
			}
		}

		// check reverse condition : inclusion of a cc in Clas
		Set<Graph> removed = new HashSet<Graph>();
		for (Graph g : ccc) {
			if (g.has(HAS_CORE_NODE))
				continue;
			if (isIncluded_(g, Clas))
				removed.add(g);
		}
		ccc.removeAll(removed);

		// finally add Clas
		ccc.add(Clas);

		//;;; options_.debug.log("@ After Stock OTFQuorum : " + ccc.size() + " " + Clas.nbVertices());
		return true;
	}

	@Override
	/**
	 * override initMultiGraph_ to add a star node in multigraph
	 */ 
	protected Graph initMultiGraph_() {

		// init using super
		//
		Graph multi = super.initMultiGraph_();
		
		// add a black star node -> without color
		//
		if (options_.maxstar > 0) {
			Vertex V = multi.addVertex();
			RColorSet.getColors(V); // will force to add an empty colorset
			C3PartUtil.addNode(V, starVertex_);
		}
		
		return multi;
	}
	
	/*
	 * required C3PartAlgo API
	 */

	public C3PartAbstractOTFQuorum(Graph dataGraph) {
		super(dataGraph);
	}

	@Override
	/**
	 * our beforeRun_ : add fake star vertex in datagraph
	 */
	protected boolean beforeRun_() {

		// add a fake star vertex in datagraph
		
		starVertex_ = datagraph_.addVertex();
		starVertex_.set(Registry.Label, "*");
		
		return super.beforeRun_();
	}
	

	/**
	 * algorithm main entry point (with datagraph)
	 * we should remove the fake star vertex from datagraph
	 */
	protected Set<Graph> run_() {

		// call our super
		
		Set<Graph> result = super.run_();
		
		// remove fake star vertex
		datagraph_.removeVertex(starVertex_);
		
		return result;
	}
	
	@Override
	/**
	 * our afterRun_ : shift colors to fit with multigraph convention
	 * note: we don't reorder tuples since this has been done during
	 * recursion.
	 */
	protected Set<Graph> afterRun_(Set<Graph> partition) {
	
		for (Graph cc : partition) {
			shiftColors_(cc, -1);
		}
		
		return partition;
	}

	
	@Override
	/**
	 * get special aggregator for OTFQuorum family
	 * TODO CenterStar aggregator
	 */
	public NodeAggregator getAggregatorFor() {

		// avoid multiple instanciation
		if (nodeAggregator_ == null) {

			String name = options_.nodeAggregatorName;
			if (C3PartOptions.CLIQUE_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new CliqueAggregator(datagraph_, options_);
			else if (C3PartOptions.CENTER_AGGREGATOR.compareToIgnoreCase(name) == 0)
				nodeAggregator_ = new CenteredAggregator(datagraph_, options_);
			else if (C3PartOptions.CC_AGGREGATOR.compareToIgnoreCase(name) == 0) {
				multicolors_ = new RColorSet();
				nodeAggregator_ = new CCAggregator(datagraph_, options_, multicolors_);
			}
			else if (C3PartOptions.DENSE_AGGREGATOR.compareToIgnoreCase(name) == 0) {
				multicolors_ = new RColorSet();
				nodeAggregator_ = new DenseCCAggregator(datagraph_, options_, multicolors_);
			}
			else
				throw new NoSuchAggregatorException("aggregator " + name + " is not available for algo " + this.getClass().getSimpleName());
		}

		return nodeAggregator_;
	}
	
	
	@Override
	/**
	 * added API
	 * make the proper optimizer for this algorithm
	 */
	public ColorOptimizer getOptimizerFor(Graph Clas) {
		String name = options_.optimizerName;
		
		if (C3PartOptions.LOCAL_OPTIMIZER.compareToIgnoreCase(name) == 0) {
			return new TerminalsStarColorOptimizer(this, datagraph_, Clas);
		}
		
		return super.getOptimizerFor(Clas);
	}
	
	
	@Override
	/**
	 * 
	 */
	protected void logAlgoInfo_() {
		
		super.logAlgoInfo_();

		options_.logger.info("maxstar:          " + options_.maxstar);
	}
}
