package helix.graph.algo.c3part;

/**
 * Compute common connected components of a data graph.
 * provides base methods for all xFS algorithms (i.e BFS and DFS).
 * 
 * this family of algorithms is intended to be design on multigraph.
 * in order to simplify your code, we implement the wrapping of <code>run_()</code>
 * to obey the C3PartAlgo interface by calling a <code>runMulti_(Graph multi, C3PartOptions options)</code>.
 * so you just have to implement this method.
 * note that for internal tests urpose, the <code>runMulti_</code> method can be called
 * directly too.
 */

//
// for readability, we use uppercase characters for variables
// holding nodes of the multigraph (e.g. V, U) and lowercase characters for
// nodes of the datagraph (also called polygraph) (e.g. v, u)
//

import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;
import helix.graph.util.IPermutation;

import java.util.Set;

public abstract class C3PartAbstractXFS extends C3PartAbstract {

	/**
	 * internal utility 
	 * recursively compute the connected component on given color containing Vertex V
	 * restricted to clas.
	 * store the result into cc
	 */
	protected final void component__(Vertex V, GSet<Vertex> clas, int color, GSet<Vertex> cc) {
		
		if (cc.contains(V))
			return;
		
		if (! clas.contains(V))
			return;
		
		cc.add(V);
		
		for (Edge E : V.edges()) {
			if (! RColorSet.hasColor(E, color))
				continue;
			Vertex U = E.getVertexTo(V);
			component__(U, clas, color, cc);
		}
	}

	/**
	 * compute the connected component on given color containing Vertex V
	 * restricted to Clas.
	 */
	protected final GSet<Vertex> component_(Vertex V, GSet<Vertex> Clas, int color) {
		
		GSet<Vertex> cc = setFactory_.make();
		
		component__(V, Clas, color, cc);
		
		return cc;
	}

	/**
	 * split Clas into its connected components on given color
	 */
	protected Set<GSet<Vertex>> splitSetColor_(GSet<Vertex> Clas, int color) {
		
		Set<GSet<Vertex>> result = setFactory_.makeSuper();
		GSet<Vertex> visited = setFactory_.make();
		
		for (Vertex V : Clas) {
			if (visited.contains(V))
				continue;
			GSet<Vertex> cc = component_(V, Clas, color);
			visited.addAll(cc);
			result.add(cc);
		}
		
		return result;
	}
	
	/**
	 * make a multigraph from datagraph
	 * using C3PartDTM algo
	 */
	protected Graph makeMultigraph_() {
		
		C3PartDTM dtm = new C3PartDTM(datagraph_);
		
		// run DTM with options. this returns a singleton = {multigraph}
		//
		dtm.setOptions(this.options_);
		Set<Graph> singleton = dtm.run();
		Graph multi = singleton.iterator().next();

		// TODO call options_.memoryWatcher.notify() since we probably use a lot of memory there
		
		// shift colors back
		// because DTM returns a proper multigraph with colors starting at 0
		// and we want the colors to fit with the one in datagraph
		//
		shiftColors_(multi, 1);
		
		return multi;
	}
	
	/**
	 * utility : check if multigraph vertices have RVertexList nodes
	 * if not, add a singleton RVertexList pointing to the vertex itself.
	 * 
	 */
	protected void assertMultiNodes(Graph multi) {
		for (Vertex V : multi.vertices()) {
			if (! V.has(C3PartUtil.Nodes)) {
				V.set(C3PartUtil.Nodes, new RVertexList(V));
			}
		}
	}
	
	/*
	 * C3PartAlgo API
	 */
	
	/**
	 * required pseudo constructor
	 */
	
	public C3PartAbstractXFS(Graph dataGraph) {
		super(dataGraph);
	}
	
	/**
	 * algorithm main entry point (with datagraph or multigraph)
	 * this family of algorithms is intended to be design on multigraph.
	 * in order to simplify your code, we implement the wrapping of <code>run_()</code>
	 * to obey the C3PartAlgo interface by calling a <code>runMulti_(Graph multi, C3PartOptions options)</code>.
	 * so you just have to implement this method.
	 */
	protected Set<Graph> run_() {

		// make single multigraph from datagraph using DTM
		//
		Graph multi;
		if (options_.ismulti) { // input is already a multigraph
			multi = datagraph_;
			assertMultiNodes(multi); // AV is this really necessary ?
		}
		else {					// normal input is a datagraph => convert it to mutigraph
			multi = this.makeMultigraph_();
		}
		
		// now perform B/D/FS algo on the multigraph to compute CCC
		// you have to implement this method
		//

		Set<GSet<Vertex>> ccc = this.runMulti_(multi, this.options_);
		
		// finaly convert CCC as subgraphs to obey the C3PartAlgo interface
		//
		return multi.physicalSplit(ccc);
	}

	@Override
	/**
	 * our afterRun_ : reorder tuples in multi vertices
	 */
	protected Set<Graph> afterRun_(Set<Graph> partition) {
	
	// reorder tuples and shift multigraphs colors properly
	//
		for (Graph cc : partition) {
			reorderTuples_(cc);
		}
		
		return super.afterRun_(partition);
	}

	@Override
	public void setOptions(C3PartOptions options) {


		if (! (options.optimizerName.equalsIgnoreCase(C3PartOptions.USER_OPTIMIZER)))  {
				options.optimizerName = C3PartOptions.USER_OPTIMIZER;
				if (options.userColors != null) {
					options.colors = new IPermutation(1, options.userColors.size());
				}
				else {
					int nbColors = RColorSet.nbColors(datagraph_);
					options.colors = new IPermutation(1, nbColors - (options.ismulti ? 0 : 1));
				}
				options.logger.severe(  "xfs family algorithms do not use color optimizer"
						               + " (will be forced to " + options.colors + ")");
		}
		
		if (options.mineltsize != 1) {
			options.mineltsize = 1;
			options.logger.severe(  "sorry... xfs family algorithms do not use the mineltsize option"
		               + " (will be forced to " + options.mineltsize + ")");
		}

		super.setOptions(options);
	}


	/*
	 * additionnal API
	 */
	
	/**
	 * to be implemented by actual algorithm
	 */
	protected abstract Set<GSet<Vertex>> runMulti_(Graph multi, C3PartOptions options);
	
 }
