package helix.graph.algo.c3part;

import helix.graph.algo.c3part.compressor.MultiGraphCompressor;
import helix.graph.algo.c3part.compressor.NoCompressor;
import helix.graph.algo.c3part.compressor.MultiIdentCompressor;
import helix.graph.algo.c3part.compressor.MultiInColorsCompressor;
import helix.graph.algo.c3part.compressor.MultiInSetColorsCompressor;
import helix.graph.algo.c3part.compressor.MultiInSetCompressor;
import helix.graph.model.Registry;
import helix.graph.util.CommandLine;
import helix.graph.util.CommandLineException;
import helix.graph.util.GraphDebugger;
import helix.graph.util.IPermutation;
import helix.graph.util.LogUtil;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * options values for C3Part algorithms
 * @author viari
 *
 */
public class C3PartOptions extends CommandLine {
	
	protected static String LOGGER_PREFIX = "C3PartLogger";
	protected static String DEBUG_PREFIX = "C3PartDebugger";

	public static String CLIQUE_AGGREGATOR = "clique";
	public static String CENTER_AGGREGATOR = "center";
	public static String CC_AGGREGATOR = "cc";
	public static String DENSE_AGGREGATOR = "dense";
	public static String TREE_AGGREGATOR = "tree"; // future use

	public static String DENSE_MODE_RELATIVE = "relative";
	public static String DENSE_MODE_ABSOLUTE = "absolute";
	public static String DENSE_MODE_QUANTILE = "quantile"; // future use
	
	public static String NO_OPTIMIZER = "off";
	public static String LOCAL_OPTIMIZER = "local";
	public static String GLOBAL_OPTIMIZER = "global";
	public static String USER_OPTIMIZER = "user";

	public static String OPTIMIZER_MODE_MIN = "min";
	public static String OPTIMIZER_MODE_MAX = "max";
	
	public static String FULLCOLOR_LOOKAHEAD = "allofcolor";
	public static String ALLPATH_LOOKAHEAD = "allpath";
	public static String ONEWAYPATH_LOOKAHEAD = "oneway";
	public static String TWOWAYPATH_LOOKAHEAD = "twoway";

	public static String NO_COMPRESSOR = "off";
	public static String IDENT_COMPRESSOR = "ident";
	public static String INCOLOR_COMPRESSOR = "incolor";
	public static String INSETCOLOR_COMPRESSOR = "insetcolor";
	public static String INSET_COMPRESSOR = "inset";
	
	protected static String DFT_AGGREGATOR = CLIQUE_AGGREGATOR;
	protected static String DFT_OPTIMIZER = LOCAL_OPTIMIZER;
	protected static String DFT_LOOKAHEAD = ONEWAYPATH_LOOKAHEAD;
	protected static String DFT_COMPRESSOR = NO_COMPRESSOR;
	protected static String DFT_OPTIMIZER_MODE = OPTIMIZER_MODE_MIN;
	protected static String DFT_DENSE_MODE = DENSE_MODE_RELATIVE;

	protected static boolean DFT_VERBOSE = false;
	protected static boolean DFT_ISMULTI = false;
	protected static int DFT_MINSIZE = 1;
	protected static int DFT_MINELTSIZE = 0;   // experimental option for C3PartOTF(Q) only
	protected static int DFT_MAXSTAR = 0;      // experimental option for C3PartOTFQuorum only
	protected static int DFT_MINCORE = 0;      // experimental option for C3PartOTFQuorum only
	protected static int DFT_NBPROC = 1;       // experimental concurrent for C3PartOTF(Q) only
	protected static int DFT_CONC_THRESH = 50; // experimental concurrent for C3PartOTF(Q) only
	protected static int DFT_DENSITY = 50;     // experimental option for CC_DENSE aggregator
	
	public boolean verbose = DFT_VERBOSE;
	public boolean ismulti = DFT_ISMULTI;	 // for C3PartXFS only
	public int minsize = DFT_MINSIZE;
	public int nbproc = DFT_NBPROC;          // experimental concurrent for C3PartOTF(Q) only
	public int concthresh = DFT_CONC_THRESH; // experimental concurrent for C3PartOTF(Q) only
	public int mineltsize = DFT_MINELTSIZE;  // experimental option for C3PartOTF(Q) only
	public int maxstar = DFT_MAXSTAR;        // experimental option for C3PartOTFQuorum only
	public int mincore = DFT_MINCORE;        // experimental option for C3PartOTFQuorum only

	public IPermutation colors = null;
	public Set<Integer> userColors = null;
	
	public String nodeAggregatorName = DFT_AGGREGATOR;
	public String optimizerName = DFT_OPTIMIZER;
	public String optimizerMode = DFT_OPTIMIZER_MODE;
	public String lookaheadName = DFT_LOOKAHEAD;
	public String ccDenseMode = DFT_DENSE_MODE;
	
	
	public List<Integer> deltaShuf = null;	// experimental option delta shuffle
	
	public Logger logger = LogUtil.getUniqueLogger(LOGGER_PREFIX);
	
	public GraphDebugger debug = new GraphDebugger(LogUtil.getUniqueLogger(DEBUG_PREFIX));
	
	public MultiGraphCompressor compressor = getCompressorFor_(DFT_COMPRESSOR);
	
	public int ccDensity = DFT_DENSITY;
	
	
	/**
	 * default constructor with user's command line arguments.
	 * use as : 
	 * <pre><code>
	 * public static void main (String[] args) {
	 * C3PartOptions options = new C3PartOptions(args);
	 * options.parse(); // will load user's arguments into options fields
	 * </code></pre>
	 */
	public C3PartOptions(String[] args) {

		// super only defines 'h' (help) option 
		super(args);

		// new options
		addOption("debug", 'g').setOptionalArgument().setDescription("[<?>=on|off] turn on/off debugger (default: off)");
		addOption("verbose", 'v').setOptionalArgument().setDescription("[<?>=on|off] turn on/off verbosity (default: " + boolToString(DFT_VERBOSE) + ")");
		addOption("optimizer", 'o').setUnlimitedArguments().setDescription("[<?>=off|local [min|max]|global [min|max]|user <colors>] select colors permutation optimization (default " +  DFT_OPTIMIZER + " " + DFT_OPTIMIZER_MODE + ")");
		addOption("ismulti", 'U').setOptionalArgument().setDescription("[<?>=on|off] tell input graph is already a multigraph (xfs algo only) (default " +  boolToString(DFT_ISMULTI) + ")");
		addOption("nbproc", 'P').setUnlimitedArguments().setDescription("[<?>=integer] nb of concurrent processors to use (1=no-concurrency 0=max-available (default " +  DFT_NBPROC + ")");
		addOption("colors", 'c').setUnlimitedArguments().setDescription("[<?>=integer+] keep only specified colors for processing (default keep all colors)");
		addOption("minsize", 's').setRequiredArgument().setDescription("[<?>=integer] ccc min size (default " +  DFT_MINSIZE + ")");
		addOption("mineltsize", 'S').setRequiredArgument().setDescription("[<?>=integer] ccc min element size (default " +  DFT_MINELTSIZE + ")");
		addOption("aggregator", 'a').setUnlimitedArguments().setDescription("[<?>=clique|center|cc|dense [[mode] density]] set node aggregator (default " + DFT_AGGREGATOR + ")");
		addOption("maxstar", 'm').setRequiredArgument().setDescription("[<?>=integer] set the maximum number of allowed stars (otfquorum algo) (default " +  DFT_MAXSTAR + ")");
		addOption("mincore", 'M').setRequiredArgument().setDescription("[<?>=integer] set the minimum number of core nodes (otfquorum algo) (default " +  DFT_MINCORE + ")");
		addOption("deltashuf", 'd').setUnlimitedArguments().setDescription("[<?>=integer+] delta for shuffle on each color. use 0 for no shuffling, use \"inf(inity)?\" for infinite shuffling. " +
				                                                           "if you specify a single value, it will apply on all colors (default: \"infinity\" (i.e. infinite shuffling on all colors))");
		addOption("lookahead", 'l').setRequiredArgument().setDescription("[<?>=terminals|allofcolor|colorpath | allpath|oneway|twoway] set lookahead connector (for \"otfql\" algo)" +
																		   " !for developers only! (default " + DFT_LOOKAHEAD + ")");
		addOption("compressor", 'z').setOptionalArgument().setDescription("[<?>=off|ident|incolor] select graph compression mode (default: " + DFT_COMPRESSOR + ")");
		LogUtil.setVerbose(logger, verbose);
		LogUtil.setHeader(logger, "+ [C3Part] ");

		LogUtil.setVerbose(debug.getLogger(), false);
		LogUtil.setHeader(debug.getLogger(), "+ [Debug] ");
	}
	
	/**
	 * constructor without any command line (used to make default options)
	 * note: of course you don't need to call parse() just after
	 */
	public C3PartOptions() {
		this(new String[0]);
	}
	
	/**
	 * constructor to copy existing options
	 */
	public C3PartOptions(C3PartOptions options)  {
		this();
		try {
			for (Field f : this.getClass().getFields()) {
				if ((f.getModifiers() & Modifier.STATIC) != 0)
					continue; // don't copy static fields
				f.set(this, f.get(options));
			}
		} catch (Exception e) {}
	}

	/**
	 * override stubs
	 */
	@Override
	public void set_help(List<String> args) {
		logger.setLevel(Level.INFO);
		logger.info(command_.getHelp());
		System.exit(0);
	}
	
	/*
	 * helper methods
	 */
	
	/*
	 * new stubs
	 */

	public void set_debug(List<String> args) {
		boolean hasDebug = (args.size() == 0) ? true : stringToBool(args.get(0));
		LogUtil.setVerbose(debug.getLogger(), hasDebug);
		if (hasDebug) {
			// TODO choose debug file;
			debug.addHandler("c3part.log"); 
			// TODO tell what is to debug
			debug.addVertexAttribute(Registry.Label);
			debug.addVertexAttribute(C3PartUtil.Nodes);
		}
	}

	public void set_verbose(List<String> args) {
		verbose = (args.size() == 0) ? true : stringToBool(args.get(0));
		LogUtil.setVerbose(logger, verbose);
	}

	public void set_optimizer(List<String> args) {
		
		optimizerName = args.get(0);

		if (   (NO_OPTIMIZER.compareToIgnoreCase(optimizerName) != 0)
				&& (LOCAL_OPTIMIZER.compareToIgnoreCase(optimizerName) != 0)
			    && (GLOBAL_OPTIMIZER.compareToIgnoreCase(optimizerName) != 0)
			    && (USER_OPTIMIZER.compareToIgnoreCase(optimizerName) != 0))
			throw new CommandLineException("unknown optimizer " + optimizerName);
		
		if (USER_OPTIMIZER.compareToIgnoreCase(optimizerName) == 0) {
			StringBuffer buf = new StringBuffer();
			for (int i = 1 ; i < args.size() ; i++) { 
				buf.append(args.get(i) + " ");
			}
			colors = new IPermutation(buf.toString());
		}

		if (   (   (LOCAL_OPTIMIZER.compareToIgnoreCase(optimizerName) == 0)
		        || (GLOBAL_OPTIMIZER.compareToIgnoreCase(optimizerName) == 0))
		    && (args.size() > 1)) {
			optimizerMode = args.get(1);
			if (   (OPTIMIZER_MODE_MIN.compareToIgnoreCase(optimizerMode) != 0)
			    && (OPTIMIZER_MODE_MAX.compareToIgnoreCase(optimizerMode) != 0))
				throw new CommandLineException("unknown optimizer mode " + optimizerMode);
		}
		
	}
	
	
	public void set_ismulti(List<String> args) {
		ismulti = (args.size() == 0) ? true : stringToBool(args.get(0));
	}
	
	public void set_nbproc(List<String> args) {
		nbproc = Integer.parseInt(args.get(0));
		if (nbproc < 0)
			throw new CommandLineException("nbproc should be >= 0");
		if (args.size() > 1)
			concthresh = Integer.parseInt(args.get(1));
		if (concthresh < 0)
			throw new CommandLineException("concurrent-threshold should be > 0");
	}
	
	public void set_colors(List<String> args) {
		userColors = new HashSet<Integer>();
		for (String str : args) {
			int color = Integer.parseInt(str);
			if (color <= 0)
				throw new CommandLineException("colors should be > 0");
			userColors.add(color);
		}
		if (userColors.size() < 2)
			throw new CommandLineException("you should keep at least two colors");
	}

	public void set_minsize(List<String> args) {
		minsize = Integer.parseInt(args.get(0));
		if (minsize < 1)
			throw new CommandLineException("minsize should be >= 1");
	}

	public void set_mineltsize(List<String> args) {
		mineltsize = Integer.parseInt(args.get(0));
		if (mineltsize < 0)
			throw new CommandLineException("mineltsize should be >= 0");
	}

	private final int getDensity_(String arg, boolean isPercent) {
		int density = Integer.parseInt(arg);
		if (isPercent) {
			if ((density < 0) || (density > 100))
				throw new CommandLineException("density should be in range [0, 100]");
		}
		else {
			if (density < 0)
				throw new CommandLineException("absolute density should be a positive integer");
		}
		return density;
	}
	
	public void set_aggregator(List<String> args) {
		
		nodeAggregatorName = args.get(0);
		
		if (       (CLIQUE_AGGREGATOR.compareToIgnoreCase(nodeAggregatorName) != 0)
			    && (CENTER_AGGREGATOR.compareToIgnoreCase(nodeAggregatorName) != 0)
		        && (CC_AGGREGATOR.compareToIgnoreCase(nodeAggregatorName) != 0)
		        && (DENSE_AGGREGATOR.compareToIgnoreCase(nodeAggregatorName) != 0))
		throw new CommandLineException("unknown aggregator " + nodeAggregatorName);

		if (DENSE_AGGREGATOR.compareToIgnoreCase(nodeAggregatorName) == 0) {
			if (args.size() == 1) {
				// use default values
			}
			else if (args.size() == 2) { // for backward compatibility : "dense nn" <=> "dense relative nn"
				ccDenseMode = DENSE_MODE_RELATIVE;
				ccDensity = getDensity_(args.get(1), true);
			}
			else {
				String mode = args.get(1).substring(0, 3);
				if (DENSE_MODE_RELATIVE.substring(0, 3).compareToIgnoreCase(mode) == 0) {
					ccDenseMode = DENSE_MODE_RELATIVE;
					ccDensity = getDensity_(args.get(2), true);
				}
				else if (DENSE_MODE_ABSOLUTE.substring(0, 3).compareToIgnoreCase(mode) == 0) {
					ccDenseMode = DENSE_MODE_ABSOLUTE;
					ccDensity = getDensity_(args.get(2), false);
				}
				else {
					throw new CommandLineException("unknown dense aggregator mode " + mode);
				}
			}
		}
	}

	public void set_maxstar(List<String> args) {
		maxstar = Integer.parseInt(args.get(0));
		if (maxstar < 0)
			throw new CommandLineException("maxstar should be >= 0");
	}

	public void set_mincore(List<String> args) {
		mincore = Integer.parseInt(args.get(0));
		if (mincore < 0)
			throw new CommandLineException("mincore should be >= 0");
	}

	public void set_deltashuf(List<String> args) {
		deltaShuf = new ArrayList<Integer>();
		for (String str : args) {
			int delta = stringToInt(str, -1);
			deltaShuf.add(delta);
		}
	}

	public void set_lookahead(List<String> args) {
		lookaheadName = args.get(0);
		if (   (FULLCOLOR_LOOKAHEAD.compareToIgnoreCase(lookaheadName)  != 0)
		    && (ALLPATH_LOOKAHEAD.compareToIgnoreCase(lookaheadName)    != 0)
		    && (ONEWAYPATH_LOOKAHEAD.compareToIgnoreCase(lookaheadName) != 0)
		    && (TWOWAYPATH_LOOKAHEAD.compareToIgnoreCase(lookaheadName) != 0))
		throw new CommandLineException("unknown lookahead " + nodeAggregatorName);
	}

	protected final MultiGraphCompressor getCompressorFor_(String name) {
		if (NO_COMPRESSOR.compareToIgnoreCase(name) == 0)
			return new NoCompressor();
		if (IDENT_COMPRESSOR.compareToIgnoreCase(name) == 0)
			return new MultiIdentCompressor();
		if (INCOLOR_COMPRESSOR.compareToIgnoreCase(name) == 0)
			return new MultiInColorsCompressor();
		if (INSETCOLOR_COMPRESSOR.compareToIgnoreCase(name) == 0)
			return new MultiInSetColorsCompressor();
		if (INSET_COMPRESSOR.compareToIgnoreCase(name) == 0)
			return new MultiInSetCompressor();
		throw new CommandLineException("unknown compressor " + name);
	}
	
	public void set_compressor(List<String> args) {
		compressor = getCompressorFor_(args.get(0));
	}

}

