package helix.graph.algo.c3part;

import java.util.HashSet;
import java.util.Set;

import helix.graph.algo.c3part.optimizer.ColorOptimizer;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RObject;
import helix.graph.model.impl.RVertex;
import helix.graph.model.impl.RVertexList;
import helix.graph.model.impl.RVertexSet;

/**
 * Utilities for C3Part algorithms
 * 
 * Note: for readability, we use uppercase characters for parameters
 * holding nodes of the multigraph (e.g. V, U) and lowercase characters for
 * nodes of the datagraph (e.g. v, u).
 * A node of the multigraph is called a 'multinode' and a node of the datagraph
 * is called a 'datanode'.
 * 
 * Note: some part of the code (marked as @concurrent) are specific to the
 * concurrent version. you may ignore them for the non-concurrent version
 * 
 */

public class C3PartUtil {

	/**
	 * the tuple associated to a node of the multigraph
	 */
	public static final String Nodes = "C3P_Nodes"; // no starting "_" please

	/**
	 * the local optimizer : experimental section
	 */
	public static final String Optimizer = "C3P_Optimizer";  // no starting "_" please

	/**
	 * the multinode associated to a datagraph node
	 * 
	 * @concurrent: this is actually a prefix to the registry key because we need this
	 * key to be thread specific.
	 */
	protected static final String MultinodePrefix = "_C3P_Multinode_";

	/**
	 * the multinodes sons of a multinode
	 * 
	 * @concurrent: this is actually a prefix to the registry key because we need this
	 * key to be thread specific.
	 * 
	 */
	protected static final String SonsPrefix = "_C3P_Sons_";

	/**
	 * the multinodes aliases of a multinode
	 * and the multinode root of multinodes
	 * 
	 * @concurrent: this is actually a prefix to the registry key because we need this
	 * key to be thread specific.
	 * 
	 */
	protected static final String AliasesPrefix = "_C3P_Alias_";
	protected static final String RootPrefix = "_C3P_Root_";

	/**
	 * @concurrent: some ThreadLocal variable to generate thread specific id's.
	 */
    private static ThreadLocal<String> MultinodeSerialNum = new ThreadLocal<String>() {
        protected synchronized String initialValue() {
            return MultinodePrefix + Thread.currentThread().getName();
        }
    };
	
	private static ThreadLocal<String> SonsSerialNum = new ThreadLocal<String>() {
        protected synchronized String initialValue() {
            return SonsPrefix + Thread.currentThread().getName();
        }
    };

    private static ThreadLocal<String> AliasesSerialNum = new ThreadLocal<String>() {
        protected synchronized String initialValue() {
            return AliasesPrefix + Thread.currentThread().getName();
        }
    };

    private static ThreadLocal<String> RootSerialNum = new ThreadLocal<String>() {
        protected synchronized String initialValue() {
            return RootPrefix + Thread.currentThread().getName();
        }
    };

    /**
     * internal storage to speedup getTrueColors() and getPseudoColors()
     */
	private static RColorSet trueColors_ = new RColorSet();
	private static RColorSet pseudoColors_ = new RColorSet();

	/**
	 * internal storage to speedup getPseudoColors()
	 */
	private static HashSet<Vertex> setV_ = new HashSet<Vertex>();
	private static HashSet<Vertex> setU_ = new HashSet<Vertex>();
	
    /**
     * API
     */
	
	/**
	 *  protected constructor : this class is static
	 */
	protected C3PartUtil() {
	}

	/**
	 * @concurrent : the actual key entry for Multinode is thread specific
	 * (for the non-concurrent version this may be a constant
	 * 
	 */
	private static final String multinodeKey_() {
//		return MultinodePrefix;
		return MultinodeSerialNum.get();
	}
	
	/**
	 * @concurrent : the actual key entry for Sons is thread specific
	 * (for the non-concurrent version this may be a constant
	 * 
	 */
	private static final String sonsKey_() {
//		return SonsPrefix;
		return SonsSerialNum.get();
	}

	/**
	 * @concurrent : the actual key entry for Alias is thread specific
	 * (for the non-concurrent version this may be a constant
	 * 
	 */
	private static final String aliasesKey_() {
//		return AliasesPrefix;
		return AliasesSerialNum.get();
	}

	/**
	 * @concurrent : the actual key entry for Root is thread specific
	 * (for the non-concurrent version this may be a constant
	 * 
	 */
	private static final String rootKey_() {
//		return RootPrefix;
		return RootSerialNum.get();
	}

	/**
	 * utility: collect V and all its aliases in set
	 */
	private static final void collectVertices_(Vertex V, Set<Vertex> set) {
		set.clear();
		set.add(V);
		for (Vertex U : getAliases(V)) {
			set.add(U);
		}
	}
	
	/** 
	 * some handy Registry manipulation functions
	 */
	
	// --------------------------
	// Handling datanodes in multinodes.
	//
	// Nodes(V) = (v1,..., vk)
	// --------------------------
	
	/**
	 * get datanodes tuple of a multinode V (i.e. Nodes(V))
	 * and create an empty RVertexList if it does not exist.
	 * 
	 * shorthand for RVertexList.getVertexList(V, Nodes)
	 */
	public static final RVertexList getNodes(Vertex V) {
		return RVertexList.getVertexList(V, Nodes);
	}
	
	/**
	 * add datanode v in tuples of a multinode V
	 */
	public static final void addNode(Vertex V, Vertex v) {
		getNodes(V).addVertex(v);
	}

	/**
	 * get last datanode of a multinode V 
	 *  Last(V) = vk / Nodes(V) = (v1, ..., vk)
	 */
	public static final Vertex getLastNode(Vertex V) {
		return getNodes(V).lastVertex();
	}
	
	
	// --------------------------
	// Handling multinodes with single datanode.
	//
	// Multinode(v) = V / Nodes(V)=(v)
	// --------------------------

	/**
	 * get the multinode V such that Nodes(V) = (v)
	 * this is used in multigraph initialisation to retrieve
	 * the multinode V that has been build from v.
	 * return null if v didn't give rise to any multinode.
	 */
	public static final Vertex getMultinode(Vertex v) {
		String key = multinodeKey_();
		return (v.has(key) ? ((RVertex) v.get(key)).get() : null);
	}

	/**
	 * record in datanode v the multinode V such that Nodes(V) = (v).
	 * this is used in multigraph initialisation to keep track in v of
	 * the multigraph vertex V that has been build from v.
	 */
	public static final void setMultinode(Vertex v, Vertex V) {
		String key = multinodeKey_();
		if (v.has(key))
			((RVertex) v.get(key)).set(V);
		else
			v.set(key, new RVertex(V));
	}

	/**
	 * remove multinode V from datanode v
	 */
	public static final void removeMultinode(Vertex v) {
		v.remove(multinodeKey_());
	}
	

	// --------------------------
	// Handling Star (*) datanode 
	// and Star containig multinodes
	//
	// --------------------------
	
	/**
	 * tell if datanode v is a star (*)
	 * 
	 * note: there are different ways to detect '*'
	 * the quickest for now is that '*' is (should be)
	 * the only colorless datanode
	 */
	public static final boolean isStar(Vertex v) {
		return (! RColorSet.hasColors(v));
	}
	
	/**
	 * count number of stars in multinode V
	 */
	public static final int countStars(Vertex V) {
		int count = 0;
		for (Vertex v : C3PartUtil.getNodes(V)) {
			if (isStar(v))
				count++;
		}
		return count;
	}

	/**
	 * tell if multinode V is a full star
	 * (also called the 'black' star)
	 * i.e if all of its datanodes are stars
	 */
	public static final boolean isFullStar(Vertex V) {
		for (Vertex v : getNodes(V)) {
			if (! isStar(v))
				return false;
		}
		return true;
	}

	/**
	 * tell if multinode V is a partial star on
	 * at least one color but not a full star
	 */
	public static final boolean isPartialStar(Vertex V) {
		for (Vertex v : getNodes(V)) {
			if (isStar(v))
				return (! isFullStar(V));
		}
		return false;
	}
	
	/**
	 * tell if multinode V has no star.
	 * such a multinode is called a 'CoreNode'
	 */
	public static final boolean isCoreNode(Vertex V) {
		for (Vertex v : getNodes(V)) {
			if (isStar(v))
				return false;
		}
		return true;
	}

	// --------------------------
	// Father - Sons relationship.
	//
	// this relationship is used during OTF multinode 
	// expansion to keep track of which multinode of
	// the initial multigraph did give rise to which
	// multinodes of the new multigraph.
	//
	// Sons(V) = {V1, V2, ...}
	// --------------------------

	/**
	 * get multinodes sons of a multinode V
	 * and create an empty RVertexSet if it does not exist.
	 * 
	 * shorthand for RVertexSet.getVertexSet(V, Sons)
	 */
	public static final RVertexSet getSons(Vertex V) {
		return RVertexSet.getVertexSet(V, sonsKey_());
	}
	
	/**
	 * get the multinode son of a multinode V, that
	 * ends with specified terminals.
	 * return null if not found
	 */
	public static final Vertex getSon(Vertex V, Vertex terminal) {
		if (! hasSon(V))
			return null;
		for (Vertex Son : getSons(V)) {
			if (getLastNode(Son).equals(terminal))
				return Son;
		}
		return null;
	}

	/**
	 * add son into multinode V
	 */
	public static final void addSon(Vertex V, Vertex Son) {
		getSons(V).addVertex(Son);
	}

	/**
	 * tell if multinode V has son(s)
	 */
	public static final boolean hasSon(Vertex V) {
		return V.has(sonsKey_());
	}
	
	/**
	 * remove all sons from multinode V
	 */
	public static final void removeSons(Vertex V) {
		V.remove(sonsKey_());
	}
	public static final void removeSons(Iterable<Vertex> iterable) {
		for (Vertex V : iterable)
			V.remove(sonsKey_());
	}

	// --------------------------
	// Root - Aliases relationship
	// 
	// This relationship is used by Compressors
	// to keep track of which multinode (root) subsumes
	// which multinodes (aliases)
	//
	// Aliases(V) = {V1, V2, ...}
	// Root(Vi) = V
	// --------------------------

	/**
	 * get multinodes aliases of a multinode V
	 * (V is called a root multinode)
	 * and create an empty RVertexSet if it does not exist.
	 * 
	 */
	public static final RVertexSet getAliases(Vertex V) {
		return RVertexSet.getVertexSet(V, aliasesKey_());
	}

	/**
	 * add alias into multinode V
	 * (also set root of V to Root)
	 */
	public static final void addAlias(Vertex V, Vertex Alias) {
		getAliases(V).addVertex(Alias);
		setRoot(Alias, V);
	}

	/**
	 * tell if multinode V has aliases
	 * i.e. if V is a root multinode
	 */
	public static final boolean hasAlias(Vertex V) {
		return (V.has(aliasesKey_()) && (getAliases(V).size() != 0));
	}

	/**
	 * remove alias from multinode
	 */
	public static final void removeAlias(Vertex V, Vertex Alias) {
		RVertexSet aliases = getAliases(V);
		aliases.removeVertex(Alias);
	}

	/**
	 * remove all aliases from multinode V
	 * (also remove root from Aliases(V))
	 */
	public static final void removeAliases(Vertex V) {
		RVertexSet aliases = getAliases(V);
		V.remove(aliasesKey_());
		for (Vertex U : aliases) {
			removeRoot(U);
		}
	}
	public static final void removeAliases(Iterable<Vertex> iterable) {
		for (Vertex V : iterable) {
			removeAliases(V);
		}
	}
	
	/**
	 * get root of multinode V
	 * return null if V is not an Alias Node
	 */
	public static final Vertex getRoot(Vertex V) {
		String key = rootKey_();
		return (V.has(key) ? ((RVertex) V.get(key)).get() : null);
	}
	
	/**
	 * record Root multinode in multinode V
	 * (also add V in Aliases(Root))
	 */
	public static final void setRoot(Vertex V, Vertex Root) {
		String key = rootKey_();
		if (V.has(key)) {
			RVertex container = (RVertex) V.get(key);
			if (container.get() != Root) {
				removeAlias(container.get(), V);
				container.set(Root);
			}
		}
		else
			V.set(key, new RVertex(Root));
		
		if (! getAliases(Root).hasVertex(V))
			addAlias(Root, V);
	}

	/**
	 * tell if multinode V has a root node
	 * i.e. if it is an alias node
	 */
	public static final boolean hasRoot(Vertex V) {
		return V.has(rootKey_());
	}

	/**
	 * remove root from multinode V
	 * (and remove V from its root aliases)
	 */
	public static final void removeRoot(Vertex V) {
		Vertex Root = getRoot(V);
		if (Root != null) {
			V.remove(rootKey_());
			getAliases(Root).removeVertex(V);
		}
	}
	public static final void removeRoot(Iterable<Vertex> iterable) {
		for (Vertex V : iterable) {
			removeRoot(V);
		}
	}

	// --------------------------
	// Handling ColorOptimizer(V)
	//
	// --------------------------

	/**
	 * get the color optimizer attached to the registry
	 */
	public static final ColorOptimizer getOptimizer(Registry r) {
		RObject robj = (RObject) r.get(Optimizer);
		return (robj == null ? null : (ColorOptimizer) robj.get());
	}

	/**
	 * set the color optimizer into the registry
	 */
	public static final void setOptimizer(Registry r, ColorOptimizer optimizer) {
		r.set(Optimizer, new RObject(optimizer));
	}
	
	// --------------------------
	// Handling true and pseudo colors between multinodes
	//
	// this is used by Compressors
	// --------------------------
	
	/**
	 * compute true colors of edge (U, V) using
	 * the datagraph (i.e not relying on an actual edge).
	 * may return an empty colorset.
	 */
	public static final RColorSet getTrueColors(Vertex V, Vertex U) {
		trueColors_.clearColors();
		RVertexList Vnodes = C3PartUtil.getNodes(V);
		RVertexList Unodes = C3PartUtil.getNodes(U);
		for (int i = 0 ; i < Vnodes.size() ; i++) {
			Vertex v = Vnodes.getVertex(i);
			if (! RColorSet.hasColors(v)) // @NonStrict
				continue;
			Vertex u = Unodes.getVertex(i);
			if (! RColorSet.hasColors(u)) // @NonStrict
				continue;
			if ((v == u) || v.hasEdgeTo(u)) {
				trueColors_.setColor(RColorSet.getColors(v).singleColor());
			}
		}
		return trueColors_;
	}	

	/**
	 * compute pseudo colors of edge (U, V) using
	 * the datagraph (i.e not relying on an actual edge).
	 * pseudo colors are the projection (OR) of all
	 * true colors of all edges between root and alias nodes.
	 * Example:
	 * V = Root(V1, V2) ; U = Root(U1, U2)
	 * pseudoColors = OR( trueColors(X,Y))
	 * for X = V, V1, V2 ; Y = U, U1, U2
	 * 
	 * note: if V and U are not roots, then this will
	 * return the same as trueColors.
	 * 
	 * may return an empty colorset.
	 */
	public static RColorSet getPseudoColors(Vertex V, Vertex U) {
		collectVertices_(V, setV_);
		collectVertices_(U, setU_);
		pseudoColors_.clearColors();
		for (Vertex VV : setV_) {
			for (Vertex UU : setU_) {
				RColorSet colors = C3PartUtil.getTrueColors(UU, VV);
				pseudoColors_.orColors(colors);
			}
		}
		return pseudoColors_;
	}
	
}
