package helix.graph.algo.c3part;

import static org.junit.Assert.*;
import helix.graph.io.DimacsExtendedReader;
import helix.graph.io.GraphReader;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.memory.MemGraphFactory;
import helix.graph.util.IPermutation;


import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Set;

import org.junit.Test;
public class TestC3PartDFS {
	
	@Test
	public void testC3PartDFSMulti() throws IOException {
		
		// fred's paper example
		//
		//   2.....4
		//  .| \   |
		//   1..3--5
		//
		String sin = "p test\nn label=v1\nn label=v2\nn label=v3\nn label=v4\nn label=v5\n" +
				     "e 1 2 colors=11\ne 1 3 colors=1\ne 2 3 colors=10\ne 2 4 colors=1\ne 3 5 colors=10\n e 4 5 colors=10\n";

		StringReader sreader = new StringReader(sin);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(sreader, factory);
		Graph graph = reader.read();
		
		assertEquals("graph size", 5, graph.nbVertices());
		assertEquals("nb colors", 2, RColorSet.nbColors(graph));

		ArrayList<Integer> colors = new ArrayList<Integer>(1);
		
		C3PartDFS algo = new C3PartDFS(graph);
		C3PartOptions options = algo.getOptions();
		
		Set<GSet<Vertex>> part;
		
		// run C3Part BFS with color 0 only
		colors.add(0);
		options.colors = new IPermutation(colors);
		part = algo.runMulti_(graph, options);
		assertEquals("partition size", 2, part.size());
		
		// run C3Part BFS with color 1 only
		colors.clear();
		colors.add(1);
		options.colors = new IPermutation(colors);
		part = algo.runMulti_(graph, options);
		assertEquals("partition size", 1, part.size());
		
		// run C3Part BFS with all colors
		colors.clear();
		colors.add(0);
		colors.add(1);
		options.colors = new IPermutation(colors);
		part = algo.runMulti_(graph, options);
		assertEquals("partition size", 3, part.size());
		
		// test result partition
		
		Vertex v1 = graph.findVertex(Registry.Label, "v1");
		Vertex v2 = graph.findVertex(Registry.Label, "v2");
		Vertex v3 = graph.findVertex(Registry.Label, "v3");
		Vertex v4 = graph.findVertex(Registry.Label, "v4");
		Vertex v5 = graph.findVertex(Registry.Label, "v5");
		
		GSet<Vertex> c1 = graph.vertexSetFactory().make();
		GSet<Vertex> c2 = graph.vertexSetFactory().make();
		GSet<Vertex> c3 = graph.vertexSetFactory().make();

		c1.add(v1);
		c1.add(v2);
		c1.add(v3);
		c2.add(v4);
		c3.add(v5);

		Set<GSet<Vertex>> expect = graph.vertexSetFactory().makeSuper();
		expect.add(c1);
		expect.add(c2);
		expect.add(c3);
		
		assertTrue("partition", expect.equals(part));
	}

	@Test
	public void testC3PartDFSLargeMulti() throws IOException {
		
		FileReader freader = new FileReader("samples/sample.multi");
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(freader, factory);
		Graph graph = reader.read();

		C3PartDFS algo = new C3PartDFS(graph);
		C3PartOptions options = algo.getOptions();
		options.colors = new IPermutation(RColorSet.nbColors(graph));
		
		Set<GSet<Vertex>> part = algo.runMulti_(graph, options);
		
		assertEquals("partition size", 32, part.size());
		
		int max = 0;
		for (GSet<Vertex> clas : part) {
			if (clas.size() > max)
				max = clas.size();
		}

		assertEquals("max clas size", 116, max);
	}
	
	
	@Test
	public void testC3PartDFSLarge() throws IOException {
		
		FileReader freader = new FileReader("samples/sample.poly");
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(freader, factory);
		Graph graph = reader.read();

		C3PartDFS algo = new C3PartDFS(graph);

		Set<Graph> part = algo.run();
		
		assertEquals("partition size", 32, part.size());
		
		int max = 0;
		for (Graph clas : part) {
			if (clas.nbVertices() > max)
				max = clas.nbVertices();
		}

		assertEquals("max clas size", 116, max);
	}
	
	
}

