package helix.graph.algo.c3part.aggregator;

/**
 * a connected component aggregator<br>
 * An acceptable multinode V=(v1,..., vn) is such that:<br>
 * <code>
 * the restriction of the datagraph to {vi} is a connected component
 * </code>
 * <br>
 * Note: this implementation is star compatible and therefore contains some
 * extra code to deal with star vertices (*).
 * The definition is the same as a non-star aggregator with the convention: 
 * <code>
 * A v * epsilon v
 * </code> 
 * This extra code is indicated below with the @NonStrict keyword.
 * 
 */
import java.util.ArrayList;
import java.util.List;

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.preprocessor.CCColorCleaner;
import helix.graph.algo.c3part.preprocessor.GraphCleaner;
import helix.graph.algo.c3part.visitor.CCByEpsilonVisitor;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class CCAggregator extends NodeAggregator {

	protected static final String CCIndexFlag = "_CCAggregatorIndex_";
	
	/**
	 * an internal list of epsilon CC (for speedup)
	 */
	protected ArrayList<GSet<Vertex>> components_ = new ArrayList<GSet<Vertex>>();

	/**
	 * internal sets of vertices (for speedup)
	 */
	protected GSet<Vertex> candidates_ = datagraph_.vertexSetFactory().make();
	protected GSet<Vertex> closeNeighbors_ = datagraph_.vertexSetFactory().make();
	protected GSet<Vertex> farNeighbors_ = datagraph_.vertexSetFactory().make();

	protected ArrayList<Vertex> allNeighbors_ = new ArrayList<Vertex>();
	
	
	/**
	 * an internal set of vertices representing a tuple (for speedup)
	 */
	protected GSet<Vertex> tuple_ = datagraph_.vertexSetFactory().make();
	protected GSet<Vertex> visited_ = datagraph_.vertexSetFactory().make();
	
	/**
	 * internal colorSets (for speedup)
	 */
	protected RColorSet currentColors_ = new RColorSet();
	protected RColorSet tupleColors_ = new RColorSet();
	
	/**
	 * current multigraph colors
	 */
	protected RColorSet multicolors_ = null;
	
	/**
	 * init epsilon connected components (for speedup)
	 */
	protected final void initCC_() {
		
		CCVisitor visitor = new CCByEpsilonVisitor(datagraph_);
		Iterable<GSet<Vertex>> ccs = visitor.traverse().getComponents();
		
		components_.clear();

		int index = 0;
		for (GSet<Vertex> cc : ccs) {
			components_.add(cc);
			// keep CC index in vertex
			for (Vertex v : cc) {
				v.set(CCIndexFlag, index);
			}
			index++;
		}
	}
	
	/**
	 * cleanup datagraph vertices
	 */
	protected final void cleanup_() {
		for (Vertex v : datagraph_.vertices()) {
			v.remove(CCIndexFlag);
		}
	}
	
	/**
	 * get CC containing the corresponding vertex
	 */
	protected final GSet<Vertex> getCCOf_(Vertex v) {
		int index = v.get(CCIndexFlag).toInt();
		return components_.get(index);
	}

	/**
	 * check if there is a CC containing all vertices in 'tupleVertices'
	 * => and having all the colors required in tupleColors.
	 * 
	 * note: for an aggregator, this is a necessary but insufficient condition
	 * for any k != n.
	 * The condition is sufficient at the end (i.e. when k = n)
	 */
	protected boolean checkCC_(Vertex v, GSet<Vertex> tupleVertices,
							   GSet<Vertex> visited,
							   RColorSet currentColors, RColorSet tupleColors) {
		// tupleVertices : V+v
		// visited : vertices visited in the current search
		// tupleColors : colors of V+v (including colors on which there's a star)
		// currentColors : current visited colors (starts with the colors on which V has a star)

		// note that v is never a Star 
		if (visited.contains(v)) // v is already visited
			return false;
		
		int vcolor = RColorSet.getColors(v).singleColor();
		
		
		// add visited color
		currentColors.setColor(vcolor);
		visited.add(v);

		// check if complete
		// ie the existence of a connected component including all k+1 vertices
		// that means it's not necessarily a winning extension but it'll still be exact 
		// (since at the last step we check the existence of a cc including all n vertices)
		// note that we could say allColors instead of tupleColors, but it would have
		// to be modulo the number of remaining possible stars, ie it would be a condition on size of currentcolors
		if (currentColors.includeColors(tupleColors)){
			return true; // end of recursion
		}
		
		// note that for a color present in tuple, we have to go through
		// the right vertex (i.e. the one in tuple) not another one
		// of the same color, we make  sure of that in collectNeighbors_
		List <Vertex> neighbors = collectNeighbors_(v, vcolor, tupleVertices, visited, tupleColors); 

		for (Vertex u : neighbors) {
			if (checkCC_(u, tupleVertices, visited, currentColors, tupleColors))
				return true;
		}
		
		return false;
	}
	
	// collect epsilon-neighbors that are not yet visited
		
	protected final List<Vertex> collectNeighbors_(Vertex v, int vcolor, GSet<Vertex> tupleVertices, GSet<Vertex> visited, RColorSet tupleColors) {
		ArrayList<Vertex> neighbors = new ArrayList<Vertex>();
		for (Vertex u : v.neighbors()) {
			int ucolor = RColorSet.getColors(u).singleColor();
			if (vcolor == ucolor) // not an epsilon link
				continue;
			if (visited.contains(u)) // already visited
				continue;
			if (tupleColors.hasColor(ucolor)) {
				// for a color present in tuple, we have to go through
				// the right vertex (i.e. the one in tuple) not another one
				// of the same color.
				if (! tupleVertices.contains(u))
					continue;
				neighbors.add(u);
			}
			else {
				neighbors.add(u);
			}
		}
		return neighbors;
	}
	
	
	/**
	 * collect espilon neighbors of v in an ordered way to speed
	 * up checkCC_
	 */
	protected final List<Vertex> collectNeighborsSave_(Vertex v, int vcolor, 
											 GSet<Vertex> tupleVertices, 
											 GSet<Vertex> visited,
											 RColorSet tupleColors) {
		closeNeighbors_.clear();
		farNeighbors_.clear();
		
		for (Vertex u : v.neighbors()) {
			//System.err.println("hihi");
			int ucolor = RColorSet.getColors(u).singleColor();
			//System.err.println("hoho");
			if (vcolor == ucolor) // not an epsilon link
				continue;
			if (visited.contains(u)) // already visited
				continue;
			if (tupleColors.hasColor(ucolor)) {
				// for a color present in tuple, we have to go through
				// the right vertex (i.e. the one in tuple) not another one
				// of the same color.
				if (! tupleVertices.contains(u))
					continue;
				closeNeighbors_.add(u);
			}
			else {
				farNeighbors_.add(u);
			}
		}

		// put close neighbors first to prioritize
		allNeighbors_.clear();
		allNeighbors_.addAll(closeNeighbors_);
		allNeighbors_.addAll(farNeighbors_);
		
		return allNeighbors_;
	}
	
	/**
	 * get first non star element in Nodes(V)
	 */
	private static final Vertex getPivot_(Vertex V) {
		for (Vertex v : C3PartUtil.getNodes(V)) {
			if (! C3PartUtil.isStar(v))
				return v;
		}
		return null;
	}
	
	
	/*
	 * API
	 */
	
	/**
	 * constructor
	 */
	public CCAggregator(Graph datagraph, C3PartOptions options, RColorSet multicolors) {
		super(datagraph, options);
		multicolors_ = multicolors;
		initCC_();
	}

	/**
	 * @Implement
	 */
	public boolean isAggregable(Vertex V, Vertex v) {
		
		if (C3PartUtil.isStar(v) && (C3PartUtil.isFullStar(V))) 
			return true;
		
		tuple_.clear();
		tuple_.addIterable(C3PartUtil.getNodes(V));
		tuple_.add(v);
		
		visited_.clear();
		
		// tuple_ : V+v
		// tupleColors_ : colors of V+v (including colors on which there's a star)
		// currentColors_ : current visited colors (starts with the colors on which V has a star)
		// multicolors_ : colors of the current MGA (ie colors of V including colors on which there's a star)

		currentColors_.clearColors();
		tupleColors_.clearColors();

		tupleColors_.setColors(RColorSet.getColors(V));
		if (! C3PartUtil.isStar(v)) 
			tupleColors_.setColor(RColorSet.getColors(v).singleColor());

		if (multicolors_ != null) {
			currentColors_.setColors(multicolors_);
			if (! C3PartUtil.isStar(v)) 
				currentColors_.xorColors(RColorSet.getColors(V)); // xor to get only the colors on which V has stars
			else 
				currentColors_.xorColors(tupleColors_);
			
			tupleColors_.orColors(multicolors_);
			
		}
		boolean ok;
		if (C3PartUtil.isStar(v)) 
			ok = checkCC_(getPivot_(V), tuple_, visited_, currentColors_, tupleColors_);
		else 
			ok = checkCC_(v, tuple_, visited_, currentColors_, tupleColors_);
		
		return ok;
	}
	

	/**
	 * @Implement
	 * get path candidates
	 * <code>
	 * candidates = {v in datagraph / (color(v) == color) and (v in CC(pivot))} <br>
	 * where<br>
	 * pivot = any non star element of Nodes(V)<br>
	 * CC(Pivot) = epsilon CC containing pivot
	 * </code> 
	 * note: in this implementation pivot = last element in Nodes(V)
	 */
	protected Iterable<Vertex> getCandidates(Vertex V, int color) {

		Vertex pivot = getPivot_(V);
		
		candidates_.clear();
		
		if (pivot == null)		// @NonStrict
			return candidates_; // no candidates for the black star

		GSet<Vertex> cc = getCCOf_(pivot);
		for (Vertex v : cc) {
			if (RColorSet.getColors(v).singleColor() == color)
				candidates_.add(v);
		}
		
		return candidates_;
	}
	
	@Override
	/**
	 * return a cleaner for CC
	 */
	public GraphCleaner getCleanerFor() {
		int quorum = RColorSet.nbColors(datagraph_) - options_.maxstar - 1;
		return new CCColorCleaner(quorum);
	}

}
