package helix.graph.algo.c3part.aggregator;

/**
 * a centered star aggregator<br>
 * An acceptable multinode V=(v1,..., vn) is such that:<br>
 * <code>
 * A vi(i>1) in Nodes(V) (vi epsilon v1)
 * </code>
 * terminals of V=(v1,..., vn) are defined as :<br>
 * <code>
 * terminals = {v in datagraph / (color(v) == color) and (v epsilon v1)}
 * </code> <br>
 * i.e the first datanode in Nodes(V) is considered as the center.
 * a CenteredAggregator is defined
 * as a PivotingAggregator with pivot = v1
 * <br>
 * Note: this implementation is star compatible and therefore contains some
 * extra code to deal with star vertices (*).
 * The definition is the same as a non-star aggregator with the convention: 
 * <code>
 * A v * epsilon v
 * </code> 
 * This extra code is indicated below with the @NonStrict keyword.
 */

import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.preprocessor.CenteredQuorumCleaner;
import helix.graph.algo.c3part.preprocessor.GraphCleaner;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;

public class CenteredAggregator extends PivotingAggregator {

	/**
	 * required constructor
	 */
	public CenteredAggregator(Graph datagraph, C3PartOptions options) {
		super(datagraph, options);
	}

	/**
	 * Check if datagraph node 'v' forms an epsilon star with 
	 * the first element in Nodes(V), considered as the pivot
	 */
	public boolean isAggregable(Vertex V, Vertex v) {

		RVertexList nodes = C3PartUtil.getNodes(V);
		Vertex u = nodes.getVertex(0);
		
		if (nodes.size()==1) { // we are on the center color : we have to check if the node has neighbors on enough different colors
			if (C3PartUtil.isStar(u))  // @NonStrict - pivot cannot be a star
				return false;
			RColorSet colors = new RColorSet();
			colors.orColors(RColorSet.getColors(u));
			for (Vertex w: u.neighbors()) {
				colors.orColors(RColorSet.getColors(w));
			}
			return (colors.countColors() - 1 >= RColorSet.nbColors(datagraph_) - options_.maxstar - 2); //-1 for u's color and -1 for epsilon color
		}
		else {
			if (C3PartUtil.isStar(v))  // @NonStrict
				return (nodes.size() + 1 - RColorSet.nbColors(V) <=  options_.maxstar);
		
			return v.hasEdgeTo(u);
		}
	}
	
	
	/**
	 * @Implement
	 * for centered aggregator, the pivot is the first element
	 * in Nodes(V). It should not be a star.
	 */
	protected final Vertex getPivot(Vertex V) {
		return  C3PartUtil.getNodes(V).getVertex(0);
	}
	
	@Override
	/**
	 * we do have a preferred color : the one at index 0
	 */
	public int getPreferredColor() {
		return options_.colors.get(0);
	}

	@Override
	/**
	 * return a cleaner for Centered
	 */
	public GraphCleaner getCleanerFor() {
		int quorum = RColorSet.nbColors(datagraph_) - options_.maxstar - 1;
		return new CenteredQuorumCleaner(quorum, getPreferredColor());
	}

}
