package helix.graph.algo.c3part.aggregator;

/**
 * a clique aggregator<br>
 * An acceptable multinode V=(v1,..., vn) is such that:<br>
 * <code>
 * A vi in Nodes(V) (v epsilon vi)
 * </code>
 * 
 * terminals of V=(v1,..., vk) are therefore defined as :<br>
 * <code>
 * terminals = {v in datagraph / (color(v) == color) and A vi in Nodes(V) (v epsilon vi)}
 * </code> 
 * for a more efficient implementation a CliqueAggregator is defined
 * as a PivotingAggregator with pivot = any vi in Nodes(V)
 * i.e: <br>
 * <code>
 * candidates = {v in datagraph / (color(v) == color) and (v epsilon pivot)}<br>
 * with pivot = any vi in Nodes(V)
 * </code> <br>
 * Note: this implementation assumes that Nodes(V) already forms an epsilon clique.
 * <br>
 * Note: this implementation is star compatible and therefore contains some
 * extra code to deal with star vertices (*).
 * The definition is the same as a non-star aggregator with the convention: 
 * <code>
 * A v * epsilon v
 * </code> 
 * This extra code is indicated below with the @NonStrict keyword.
 */

import java.util.ArrayList;

import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.preprocessor.GraphCleaner;
import helix.graph.algo.c3part.preprocessor.QuorumCleaner;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class CliqueAggregator extends PivotingAggregator {

	protected GSet<Vertex> closeNeighbors_ = datagraph_.vertexSetFactory().make();
	protected GSet<Vertex> farNeighbors_ = datagraph_.vertexSetFactory().make();

	/**
	 * an internal set of vertices representing a tuple (for speedup)
	 */
	protected GSet<Vertex> tuple_ = datagraph_.vertexSetFactory().make();
	protected GSet<Vertex> visited_ = datagraph_.vertexSetFactory().make();
	
	/**
	 * internal colorSets (for speedup)
	 */
	protected RColorSet currentColors_ = new RColorSet();
	protected RColorSet tupleColors_ = new RColorSet();
	
	/**
	 * current multigraph colors
	 */
	protected RColorSet multicolors_ = new RColorSet();
	
	protected ArrayList<Vertex> allNeighbors_ = new ArrayList<Vertex>();
	/**
	 * required constructor
	 */
	public CliqueAggregator(Graph datagraph, C3PartOptions options) {
		super(datagraph, options);
	}

	/**
	 * Check if datagraph node 'v' forms an epsilon clique with 
	 * all elements in Nodes(V)
	 * 
	 * note: star ghost is considered having epsilon links with any vertex
	 *       (including itself)
	 * 
	 * note: since Nodes(V) already forms a clique, we just have to check
	 * that v is epsilon connected to each element
	 */
	public boolean isAggregable(Vertex V, Vertex v) {
		if (C3PartUtil.isStar(v))  // @NonStrict
			return true;

		for (Vertex u : C3PartUtil.getNodes(V)) {
			if (C3PartUtil.isStar(u)) // @NonStrict
				continue;
			if (! (v.hasEdgeTo(u)))
				return false;
		}
		return true;
	}
	
	/**
	 * @Implement
	 * for clique aggregator, the pivot is any non-star element in Nodes(V).
	 */
	
	protected Vertex getPivot(Vertex V) {
		for (Vertex v : C3PartUtil.getNodes(V)) {
			if (! C3PartUtil.isStar(v))
				return v;
		}
		// huh... we got a super star ! just return a star 
		// this will not hurt since it's not connected to anything
		return C3PartUtil.getLastNode(V); 
	}

	@Override
	/**
	 * return a cleaner for Clique
	 */
	public GraphCleaner getCleanerFor() {
		int quorum = RColorSet.nbColors(datagraph_) - options_.maxstar - 1;
		return new QuorumCleaner(quorum);
	}

}
