package helix.graph.algo.c3part.aggregator;

/**
 * a dense connected component aggregator<br>
 * An acceptable multinode V=(v1,..., vn) is such that:<br>
 * <code>
 * the restriction of the datagraph to {vi} is a connected component
 * and each of the nodes in vi has at least n * 50% epsilon-links to other nodes in vi
 * </code>
 * 
 */

import java.text.DecimalFormat;

import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class DenseCCAggregator extends CCAggregator {

	/**
	 * required density :
	 * 
	 * iDensity_ : density % (in relative mode)
	 *             nb of links (in absolute mode)
	 *             
	 * dDensity_ : link density (in relative mode)
	 *             no-meaning in absolute mode
	 *             
	 * nbTotal_  : total number of colors in datagraph - 1
	 * 
	 * minNonstar_ : min number of non stars required
	 * 
	 * isAbsolute_ : absolute/relative mode
	 */
	protected double dDensity_;
	protected int iDensity_;
	protected int nbTotal_;
	protected int minNonstar_;
	protected boolean isAbsolute_;
	
	/**
	 * internal: output format for density
	 */
	protected static DecimalFormat format_ = new DecimalFormat("0.00");
	
	
	
	/**
	 * internal utility : compute number of required link per node
	 *                    currently in tuple.
	 * the important property here is that nbRequiredLinks_
	 * is monotonically increasing.
	 */
	protected final int nbRequiredLinks_(Vertex V, Vertex v) {
		int tupleSize = C3PartUtil.getNodes(V).size() +1;
		int nbRemains = nbTotal_ - tupleSize; // nb remaining colors after V and v 
		int nbRequired;
		
		if (isAbsolute_) {
			nbRequired = iDensity_ - nbRemains; // iDensity_ is an absolute min degree
		}
		else { // we want final degree >= density * (final nbNonStars -1)
			   // we know final degree <= current degree + nbRemains
			   // and final nbNonStars >= current nbNonStars
			   // that implies current degree + nbRemains >= density * (current nbNonStars -1)
			   // ie current degree >= density * (current nbNonStars -1) - nbRemains
 			int nbNonStars = RColorSet.nbColors(V);
			if (!C3PartUtil.isStar(v)) nbNonStars++;
			nbRequired = (int) Math.floor(dDensity_ * (nbNonStars-1)) - nbRemains; // in a dDensity-quasi-clique, the degree must be > dDensity * (nbNonStars-1)
		}

		return nbRequired;
	}
	
	/**
	 * internal utility: check local density condition
	 */
	protected boolean isLocallyDense_(Vertex V, Vertex v) {

		if (C3PartUtil.isFullStar(V)) return true;
		
		// compute number of internal links required
		// the important property here is that nbRequiredLinks
		// is monotonically increasing.

		int nbRequiredLinks = nbRequiredLinks_(V,v);
		
		if (nbRequiredLinks <= 0)  // too early to decipher
			return true;

		// go ahead : check links
		
		tuple_.clear();
		tuple_.addIterable(C3PartUtil.getNodes(V));
		tuple_.add(v);

		for (Vertex u1 : tuple_) {
			if (C3PartUtil.isStar(u1))
				continue;
			int nbEdges = 0;
			for (Vertex u2 : tuple_) {
				if (C3PartUtil.isStar(u2) || (u1 == u2))
					continue;
				if (u1.hasEdgeTo(u2))
					nbEdges++;
			}
			if (nbEdges < nbRequiredLinks)
				return false;
		}
		return true;
	}
	
	/*
	 * API
	 */

	/**
	 * constructor
	 */
	public DenseCCAggregator(Graph datagraph, C3PartOptions options, RColorSet multicolors) {
		super(datagraph, options, multicolors);
		nbTotal_ = RColorSet.nbColors(datagraph_) - 1; // do not count color 0 (epsilon)
		isAbsolute_ = (options.ccDenseMode == C3PartOptions.DENSE_MODE_ABSOLUTE);
		iDensity_ = options.ccDensity;
		minNonstar_ = nbTotal_ - options.maxstar;
		dDensity_ = ((double) options.ccDensity) / 100d;
	}
	
	/**
	 * override isAggregable to add the density condition
	 */
	@Override
	public boolean isAggregable(Vertex V, Vertex v) {		
		return ((isLocallyDense_(V, v)) && (super.isAggregable(V, v)));
	}
	
	/**
	 * override getParamInfo to give information about density
	 */
	public String getParamInfo() {
		return "density "
			   + (isAbsolute_ ? C3PartOptions.DENSE_MODE_ABSOLUTE : C3PartOptions.DENSE_MODE_RELATIVE)
		       + iDensity_;
	}
}