package helix.graph.algo.c3part.aggregator;

import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.preprocessor.GraphCleaner;
import helix.graph.algo.c3part.preprocessor.NoCleaner;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

/**
 * A class to define the way datagraph nodes are aggregated in
 * a multinode.
 * The main method is <code>getTerminals(V, color)</code> that returns the
 * set of datagraph terminals nodes {vi} of given color that can be aggregated
 * to a given multigraph vertex V. This is achieved by calling the <code>isAggregable</code>
 * method  (that tells if a datagraph node 'v' can be aggregated to a multigraph
 * vertex 'V') on a proper set of candidates nodes.
 * <br>
 * To subclass, you have to implement two methods:<br>
 * The <code>isAggregable(V, v)</code> public method tells if datanode v can be
 * aggregated to multinode V<br>
 * The <code>getCandidates(V)</code> protected method returns an Iterable of
 * datanodes that are candidates to be terminals<br>
 * The definition of terminals is therefore:<br>
 * <code>
 * terminals = {v in getCandidates(V) / isAggregable(V, v)}
 * </code> 
 * <br>
 * Typical subclasses are cliqueAggregator, centeredAggregator
 *
 */
public abstract class NodeAggregator {

	protected Graph datagraph_;
	protected C3PartOptions options_;

	/**
	 * required constructor
	 */
	public NodeAggregator(Graph datagraph, C3PartOptions options) {
		datagraph_ = datagraph;
		options_ = options;
	}
	
	/**
	 * Check if datagraph node 'v' can be aggregated to multigraph
	 * vertex V. 
	 */
	public abstract boolean isAggregable(Vertex V, Vertex v);

	/**
	 * returns a set of datagraph nodes that are candidates
	 * to be terminals.
	 */
	protected abstract Iterable<Vertex> getCandidates(Vertex V, int color);
	
	/**
	 * returns the set of datagraph terminals nodes {v} of given color
	 * that can be aggregated to a given multigraph vertex V = (v1,..., vk).
	 *
	 * this is the default method : it calls the <code>isAggregable</code>
	 * on a set of candidates nodes returned by the <code>getCandidates_</code> 
	 * internal method :
	 * v is terminal <=> v is candidate and isAggregable(V, v)
	 */
	public GSet<Vertex> getTerminals(Vertex V, int color) {
		
		GSet<Vertex> terminals = datagraph_.vertexSetFactory().make();

		for (Vertex v : this.getCandidates(V, color)) {
			// check aggregation to tuple elements
			if (this.isAggregable(V, v))
				terminals.add(v);
		}

		return terminals;
	}
	
	/**
	 * accessor to current options
	 */
	public C3PartOptions getOptions() {
		return options_;
	}
	
	/**
	 * tells if this aggregator has some preferred color (e.g. pivot)
	 * (this is mostly useful for ColorOptimizer's to known which
	 * color to start with)
	 * default : return < 0 for no preferred color
	 */
	public int getPreferredColor() {
		return -1; // means no preferred color
	}
	
	/**
	 * get an instance of a datagraph cleaner for this aggregator
	 * 
	 * default: return a NoCleaner 
	 */
	public GraphCleaner getCleanerFor() {
		return new NoCleaner();
	}

	/**
	 * get optional parameters info
	 * this is used only for logging purpose (for user information)
	 * so default is to return an empty string
	 */
	public String getParamInfo() {
		return "";
	}

}
