package helix.graph.algo.c3part.aggregator;

/**
 * a pivoting aggregator is a nodeAggregator for which candidates
 * can be computed from a single 'pivot' in Nodes(V).
 * The definition of candidates is therefore:<br>
 * <code>
 * candidates = {v in datagraph / (color(v) == color) and (v epsilon pivot)}
 * </code> 
 * 
 * This pivot is provided by the <code>getPivot</code> method
 */

import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public abstract class PivotingAggregator extends NodeAggregator {
	
	/**
	 * an internal set of candidates (for speedup)
	 */ 
	protected GSet<Vertex> candidates_ = datagraph_.vertexSetFactory().make();

	/**
	 * constructor - same as super
	 */
	public PivotingAggregator(Graph datagraph, C3PartOptions options) {
		super(datagraph, options);
	}

	/**
	 * return the pivot
	 */
	protected abstract Vertex getPivot(Vertex V);
	
	/**
	 * @Implement
	 * for a pivot aggregator, candidates are the vertices
	 * of given color that are epsilon connected to the pivot.
	 */
	protected Iterable<Vertex> getCandidates(Vertex V, int color) {
		
		Vertex v = getPivot(V);
		
		candidates_.clear();
		
		for (Vertex u : v.neighbors()) {  // foreach u connected to v
			// check proper color
			if (RColorSet.hasColor(u, color))
				candidates_.add(u);
		}
		
		return candidates_;
	}
}
