package helix.graph.algo.c3part.compressor;

import java.util.ArrayList;
import java.util.logging.Logger;

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RVertexSet;

/**
 * A basic skeleton for compressors with default implementations.<br>
 * You mostly have to implement the two methods: <br>
 * <code>subsume(v, u) : tells if v S u</code><br>
 * <code>restore(u) : restore the compressed vertex u to its initial state</code><br>
 * <br>
 */
public abstract class AbstractCompressor implements GraphCompressor {

	/**
	 * placeholder to store optional temporary set
	 */
	protected GSet<Vertex> storedSet_;
	protected GSet<Vertex> retrievedSet_;

	/**
	 * an empty RVertexSet
	 */
	protected final Iterable<Vertex> emptySet_ = new RVertexSet();
	
	
	/*
	 * default API : compression
	 */

	/**
	 * compress single vertex v on root vertex root
	 */
	public void compressVertex(Vertex v, Vertex root, Graph graph) {
		// check if we can compress
		if (v == root)
			throw new NotCompressibleVertexException("cannot self compress");
		if (isAlias(root))
			throw new NotCompressibleVertexException("invalid root vertex");
		
		// if v is itself a root we need to move its aliases
		if (isRoot(v)) {
			GSet<Vertex> aliases = graph.vertexSetFactory().make(getAliases(v));
			for (Vertex u : aliases) {
				C3PartUtil.setRoot(u, root);
			}
			C3PartUtil.removeAliases(v);
		}
		
		// disconnect vertex v
		graph.removeVertex(v);
		// add v in root aliases and set Root(v) to root
		C3PartUtil.addAlias(root, v);
	}
	
	/**
	 * compress graph
	 * @Implement
	 */
	public void compress(Graph graph) {
		// we need these sets to avoid concurrent access in graph
		GSet<Vertex> vSet = graph.vertexSetFactory().make();
		GSet<Vertex> uSet = graph.vertexSetFactory().make();

		vSet.addIterable(graph.vertices());
		
		for (Vertex v : vSet) {
			if (! graph.hasVertex(v)) // already compressed
				continue;
			uSet.clear();
			uSet.addIterable(v.neighbors());
			for (Vertex u : uSet) {
				if (subsumes(v, u, graph)) {
					this.compressVertex(u, v, graph);
				}
			}
		}
	}

	/**
	 * same as <code>compress</code> with logging compression rate
	 * @Implement
	 */
	public void compress(Graph graph, Logger logger) {
		int n0 = graph.nbVertices();
		int m0 = this.nbVertices(graph);
		int e0 = graph.nbEdges();
		
		this.compress(graph); // call compress
		
		int n1 = graph.nbVertices();
		int m1 = this.nbVertices(graph);
		int e1 = graph.nbEdges();

		if (m0 != m1) {
			logger.severe("* Warning * compress " + m0 + " != " + m1);
		}

		if (n0 == n1)
			return;
		
		String header = this.getClass().getSimpleName();

		float nperc = (float) ((n0-n1) * 100) / (float) (n0 != 0 ? n0 : 1);
		float eperc = (float) ((e0-e1) * 100) / (float) (e0 != 0 ? e0 : 1);
		logger.info(header + " Vertices " + n0 + "[" + m0 + "] -> " + n1 + "[" + m1 + "] (" + String.format("%.1f", nperc) +"%) " 
				           + " Edges "   + e0 + " -> " + e1 + " (" + String.format("%.1f", eperc) +"%)");
	}

	/*
	 * default API : expansion
	 */
	
	/**
	 * expand graph to its uncompressed state
	 * @Implement
	 */
	public void expand(Graph graph) {
		// we need these sets to avoid concurrent access in graph
		GSet<Vertex> vertices = graph.vertexSetFactory().make(graph.vertices());
		GSet<Vertex> aliases = graph.vertexSetFactory().make();
		for (Vertex v : vertices) {
			if (isRoot(v)) {
				aliases.clear();
				aliases.addIterable(C3PartUtil.getAliases(v));
				for (Vertex u : aliases) {
					this.expandVertex(u, graph);
				}
			}
		}
	}
	
	/*
	 * default API : misc utilities
	 */

	/**
	 * return all vertices in compressed graph.
	 * for efficiency, vertices should be returned
	 * in an ordered way : any root vertex should be immediately 
	 * followed by it aliases.
	 * @Implement
	 */
	public Iterable<Vertex> getVertices(Graph graph) {
		ArrayList<Vertex> list = new ArrayList<Vertex>();
		for (Vertex v : graph.vertices()) {
			list.add(v);
			if (isRoot(v)) {
				for (Vertex u : this.getAliases(v)) {
					list.add(u);
				}
			}
		}
		return list;
	}

	/**
	 * return the number of vertices in graph
	 * (including compressed one)
	 * @Implement
	 */
	public int nbVertices(Graph graph) {
		int count = 0;
		for (Vertex v : graph.vertices()) {
			count++;
			if (isRoot(v)) {
				count += C3PartUtil.getAliases(v).size();
			}
		}
		return count;
	}
	
	/**
	 * return neighbors of vertex v, including
	 * aliases of neighbors
	 * @Implement
	 */
	public Iterable<Vertex> getNeighbors(Vertex v) {
		ArrayList<Vertex> list = new ArrayList<Vertex>();
		for (Vertex u : v.neighbors()) {
			list.add(u);
			if (isRoot(u)) {
				for (Vertex w : getAliases(u)) {
					list.add(w);
				}
			}
		}
		/* AV: ne sert a rien
		if (isRoot(v)) {
			for (Vertex w : getAliases(v)) {
				list.add(w);
			}
		}
		*/
		return list;
	}
	
	/**
	 * tell if vertex is an Alias (i.e. compressed) vertex.
	 * @Implement
	 */
	public boolean isAlias(Vertex v) {
		return C3PartUtil.hasRoot(v);
	}
	
	/**
	 * tell if vertex is a root vertex.
	 * i.e. if it owns at least one alias
	 * @Implement
	 */
	public boolean isRoot(Vertex v) {
		return (    C3PartUtil.hasAlias(v)
		        && (C3PartUtil.getAliases(v).size() != 0));
	}

	/**
	 * tell if vertex is a normal vertex.
	 * i.e. neither root nor alias
	 * @Implement
	 */
	public boolean isNormal(Vertex v) {
		return (! (isRoot(v) || isAlias(v)));
	}
	
	/**
	 * get root of a compressed vertex
	 * @Implement
	 */
	public Vertex getRoot(Vertex v) {
		return C3PartUtil.getRoot(v);
	}
	
	/**
	 * get aliases of root vertex
	 * @Implement
	 */
	public Iterable<Vertex> getAliases(Vertex v) {
		if (! C3PartUtil.hasAlias(v))
			return emptySet_;
		return C3PartUtil.getAliases(v);
	}

	/*
	 * Optional API
	 */

	/**
	 * expand graph without edges
	 * @Implement
	 */
	public void flatten(Graph graph) {
		// we need these sets to avoid concurrent access in graph
		GSet<Vertex> vertices = graph.vertexSetFactory().make(graph.vertices());
		GSet<Vertex> aliases = graph.vertexSetFactory().make();
		for (Vertex v : vertices) {
			graph.disconnectVertex(v);
			if (isRoot(v)) {
				aliases.clear();
				aliases.addIterable(C3PartUtil.getAliases(v));
				for (Vertex u : aliases) {
					C3PartUtil.removeRoot(u);
					graph.addPhysicalVertex(u);
				}
			}
		}
	}
	
	/**
	 * temporarily stores a copy of set of vertices
	 * optional operation
	 * @Implement
	 */
	public void storeSet(GSet<Vertex> set) {
		if (storedSet_ == null) {
			storedSet_ = set.factory().make();
			retrievedSet_ = set.factory().make();
		}	
		storedSet_.clear();
		storedSet_.addAll(set);
	}
	
	/**
	 * retrieve a copy of previously stored set
	 * optional operation
	 * @Implement
	 */
	public GSet<Vertex> retrieveSet() {
		retrievedSet_.clear();
		retrievedSet_.addIterable(storedSet_);
		return retrievedSet_;
	}
	
	/**
	 * get optional parameters info
	 * @Implement
	 */
	public String getParamInfo() {
		return "";
	}
	
}
