package helix.graph.algo.c3part.compressor;

/**
 * a base skeleton for multigraph compressor with the following partial subsumption:<br>
 * <code>
 * V S U <=> Colors(U) = Colors(V) = Colors(Edge(U,V))  // @NonStrict
 *           and 
 *           subsumes_(V,U)
 * </code><br>
 * // @NonStrict : for OTF, the conditions reads Colors(Edge(U,V)) = AllColors<br>
 * 
 * the expandVertex base mechanism is:<br>
 * <code>
 *   remove V from Root aliases
 *   add back V in graph
 *   for W in patternSet_(Root)
 *     restoreEdge_(V, W)
 *   connect V back to Root
 * </code><br>
 * note: <code>patternSet_(X)</code> is a set of vertices such
 * that <code>Neighbors(X) c patternSet_(X)</code><br>
 * by default patternSet_(X) = Neighbors(X)<br>
 * 
 * In subclasses, you therefore have to provide the three delegate methods:<br>
 * <code>patternSet_, subsumes_ and restoreEdge_</code>
 * 
 */

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public abstract class AbstractMultiCompressor extends AbstractCompressor implements MultiGraphCompressor {

	/*
	 * API
	 */

	/**
	 * delegate subsumption
	 */
	public abstract boolean subsumes_(Vertex V, Vertex U, Graph multi);

	/**
	 * delegate edge restoration
	 * restore (if needed) edge between V and W 
	 */
	public abstract void restoreEdge_(Vertex V, Vertex W, Vertex Root, Graph multi);

	/**
	 * delegate pattern set
	 */
	public abstract Iterable<Vertex> patternSet_(Vertex V, Graph multi);
	
	/**
	 * tells if vertex V subsumes vertex U
	 * @Implement
	 */
	public boolean subsumes(Vertex V, Vertex U, Graph multi) {
		Edge UV = V.getEdgeTo(U);
		if (UV == null)
			return false;
		RColorSet vcolors = RColorSet.getColors(V);
		if (! vcolors.equals(RColorSet.getColors(U)))
			return false;
		if (! vcolors.equals(RColorSet.getColors(UV)))
			return false;
		return this.subsumes_(V, U, multi);
	}

	/**
	 * restore compressed vertex v to its initial state
	 * @Implement
	 */
	public void expandVertex(Vertex V, Graph multi) {
		// get root of V
		Vertex Root = C3PartUtil.getRoot(V);
		if (Root == null)
			throw new NotExpandableVertexException();
		// remove v from Root aliases
		C3PartUtil.removeRoot(V);
		// add back V in graph
		multi.addPhysicalVertex(V);
		// restore Edges from V to Root neighbors
		for (Vertex W : patternSet_(Root, multi)) {
			this.restoreEdge_(V, W, Root, multi);
		}
		// connect V back to its Root 
		Edge E = multi.addEdge(Root, V);
		RColorSet.getColors(E).setColors(RColorSet.getColors(Root));
	}

	/**
	 * Get color of edge (V,U) if it exists,
	 * else compute the colors it should have
	 * according to this compressor.
	 * This may return emty colorSet is V and U are not connected.
	 * by default this is the true colors (computed from datagraph).
	 * Please override for inSet compressors
	 * @Implement
	 */
	public RColorSet getColors(Vertex V, Vertex U) {
		Edge E = V.getEdgeTo(U);
		return (E != null ? RColorSet.getColors(E) : C3PartUtil.getTrueColors(V, U));
	}
}
