package helix.graph.algo.c3part.compressor;

import java.util.logging.Logger;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

/**
 * A graph compressor is used to compress vertices and edges of graphs.<br>
 * It is based on a <b>compression condition</b> that is expressed by
 * a subsumption relation <code>S</code> between vertices.<br>
 * v S u => u can be removed from graph, u is stored in Alias(v), v is stored as Root(u).
 * (note that S is either a partial order or an equivalence relation)
 * The storage allows for further expansion of the graph to its initial state.<br>
 * 
 */
public interface GraphCompressor {

	/*
	 * Generic API
	 */

	/**
	 * tells if vertex v subsumes vertex u
	 */
	public boolean subsumes(Vertex v, Vertex u, Graph graph);

	/**
	 * compress single vertex v on root vertex
	 */
	public void compressVertex(Vertex v, Vertex root, Graph graph);
	
	/**
	 * compress graph
	 */
	public void compress(Graph graph);

	/**
	 * same as <code>compress</code> with logging compression rate
	 */
	public void compress(Graph graph, Logger logger);
	
	/**
	 * restore compressed alias vertex v to its initial state
	 */
	public void expandVertex(Vertex v, Graph graph) throws NotExpandableVertexException;
	
	/**
	 * expand graph to its uncompressed (initial) state
	 */
	public void expand(Graph graph);
	
	/**
	 * return all vertices in compressed graph.
	 * for efficiency, vertices should be returned
	 * in an ordered way : any root vertex should be immediately 
	 * followed by it aliases.
	 */
	public Iterable<Vertex> getVertices(Graph graph);

	/**
	 * return the number of vertices in graph
	 * (including compressed one)
	 */
	public int nbVertices(Graph graph);
	
	/**
	 * return neighbors of vertex v, including
	 * aliases of neighbors
	 */
	public Iterable<Vertex> getNeighbors(Vertex v);

	/**
	 * tell if vertex is an Alias vertex.
	 */
	public boolean isAlias(Vertex v);
	
	/**
	 * tell if vertex is a root vertex.
	 * i.e. if it owns at least one alias
	 */
	public boolean isRoot(Vertex v);

	/**
	 * tell if vertex is a normal vertex.
	 * i.e. neither root nor alias
	 */
	public boolean isNormal(Vertex v);
	
	/**
	 * get root of a compressed vertex
	 */
	public Vertex getRoot(Vertex v);
	
	/**
	 * get aliases of root vertex
	 * return empty set if v is not root
	 */
	public Iterable<Vertex> getAliases(Vertex v);

	/*
	 * optional API
	 */

	/**
	 * expand graph without edges
	 * AV TODO temporary placeholder
	 */
	public void flatten(Graph graph);

	/**
	 * temporarily stores a set of vertices
	 * this is an optional operation
	 */
	public void storeSet(GSet<Vertex> set);
	
	/**
	 * retrieve previously stored set
	 * this is an optional operation
	 */
	public GSet<Vertex> retrieveSet();
	
	/**
	 * get optional parameters info
	 */
	public String getParamInfo();
}
