package helix.graph.algo.c3part.compressor;

/**
 * a multigraph compressor with the following subsumption:<br>
 * <code>
 * V S U <=> Colors(U) = Colors(V) = Colors(Edge(U,V))  // @NonStrict
 *           and 
 *           Neighbors(V) = Neighbors(U)
 *           and 
 *           A W in Neighbors(V), Colors(Edge(U,W)) c Colors(Edge(V,W))
 *           
 * Note that S is a partial order.
 * // @NonStrict : for OTF, the conditions reads Colors(Edge(U,V)) = AllColors<br>
 * </code>
 */

import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class MultiInColorsCompressor extends AbstractMultiCompressor {
	
	/**
	 * internal utility
	 * check inclusion of colors
	 */
	private final boolean includeColors_(Registry r1, Registry r2) {
		return RColorSet.getColors(r1).includeColors(RColorSet.getColors(r2));
	}

	/**
	 * delegate patternSet
	 * @Implement
	 */
	public final Iterable<Vertex> patternSet_(Vertex V, Graph multi) {
		return V.neighbors();
	}
	
	/**
	 * delegate subsumption
	 * @Implement
	 */
	public boolean subsumes_(Vertex V, Vertex U, Graph graph) {
		for (Vertex W : V.neighbors()) {
			if (W == U)
				continue;
			Edge UW = U.getEdgeTo(W);
			if (UW == null)
				return false;
			if (! includeColors_(V.getEdgeTo(W), UW))
				return false;
		}
		for (Vertex W : U.neighbors()) {
			if (W == V)
				continue;
			Edge VW = V.getEdgeTo(W);
			if (VW == null)
				return false;
		}
		return true;
	}
	
	/**
	 * delegate restoration
	 * restore (if needed) edge between V and W 
	 * @Implement
	 */
	public void restoreEdge_(Vertex V, Vertex W, Vertex Root, Graph multi) {
		Edge E = multi.addEdge(V, W);
		E.copy(Root.getEdgeTo(W)); // restore root edge attributes
		
		// restore proper color
		RColorSet colors = this.getColors(V, W);
		RColorSet.getColors(E).setColors(colors);
	}
}
