package helix.graph.algo.c3part.compressor;

/**
 * a multigraph Compressor with the following subsumption:<br>
 * <code>
 * V S U <=> Colors(U) = Colors(V) = Colors(Edge(U,V))  // @NonStrict
 *           and 
 *           Neighbors(U) c Neighbors(V)
 *           
 * Note that S is a partial order relation.
 * 
 * Important note: this compressor uses pseudo colors for root
 * </code>
 */

import java.util.HashSet;
import java.util.Set;

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class MultiInSetCompressor extends AbstractMultiCompressor {
	
	/*
	 * utilities
	 */
	
	/**
	 * utility: collect V and all its aliases in set
	 */
	private final void collectVertices_(Vertex V, Set<Vertex> set) {
		set.clear();
		set.add(V);
		for (Vertex U : C3PartUtil.getAliases(V)) {
			set.add(U);
		}
	}

	/**
	 * get pseudo colors between two vertices V and U
	 * according to all of their aliases.
	 * just a shorthand for C3PartUtil.getPseudoColors
	 */
	private final RColorSet getPseudoColors_(Vertex V, Vertex U) {
		return C3PartUtil.getPseudoColors(V, U);
	}
	
	/**
	 * set pseudo colors between two vertices V and U
	 * according to all of their aliases
	 */
	protected void setPseudoColors_(Vertex V, Vertex U) {
		RColorSet UVcolors = RColorSet.getColors(V.getEdgeTo(U));
		UVcolors.setColors(getPseudoColors_(V, U));
	}
	
	/**
	 * update colors of V-U edge when adding W as an alias of V
	 */
	protected void updatePseudoColors_(Vertex V, Vertex U, Vertex W) {
		RColorSet UVcolors = RColorSet.getColors(V.getEdgeTo(U));
		UVcolors.orColors(getPseudoColors_(W, U));
	}
	
	/**
	 * set pseudo colors around Root
	 * according to all aliases in Root and its neighbors
	 */
	private final void setRootColors_(Vertex Root) {
		for (Vertex U : Root.neighbors()) {
			setPseudoColors_(Root, U);
		}
	}
	
	/**
	 * update pseudo colors around Root
	 * according to alias V in Root
	 */
	private final void updateRootColors_(Vertex Root, Vertex V) {
		for (Vertex U : Root.neighbors()) {
			updatePseudoColors_(Root, U, V);
		}
	}

	/*
	 * API
	 */

	/**
	 * delegate patternSet
	 * @Implement
	 */
	public final Iterable<Vertex> patternSet_(Vertex V, Graph multi) {
		return V.neighbors();
	}
	
	/**
	 * delegate subsumption
	 */
	public boolean subsumes_(Vertex V, Vertex U, Graph graph) {
		// first check true colors between V and U
		if (! RColorSet.getColors(V).equals(C3PartUtil.getTrueColors(V, U)))
			return false;
		// now check inclusion of neighbors
		for (Vertex W : U.neighbors()) {
			if (W == V)
				continue;
			Edge VW = V.getEdgeTo(W);
			if (VW == null)
				return false;
		}
		return true;
	}
	
	/**
	 * delegate restoration
	 * restore (if needed) edge between V and W 
	 * @Implement
	 */
	public void restoreEdge_(Vertex V, Vertex W, Vertex Root, Graph multi) {

		// get actual color
		
		RColorSet colors = getColors(V, W);
		
		if (colors.isEmpty())
			return;

		// add edge
		
		multi.addEdge(V, W);
		setPseudoColors_(V, W);
	}
	
	/**
	 * compress single vertex v on root vertex root
	 * setup pseudo colors of roots
	 */
	// private variable to speedup
	private HashSet<Vertex> setAlias_ = new HashSet<Vertex>();
	@Override
	public void compressVertex(Vertex V, Vertex Root, Graph multi) {
		// keep V and its (potential) aliases
		this.collectVertices_(V, setAlias_);

		// compress as usual
		super.compressVertex(V, Root, multi);

		// and recolorize our root
		for (Vertex A : setAlias_) {
			this.updateRootColors_(Root, A);
		}
	}
	
	/**
	 * expand single vertex
	 * and recompute root colors
	 */
	@Override
	public void expandVertex(Vertex V, Graph multi) {
		Vertex Root = this.getRoot(V);
		super.expandVertex(V, multi);
		// we cannot just update here... we need to recompute colors
		this.setRootColors_(Root);
		this.setRootColors_(V);
	}

	/**
	 * override getColors
	 * get proper color between two vertices
	 * if there is a physical edge return this color
	 * else compute pseudocolors.
	 */
	@Override
	public RColorSet getColors(Vertex V, Vertex U) {
		Edge e = V.getEdgeTo(U);
		return (e != null ? RColorSet.getColors(e) : getPseudoColors_(V,U)); 
	}
	
	/*
	 * Improvements
	 */
	
	/**
	 * compress graph
	 * -> avoid recoloring root at each compression
	 */
	@Override
	public void compress(Graph graph) {
		// we need these sets to avoid concurrent access in graph
		GSet<Vertex> vSet = graph.vertexSetFactory().make();
		GSet<Vertex> uSet = graph.vertexSetFactory().make();

		vSet.addIterable(graph.vertices());
		
		for (Vertex V : vSet) {
			if (! graph.hasVertex(V)) // already compressed
				continue;
			uSet.clear();
			uSet.addIterable(V.neighbors());
			boolean compressed = false;
			for (Vertex U : uSet) {
				if (subsumes(V, U, graph)) {
					compressed = true;
					super.compressVertex(U, V, graph);
				}
			}
			if (compressed) {
				setRootColors_(V);
			}
		}
	}
	
	
}
