package helix.graph.algo.c3part.lookahead;

/**
 * a lookahead connector providing connected component of *all datanodes* in
 * multigraph tuples (i.e. not only terminals) 
 * restricted to nodes of given color and following *any link* 
 * See OneWayPathConnector and TwoWayPathConnector for more restrictive
 * versions
 */

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class AllPathConnector extends ExtendedConnector {

	protected static LookaheadConnector instance_ = new AllPathConnector();

	/**
	 * singleton class
	 */
	protected AllPathConnector() {
	}

	public static LookaheadConnector getInstance() {
		return instance_;
	}

	/**
	 * utility: get the connected component of datanode v
	 * restricted to nodes of given color and following any link.
	 * note: improvement : forbidColors : set of colors that can be skipped during visit.
	 * may be empty for tests
	 */
	private final void allPathConnected_(Vertex v, int color, RColorSet forbidColors, GSet<Vertex> visited, GSet<Vertex> store) {
		if (forbidColors.hasColor(RColorSet.getColors(v).singleColor()))
			return;
		if (visited.contains(v))
			return;
		visited.add(v);
		if (RColorSet.hasColor(v, color))
			store.add(v);
		for (Vertex u : v.neighbors()) {
			allPathConnected_(u, color, forbidColors, visited, store);
		}
	}

	/**
	 * @Implement
	 * get the connected component of all datanodes in multigraph tuples
	 * restricted to nodes of given color and following any link.
	 * store them into 'store'
	 */
	public void getLookAhead(Graph datagraph, Graph multigraph, int color, GSet<Vertex> store) {
		GSet<Vertex> former = getConnectors(multigraph);
		GSet<Vertex> visited = datagraph.vertexSetFactory().make();
		RColorSet forbidColors = new RColorSet(RColorSet.getColors(multigraph));
		forbidColors.clearColor(color);
		for (Vertex v : former)
			allPathConnected_(v, color, forbidColors, visited, store);
	}

}
