package helix.graph.algo.c3part.lookahead;

/**
 * In OTFQLocal, during multigraph expansion with a new color 'color'
 * we need to add 'lookahead' stars in front of some datanodes
 * (i.e. to produce new multinodes of the form (*,*...*, v).
 * These datanodes 'v' are called 'pseudoterminals'.
 * They are of color 'color' and are connected 'in some way' to datanodes
 * in Nodes(V).
 * 
 * The LookAheadConnector subclasses provide various algorithms
 * to collect these pseudoterminals in order to test various
 * algorithms or implementations.
 * 
 * This class and subclasses are therefore intended for developers
 * to experiment with various lookahead connectors.
 * 
 * important subclasses are:<br>
 * 
 * FullColorConnector  : defines the pseudoterminals as all datagraph vertices of color 'color'<br>
 * ColorPathConnector  : defines the pseudoterminals as all vertices of color 'color' connected
 * 						 to current terminals by link of the same color<br>
 * AllPathConnector    : defines the pseudoterminals as all vertices of color 'color' connected
 *                       to any vertex in multigraph tuples by any link<br>
 * TwoWayPathConnector : defines the pseudoterminals as all vertices of color 'color' connected
 *                       to any vertex in multigraph tuples by a 'two-ways' path<br>
 * OneWayPathConnector : defines the pseudoterminals as all vertices of color 'color' connected
 *                       to any vertex in multigraph tuples by a 'one-way' path<br>
 * <br>
 * In addition the following subclass is used for special cases: <br>
 * TerminalsConnector  : defines the pseudoterminals as current terminals<br>
 * <br>
 * the generic call is <code>getLookahead(Graph datagraph, Graph multigraph, int color)</code>
 * where datagraph is the current datagraph, multigraph is the current multigraph (Clas) beeing 
 * expanded (note that the first step expansion should have been done already), color is the
 * current color being added (i.e. the latest color added in multigraph).
 */

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;

public abstract class LookaheadConnector {

	/**
	 * API
	 */

	/**
	 * get all datanodes from multigraph (except stars) that may be connected
	 * to pseudoterminals.
	 * 'connector' means either terminals or all datanodes in multinodes
	 * depending upon the LookAheadConnector definition.
	 * you must implement this method
	 */
	protected abstract GSet<Vertex> getConnectors(Graph multigraph);

	/**
	 * get the pseudoterminals and append them into 'store'.
	 * you must implement this method
	 * @param datagraph Graph current datagraph
	 * @param multigraph Graph current multigraph (clas) beeing expanded
	 * @param color int color of terminals to find
	 * @param store GSet[Vertex] set to store results
	 */
	public abstract void getLookAhead(Graph datagraph, Graph multigraph, int color, GSet<Vertex> store);

	/**
	 * get the pseudoterminals and return a new GSet[Vertex]
	 * @param datagraph Graph current datagraph
	 * @param multigraph Graph current multigraph (clas) beeing expanded
	 * @param color int color of terminals to find
	 */
	public GSet<Vertex> getLookahead(Graph datagraph, Graph multigraph, int color) {
		GSet<Vertex> store = datagraph.vertexSetFactory().make();
		this.getLookAhead(datagraph, multigraph, color, store);
		return store;
	}
	
}
