package helix.graph.algo.c3part.lookahead;

/**
 * a lookahead connector providing the connected component of *all datanodes* in
 * current multigraph (i.e. not only terminals) 
 * restricted to nodes of given color and following the two-ways path.
 * A 'two-ways' path is an optional epsilon path (called 'before')
 * followed by an optional path on the same color (not necessarily 'color', called pending),
 * followed by an optional epsilon path (called 'after').
 * example: v -e-> u -ucolor-> u' -e-> v'
 */

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class TwoWayPathConnector extends ExtendedConnector {

	protected static LookaheadConnector instance_ = new TwoWayPathConnector();

	/**
	 * singleton class
	 */
	protected TwoWayPathConnector() {
	}

	public static LookaheadConnector getInstance() {
		return instance_;
	}

	/**
	 * get datanodes of color 'color' that are connected to v by following
	 * a 'two-ways' path. A 'two-ways' path is an optional epsilon path (called 'before')
	 * followed by an optional path on the same color (not necessarily 'color', called pending),
	 * followed by an optional epsilon path (called 'after').
	 * example: v -e-> u -ucolor-> u' -e-> v'
	 * 
	 * note: improvement : forbidColors : set of colors that can be skipped during a pending
	 * (or after ?) path. may be empty for tests
	 * 
	 */
	protected void twoWaysConnected_(Vertex v, int color, RColorSet forbidColors, 
			                         boolean before, boolean pending, 
			                         GSet<Vertex> visitedBefore, GSet<Vertex> visitedPending, 
			                         GSet<Vertex> visitedAfter, GSet<Vertex> store) {

		int vcolor = RColorSet.getColors(v).singleColor();
		
		if (before) {
			if (visitedBefore.contains(v))
				return;
			visitedBefore.add(v);
		}
		else if (pending) {
			if (forbidColors.hasColor(vcolor))
				return;
			if (visitedPending.contains(v))
				return;
			visitedPending.add(v);
		}
		else {
			if (forbidColors.hasColor(vcolor))
				return;
			if (visitedAfter.contains(v))
				return;
			visitedAfter.add(v);
		}

		if (RColorSet.hasColor(v, color))
			store.add(v);
		
		for (Vertex u : v.neighbors()) {
			if (RColorSet.hasColor(u, vcolor)) { // same color
				if (before || pending)
					twoWaysConnected_(u, color, forbidColors, false, true, 
							          visitedBefore, visitedPending, visitedAfter, store);
			}
			else {								 // epsilon link 
				twoWaysConnected_(u, color, forbidColors, before, false,
						          visitedBefore, visitedPending, visitedAfter, store);
			}
		}
	}

	/**
	 * @Implement
	 * get the connected component of all datanodes in in multigraph tuples
	 * restricted to nodes of given color and following a two-ways path.
	 * a 'two-ways' path. A 'two-ways' path is an optional epsilon path (called 'before')
	 * followed by an optional path on the same color (not necessarily 'color', called pending),
	 * followed by an optional epsilon path (called 'after').
	 * example: v -e-> u -ucolor-> u' -e-> v'.
	 * store them into 'store'
	 */
	public void getLookAhead(Graph datagraph, Graph multigraph, int color, GSet<Vertex> store) {
		GSet<Vertex> former = getConnectors(multigraph);
		GSet<Vertex> visitedBefore = datagraph.vertexSetFactory().make();
		GSet<Vertex> visitedPending = datagraph.vertexSetFactory().make();
		GSet<Vertex> visitedAfter = datagraph.vertexSetFactory().make();
		RColorSet forbidColors = new RColorSet(RColorSet.getColors(multigraph));
		forbidColors.clearColor(color);
		for (Vertex v : former)
			twoWaysConnected_(v, color, forbidColors, true, false, 
					          visitedBefore, visitedPending, visitedAfter, store);
	}
	
}
