package helix.graph.algo.c3part.optimizer;

/**
 * an abstract class to provide basic methods for 'embedded optimizers'.
 * 
 * an 'embedded optimizer' relies on a (optional) parent optimizer, embedded
 * in clas (and retrieved thru clas.get(C3PartUtil.Optimizer)). The parent optimizer
 * can, itself, be embedded. There is therefore a lineage of optimizers, called its
 * ancestors.
 * The basic property of an embedded optimizer is that it will produce the same
 * first colors as its ancestors -see note below-. When these colors are exhausted,
 * it will pickup the next colors (called its proper color) by calling
 * the <code>nextColor_()</code> method.
 * if clas is null or has no color defined yet (and therefore no embedded parent optimizer),
 * then the optimizer call the <code>firstColor_()</code> method.
 * 
 * you have therefore to implement <code>nextColor_()</code> and <code>firstColor_()</code> methods.
 * 
 * important note:  In order to work properly embedded optimizer have slightly different
 * interpretation of the <code>getUsedColors()</code> method. It actually returns the
 * list of <code>getUsedColors()</code> of the parent optimizer (recursively this is therefore
 * the largest list of colors produced by any of the ancestors) followed by the proper 
 * colors produced by this optimizer.
 * This means that, even if an embedded optimizer has never called the <code>next()</code>
 * then the <code>getUsedColors()</code> method will return the (possibly) non empty list.
 * <br>
 * note: by construction the <code>firstColor_()</code> method is called only once for
 * the ancestor in the lineage. 
 * <br>
 * note: the internal Comparator class is used to revert inequality tests in some
 * optimizer (i.e choosing min or max)
 * 
 */

import helix.graph.algo.c3part.C3PartAlgo;
import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.model.Graph;
import helix.graph.model.impl.RColorSet;
import helix.graph.util.EmptyIterator;

import java.util.Iterator;
import java.util.NoSuchElementException;

public abstract class EmbeddedColorOptimizer extends AbstractColorOptimizer {

	/**
	 * internal classes for min/max computations
	 */
	
	protected abstract class Comparator {
		public abstract boolean compare(int x, int y);
		public abstract int first(int array[], int from);
		public abstract String info();
	}
	protected class MinComparator extends Comparator {
		public final boolean compare(int x, int y) {
			return (x < y);
		}
		public int first(int array[], int from) {
			return OptimizerUtil.indexMin(array, from);
		}
		public String info() {
			return C3PartOptions.OPTIMIZER_MODE_MIN;
		}
	}
	protected class MaxComparator extends Comparator {
		public final boolean compare(int x, int y) {
			return (x > y);
		}
		public int first(int array[], int from) {
			return OptimizerUtil.indexMax(array, from);
		}
		public String info() {
			return C3PartOptions.OPTIMIZER_MODE_MAX;
		}
	}
	
	
	protected int maxIndex_;
	protected int currentIndex_;

	protected RColorSet clasColors_;

	protected Iterator<Integer> usedColorsIterator_;

	protected Comparator comparator_;

	/**
	 * required constructor
	 */
	public EmbeddedColorOptimizer(C3PartAlgo algo, Graph datagraph, Graph clas) {
		
		currentIndex_ = 0;
		maxIndex_ = RColorSet.nbColors(datagraph) - 1;
		
		if (clas == null) {	// this is the first color (no previous guess)
			clasColors_ = new RColorSet();
			usedColorsIterator_ = new EmptyIterator<Integer>();
		}
		else {				// get our previous optimizer
			clasColors_ = new RColorSet(RColorSet.getColors(clas));
			//
			// get previous optimizer and its used colors
			//
			ColorOptimizer embedded = C3PartUtil.getOptimizer(clas);
			if (embedded != null) {
				usedColors_.addAll(embedded.getUsedColors());
			}
			usedColorsIterator_ = usedColors_.iterator();
		}

		// this may be useless for some optimizer, but does not hurt
		String mode = algo.getPhysicalOptions().optimizerMode;
		if (C3PartOptions.OPTIMIZER_MODE_MAX.equalsIgnoreCase(mode))
			comparator_ = new MaxComparator();
		else
			comparator_ = new MinComparator();
	
	}

	/*
	 * Iterator API
	 */

	/**
	 * default hasNext()
	 */
	public boolean hasNext() {
		return (currentIndex_ < maxIndex_);
	}
	
	/**
	 * default next()
	 */
	public Integer next() {

		if (! hasNext()) {
			throw new NoSuchElementException();
		}
		
		if (this.giveup_) {
			System.err.println("[EmbeddedColorOptimizer] Warning next() call after giving up");
		}

		int color; // color to guess
			
		currentIndex_++;
		
		if (usedColorsIterator_.hasNext()) {	// previous colors from iterator
			color = usedColorsIterator_.next();
		}
		else {									// next color to guess
			if (clasColors_.isEmpty()) {	   		// this is the first color for clas
			   color = this.firstColor_();
			}
			else {
			   color = this.nextColor_();	 	   	// this is the next color for clas
			}
			
			this.storeColor_(color);				// keep this color in producedColors
		}

		return color;
	}

	/*
	 * Added API
	 */

	/**
	 * you have to implement this : guess the first color for the clas
	 */
	protected abstract int firstColor_();

	/**
	 * you have to implement this : guess the next color for the clas
	 */
	protected abstract int nextColor_();
	
}
