package helix.graph.algo.c3part.optimizer;

/**
 * an embedded optimizer which provide nextColor_() based on a global
 * (i.e. Clas independent) optimization. This optimizer chooses the
 * first color that yields the minimum number of vertices and then
 * the following colors that have the smaller number of epsilon links
 * to the previous one.
 * 
 */

import helix.graph.algo.c3part.C3PartAlgo;
import helix.graph.algo.c3part.aggregator.NodeAggregator;
import helix.graph.algo.c3part.reporter.GraphStatistics;
import helix.graph.model.Graph;
import helix.graph.model.impl.RColorSet;

public class GlobalColorOptimizer extends EmbeddedColorOptimizer {

	protected Graph datagraph_; 
	protected NodeAggregator aggregator_;
	
	public GlobalColorOptimizer(C3PartAlgo algo, Graph datagraph, Graph clas) {
		
		super(algo, datagraph, clas);
		
		datagraph_ = datagraph;

		aggregator_ = algo.getAggregatorFor();
	}
	
	/*
	 * required API
	 */

	@Override
	/**
	 * override storeColor_ to prevent storing (see below)
	 */
	protected void storeColor_(int color) {
		// don't store anything
	}
	
	/**
	 * provides first color based on the datagraph: choose color with the minimum
	 * number of vertices.
	 */
	
	/* implementation note:
	 * for embedded optimizers, the firstColor_ method is only called
	 * once (for the ancestor). The trick is therefore to determine the
	 * color permutation at this step and to register the whole permutation,
	 * once for all, in the usedColors. In that way, the nextColor_() method
	 * will actually never be called on the offspring.
	 * We should also prevent storeColor_ to store the colors each time.
	 */

	protected int firstColor_() {
		
		if (! this.usedColors_.isEmpty()) { // calculation already done
			return this.usedColors_.get(0);
		}

		int firstColor;
		
		// get first color
		//
		if (aggregator_.getPreferredColor() > 0) { 
			// check if our aggregator has some fix color
			firstColor = aggregator_.getPreferredColor();
		}
		else {
			// choose first color
			int countColors[] = GraphStatistics.verticesPerColor(datagraph_);
			firstColor = comparator_.first(countColors, 1); // skip color 0
		}
		
		usedColors_.add(firstColor);
		
		// precompute other colors once for all
		// and keep them in our usedColors so our offsprings will never have
		// to compute anything
		//
		int nbColor = RColorSet.nbColors(datagraph_);
		int matrix[][] = GraphStatistics.epsilonMatrix(datagraph_);
		
		// select other elements
		int pivot = firstColor;
		while (usedColors_.size() < nbColor-1) {
			int iminmax = 0, vminmax = -1;
			for (int i = 1 ; i < nbColor ; i++) {
				if (usedColors_.contains(i)) 
					continue;
				if ((vminmax < 0) || comparator_.compare(matrix[pivot][i], vminmax)) {
					vminmax = matrix[pivot][i];
					iminmax = i;
				}
			}
			pivot = iminmax;
			usedColors_.add(iminmax);
		}
				
		return firstColor;
	}


	/**
	 * this method will actually never be called
	 */
	protected int nextColor_() {
		System.err.println("[GlobalColorOptimizer] nextColor_ should never be called !");
		return 0;
	}
	
	@Override
	/**
	 * preview
	 */
	public String preview() {
		StringBuffer buffer = new StringBuffer();
		this.firstColor_(); // force calculation
		for (int color : usedColors_) {
			buffer.append(" " + color);
		}
		return "[" + buffer.substring(1) + "]";
	}
	
	@Override
	/**
	 * optional parameters info
	 */
	public String getParamInfo() {
		return comparator_.info();
	}
}
	