package helix.graph.algo.c3part.optimizer;

/**
 * an embedded optimizer which provide nextColor_() based on the number
 * of terminals extending current class. This optimizer chooses the
 * color that yields the minimum number of terminals (including 0 terminals)
 */

import helix.graph.algo.c3part.C3PartAlgo;
import helix.graph.algo.c3part.aggregator.NodeAggregator;
import helix.graph.algo.c3part.reporter.GraphStatistics;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

import java.util.Iterator;

public class TerminalsColorOptimizer extends EmbeddedColorOptimizer {

	protected Graph datagraph_; 
	protected GSet<Vertex> clasVertices_;
	protected NodeAggregator aggregator_;

	protected RColorSet graphColors_;
	
	
	public TerminalsColorOptimizer(C3PartAlgo algo, Graph datagraph, Graph clas) {
		
		super(algo, datagraph, clas);
		
		datagraph_ = datagraph;

		aggregator_ = algo.getAggregatorFor();

		graphColors_ = RColorSet.getColors(datagraph);
		
		if (clas != null) {
			clasVertices_ = clas.vertexSetFactory().make(clas.vertices());
		}
	}
	
	@Override
	public void giveup() {
		super.giveup();
		clasVertices_ = null;
	}
	
	/*
	 * required API
	 */
	/**
	 * provides first color based on the datagraph. choose color with the minimum
	 * number of vertices
	 */
	protected int firstColor_() {
		
		// check if our aggregator has some fix color
		//
		if (aggregator_.getPreferredColor() > 0) { 
			return aggregator_.getPreferredColor();
		}

		// ok go ahead to find a first color
		//
		int countColors[] = GraphStatistics.verticesPerColor(datagraph_);
		return comparator_.first(countColors, 1); // skip color 0
	}


	/**
	 * provides next color based on the number
	 * of terminals extending current class. This optimizer chooses the
	 * color that yields the minimum number of terminals (including 0)
	 */
	protected int nextColor_() {

		RColorSet restColors = new RColorSet(clasColors_);
		restColors.xorColors(graphColors_);
		restColors.clearColor(0); // don't use this color

		Iterator<Integer> rest = restColors.iterator();
		int colorminmax = rest.next();
		int nminmax = getSizeTerminals_(colorminmax);
		while (rest.hasNext()) {
			int color = rest.next();
			int n = getSizeTerminals_(color);
			if (comparator_.compare(n, nminmax)) {
				nminmax = n;
				colorminmax = color;
			}
		}

		return colorminmax;
	}
	
	/**
	 * utility : compute size of color terminals
	 */
	
	protected int getSizeTerminals_(int color) {

		GSet<Vertex> terminals = clasVertices_.factory().make();
		
		for (Vertex V : clasVertices_) {
			terminals.addAll(aggregator_.getTerminals(V, color));
		}

		return terminals.size();
	}

	@Override
	/**
	 * optional parameters info
	 */
	public String getParamInfo() {
		return comparator_.info();
	}
	
	
}
	