package helix.graph.algo.c3part.preprocessor;

/**
 * chromatize multigraph : only keep edges that have specified colors.
 * note: color 0 is assumed to be unused in multigraph (epsilon relation)
 *       so colors start at 1.
 * note: edge colors may be > m=maxColor in source <b>datagraph</b>,
 *       this is called shifted colors (see <code>MultiGraphCloser</code>)
 *       because of this, all colors computation are performed modulo m
 *       and shifted colors are removed at the end of the process.
 */
import java.util.List;

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Edge;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class Chromatizer extends MultigraphCleaner {
	
	protected int maxColor_;
	protected RColorSet requiredColors_ = new RColorSet();
	protected RColorSet tmpColors_ = new RColorSet(); // to speed up
	protected RColorSet xorColors_ = new RColorSet(); // to speed up

	/**
	 * tell if edge has all required colors
	 */
	protected final boolean hasRequiredColors_(Edge e) {
		//
		// first we perform a quick check using e colors only
		//
		// check if requiredColors are included in colors(e)
		//
		tmpColors_.clearColors();
		for (int color : RColorSet.getColors(e)) {
			if (color == 0) continue;
			int folded = ((color - 1) % maxColor_) + 1;
			tmpColors_.setColor(folded);
		}
		tmpColors_.andColors(requiredColors_);     // intersection

		if (tmpColors_.equals(requiredColors_))    // inclusion
			return true;
		//
		// required colors are not included in found colors
		// we now have to check for stars
		//
		// the following trick assumes that if a color is required
		// but absent from e.V1 or e.V2 then this color is
		// represented by a star.
		//
		xorColors_.setColors(requiredColors_);
		xorColors_.xorColors(RColorSet.getColors(e.v1()));
		tmpColors_.orColors(xorColors_);
		xorColors_.setColors(requiredColors_);
		xorColors_.xorColors(RColorSet.getColors(e.v2()));
		tmpColors_.orColors(xorColors_);
		tmpColors_.andColors(requiredColors_);

		return tmpColors_.equals(requiredColors_);
	}

	/**
	 * remove shifted colors
	 */
	protected final void cleanupColors_(Edge e) {
		RColorSet edgeColors = RColorSet.getColors(e);
		tmpColors_.setColors(edgeColors);
		for (int color : tmpColors_) {
			if (color <= maxColor_) continue;
			edgeColors.clearColor(color);
		}
	}

	/**
	 * default constructor
	 * @param colors required colors
	 * @param maxColor maximum color to use 
	 * normally this should equal the number of colors in the source <b>datagraph</b> minus 1
	 * (because color 0 is unused)
	 */
	public Chromatizer(RColorSet colors, int maxColor) {
		requiredColors_ = new RColorSet(colors);
		maxColor_ = maxColor;
	}

	/**
	 * alternate constructor : use the delta values in user's delta list
	 * to specify which colors are required. a negative delta value indicates
	 * infinite closure and therefore means that the color is actually not
	 * required.
	 * @param colors required colors
	 * @param maxColor maximum color to use 
	 * normally this should equal the number of colors in the source <b>datagraph</b> minus 1
	 * (because color 0 is unused)
	 * @param delta delta list. List size should be >= maxColor or 1.
	 * In the latter case, the same delta is assumed for all colors.
	 */
	public Chromatizer(RColorSet colors, int maxColor, List<Integer> delta) {
		delta = GraphCloser.makeDeltaList(delta, maxColor);
		RColorSet required = new RColorSet(colors);
		for (int color = 0 ; color < delta.size() ; color++) {
			if (delta.get(color) < 0)
				required.clearColor(color+1);
		}
		requiredColors_ = required;
		maxColor_ = maxColor;
	}
	
	/**
	 * implement our run() operation
	 */
	//@Implements
	public void run(Graph graph) {
		
		GSet<Edge> removed = graph.edgeSetFactory().make();

		// -1- get edges to remove
		//
		for (Vertex v : graph.vertices()) {
			for (Edge e : v.edges()) {
				if (! hasRequiredColors_(e))
					removed.add(e);
			}
		}
		
		// -2- remove edges
		for (Edge e : removed) {
			graph.removeEdge(e);
		}
		
		// -3- remove shifted colors
		for (Vertex v : graph.vertices()) {
			for (Edge e : v.edges()) {
				cleanupColors_(e);
			}
		}
	}

	/**
	 * override logParams
	 */
	@Override
	public void logParams() {
		super.logParams();
		log("required colors : " + requiredColors_.toString());
	}

}
