package helix.graph.algo.c3part.preprocessor;

import java.util.List;

import helix.graph.algo.DeltaClosure;
import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.impl.RColorSet;

/**
 * a class to compute delta closure of datagraphs.<br>
 * in a datagraph all vertices are assumed to be colored (not necessarily all edges)
 * and monochromatic.<br>
 * note: color 0 is reserved for the epsilon relation and is specified on edges.<br>
 * note: this class does not close on the epsilon relation. to do so, use the
 * <code>EpsilonCloser</code> class instead.
 */

public class DataGraphCloser extends GraphCloser {
	
	/**
	 * internal constraint to close on all colors (but 0)
	 * for speed up in the particular case where all deltas are the same
	 */
	private class AllColorsConstraint extends DeltaClosure.Constraint {
		public boolean accept(Edge e) {
			// accept edge (u,v) if both vertices are of the same color
			int color1 = RColorSet.getColors(e.v1()).singleColor();
			int color2 = RColorSet.getColors(e.v2()).singleColor();
			return (color1 == color2);
		}
		// note: no validate() required since both vertices have the same color
		public void register(Edge e) {
			// color edge with same color as vertices
			int color = RColorSet.getColors(e.v1()).singleColor();
			RColorSet.getColors(e).setColor(color);
		}
	}

	/**
	 * internal constraint to close on single vertex color
	 */
	private class SingleColorConstraint extends DeltaClosure.Constraint {
		int color_;
		public SingleColorConstraint(int color) {
			color_ = color;
		}
		public boolean accept(Edge e) {
			int color;
			// accept edge (u,v) if both vertices are of the specified color
			color = RColorSet.getColors(e.v1()).singleColor();
			if (color != color_)
				return false;
			color = RColorSet.getColors(e.v2()).singleColor();
			return (color == color_);
		}
		// note: no validate() required since both vertices have the same color
		public void register(Edge e) {
			// color edge with same color as vertices
			RColorSet.getColors(e).setColor(color_);
		}
	}

	/**
	 * API
	 */
	
	/**
	 * close on all colors (except color 0) with
	 * deltas given in delta list. To avoid closing
	 * on a color, use a delta value of 0 for this color.
	 * @param delta list. List size
	 * should be >= maxcolor in the datagraph or 1.
	 * In the latter case, the same delta value is used for
	 * all colors.
	 */
	public DataGraphCloser(List<Integer> delta) {
		delta_ = initDelta_(delta);
	}

	/**
	 * implement our run() operation
	 */
	//@Implement
	public void run(Graph graph) {

		int maxColor = RColorSet.getColors(graph).maxColor();
		
		List<Integer> deltaList = makeDelta_(delta_, maxColor);
		
		if (sameDelta_(deltaList)) {
			// use same delta for all colors -> speedup
			int delta = deltaList.get(0) + 1; // +1 for standard interpretation
			if (delta <= 1) return;
			DeltaClosure.Constraint constraint = new AllColorsConstraint();
			DeltaClosure closure = new DeltaClosure(graph, delta, constraint);
			closure.run();
		}
		
		else {
			// use different delta for each color
			for (int color : RColorSet.getColors(graph)) {
				if (color == 0) continue;
				int delta = deltaList.get(color-1) + 1; // +1 for standard interpretation
				if (delta <= 1) continue;
				DeltaClosure.Constraint constraint = new SingleColorConstraint(color);
				DeltaClosure closure = new DeltaClosure(graph, delta, constraint);
				closure.run();
			}
		}
	}
}
