package helix.graph.algo.c3part.preprocessor;

/**
 * a class implementing a datagraph preprocessor.
 * call by <code>run(graph)</code><br> or <code>run(graph, logger)</code><br>
 * note: the graph is modified<br>
 * typical subclasses are GraphCleaner's and GraphDeltaCloser's
 */
import java.util.logging.Logger;

import helix.graph.model.Graph;
import helix.graph.util.LogUtil;

public abstract class GraphPreprocessor {
	
	protected Logger logger_ = null;
	
	protected String header_ = "[" + this.getClass().getSimpleName() + "] ";

	protected void log(String msg) {
		if (logger_ != null) 
			logger_.info(header_ + " " + msg);
	}
	
	/**
	 * preprocess datagraph
	 * note: the datagraph argument is modified
	 */
	public abstract void run(Graph graph);
	
	/**
	 * same but with logging information
	 */
	public void run(Graph graph, Logger logger) {
		
		if (! LogUtil.isVerbose(logger)) {
			this.run(graph);
			return;
		}
		
		logger_ = logger;

		int n0 = graph.nbVertices();
		int e0 = graph.nbEdges();

		this.run(graph);

		int n1 = graph.nbVertices();
		int e1 = graph.nbEdges();

		this.logParams();
		
		if (n1 == n0) {
			log("nb vertices : " + n0 + " unchanged");
		}
		else {
			String info = (n1 < n0) ? " removed " : " added ";
			int change = Math.abs(n0-n1);
			float perc = (float) (change * 100) / (float) (n0+(n0==0?1:0));
			log("nb vertices : " + n0 + " -> " + n1 + info + change + " (" +  String.format("%.1f", perc) + " %)");
		}

		if (e1 == e0) {
			log("nb edges : " + e0 + " unchanged");
		}
		else {
			String info = (e1 < e0) ? " removed " : " added ";
			int change = Math.abs(e0-e1);
			float perc = (float) (change * 100) / (float) (e0+(e0==0?1:0));
			log("nb edges : " + e0 + " -> " + e1 + info + change + " (" +  String.format("%.1f", perc) + " %)");
		}
	}
	
	/**
	 * default log for preprocessor parameters
	 */
	public void logParams() {
		// nothing to log
	}
}
