package helix.graph.algo.c3part.preprocessor;

import helix.graph.algo.DeltaClosure;
import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.impl.RColorSet;

import java.util.List;

/**
 * a class to compute delta closure of multigraph.<br>
 * this is used in conjunction with a Chromatizer to implements delta shuffling<br>
 * in a multigraph only edges are assumed to be colored (not necessarily vertices)
 * and polychromatic.<br>
 * note: color 0 is assumed to be unused (epsilon relation)<br>
 * For later retrieval, the colors of newly colored edges are 
 * shifted by a specified amount. By default this shift amount equals
 * the <code>maxcolor</code> parameter given in constructor and should therefore
 * equal the maximal color in the source <b>datagraph</b> i.e. the
 * number of colors in the source <b>datagraph</b> minus one.
 */

public class MultiGraphCloser extends GraphCloser {
	
	protected int shift_ = 0;

	/**
	 * internal constraint to close on single color
	 */
	private class SingleColorConstraint extends DeltaClosure.Constraint {
		int color_;
		int shift_;
		RColorSet colors_ = new RColorSet(); // just a speed up (not part of algo)
		
		public SingleColorConstraint(int color, int shift) {
			color_ = color;
			shift_ = shift;
		}
		public boolean accept(Edge e) {
			// accept edge if color is set in edge
			// [AV] TODO OTFQuorum : voir ce qui se passe avec les stars...
			// [AV] -> ca semble ok comme ca : ne pas fermer si star
			return RColorSet.getColors(e).hasColor(color_);
		}
		// no validate() required
		public void register(Edge e) {
			// color (new) edge with shifted color
			RColorSet.getColors(e).setColor(color_ + shift_);
		}
	}

	/**
	 * API
	 */

	/**
	 * constructor
	 * close on all colors (except color 0) with
	 * deltas given in delta list. To avoid closing
	 * on a color, use a delta value of 0 for this color.
	 * @param maxColor maximum color to use 
	 * normally this should equal the number of colors in the source <b>datagraph</b> minus 1
	 * (because color 0 is unused)
	 * @param delta delta list. List size should be >= maxColor or 1.
	 * In the latter case, the same delta is assumed for all colors.
	 * @param shift specified shift. Usually this should be <code>maxColor</code> or 0
	 */
	public MultiGraphCloser(List<Integer> delta, int maxColor, int shift) {
		delta_ = makeDelta_(initDelta_(delta), maxColor);
		shift_ = shift;
	}
	/**
	 * special constructor with shift = maxColor
	 */
	public MultiGraphCloser(List<Integer> delta, int maxColor) {
		this(delta, maxColor, maxColor);
	}
	
	/**
	 * implement our run() operation
	 */
	//@Implement
	public void run(Graph graph) {

		// note: the maxColor of the multigraph is <= maxColor of the datagraph
		int maxColor = RColorSet.getColors(graph).maxColor();

		if (maxColor > delta_.size())
			throw new PreprocessorException("multigraph has to many colors " + maxColor + " (expected max " + delta_.size() + ")");
		
		for (int color : RColorSet.getColors(graph)) {
			if (color == 0) continue;
			int delta = delta_.get(color-1) + 1; // +1 for standard interpretation
			if (delta <= 1) continue;
			DeltaClosure.Constraint constraint = new SingleColorConstraint(color, shift_);
			DeltaClosure closure = new DeltaClosure(graph, delta, constraint);
			closure.run();
		}
	}
}
