package helix.graph.algo.c3part.preprocessor;

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.visitor.CCByColorsVisitor;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertex;

/**
 * Experimental section
 */
public class TandemCleaner extends DatagraphCleaner {

	protected final String FatherFlag = "_TandemAliasFather_";

	/**
	 * costructor parameters
	 */
	protected int minSize_;
	protected float minFraction_;

	/*
	 * internal utilities
	 */
	protected final void setFather_(Vertex v, Vertex father) {
		v.set(FatherFlag, new RVertex(father));
	}

	protected final Vertex getFather_(Vertex v) {
		return v.has(FatherFlag) ? ((RVertex) v.get(FatherFlag)).get() : null;
	}

	protected final Vertex getAncestor_(Vertex v) {
		Vertex father = v;
		while (father != null) {
			v = father;
			father = getFather_(v);
		}
		return v;
	}

	protected final void getEpsilon_(Vertex v, GSet<Vertex> epsilon) {
		epsilon.clear();
		int vcolor = RColorSet.getColors(v).singleColor();
		for (Vertex u : v.neighbors()) {
			int ucolor = RColorSet.getColors(u).singleColor();
			if (vcolor != ucolor)
				epsilon.add(u);
		}
	}

	/**
	 * reconnect non-epsilon neighbors of a removable vertex v to its father
	 * taking care of alias elements
	 */ 
	protected final void reconnectNeighbors_(Graph graph, Vertex v) {
		Vertex father = getFather_(v);
		int vcolor = RColorSet.getColors(v).singleColor();
		for (Vertex u : v.neighbors()) {
			int ucolor = RColorSet.getColors(u).singleColor();
			if (ucolor != vcolor)
				continue;  // not epsilon 
			Vertex ancestor = getAncestor_(u);
			if (ancestor == father)
				continue;
			if (graph.hasEdge(father, ancestor))
				continue;
			Edge e = graph.addEdge(father, ancestor);
			RColorSet.getColors(e).setColor(ucolor);
		}
	}
	
	/**
	 * compute intersection size of two sets
	 */
	protected final int interSize_(GSet<Vertex> a, GSet<Vertex> b) {
		boolean amin = (a.size() < b.size());
		GSet<Vertex> min = amin ? a : b;
		GSet<Vertex> max = amin ? b : a;
		int count = 0;
		for (Vertex v : min) {
			if (max.contains(v))
				count++;
		}
		return count;
	}
	
	/**
	 * constraint on epsilon sets
	 */
	protected boolean EpsilonSetConstraint_(GSet<Vertex> vepsilon, GSet<Vertex> uepsilon) {
		
		// prop -1-
		// return vepsilon.equals(uepsilon);

		// prop -2-
		//return vepsilon.containsAll(uepsilon);
		
		// prop -3-
		// TODO use parameters not constant values
		if (uepsilon.size() > vepsilon.size())
			return false; // u should be smaller
		
		if (uepsilon.size() < minSize_)
			return false;
		
		float frac = (float) interSize_(vepsilon, uepsilon) / (float) uepsilon.size();
		
		return (frac >= minFraction_);
	}
	
	
	
	protected final void cleanupCC_(GSet<Vertex> cc, GSet<Vertex> removed,
							  GSet<Vertex> vepsilon, GSet<Vertex> uepsilon) {
		for (Vertex v : cc) {
			if (removed.contains(v))
				continue;
			int vcolor = RColorSet.getColors(v).singleColor();
			getEpsilon_(v, vepsilon);
			for (Vertex u : v.neighbors()) {
				int ucolor = RColorSet.getColors(u).singleColor();
				if (vcolor != ucolor)
					continue;
				if (removed.contains(u))
					continue;
				getEpsilon_(u, uepsilon);
				if (EpsilonSetConstraint_(vepsilon, uepsilon)) {
					String vlabel = v.get(Registry.Label).toString();
					String ulabel = u.get(Registry.Label).toString();
					v.set(Registry.Label, vlabel + "//" + ulabel);
					removed.add(u);
					u.set(FatherFlag, new RVertex(v));
				}
				
			}
		}
	}
	
	/*
	 * API
	 */
	
	/**
	 * constructors
	 */
	public TandemCleaner(int minSize, float minFraction) {
		minSize_ = minSize;
		minFraction_ = minFraction;
	}

	public TandemCleaner(int minSize, int percentFraction) {
		this(minSize, ((float) percentFraction) / 100F);
	}
	
	/**
	 * run cleaner
	 */
	public void run(Graph graph) {
		
		// -1- compute CC by colors (each cc is a primary graph)
		CCVisitor visitor = new CCByColorsVisitor(graph);
		Iterable<GSet<Vertex>> ccs = visitor.traverse().getComponents();
		
		// -2- get tandem vertices
		GSet<Vertex> removed = graph.vertexSetFactory().make();
		GSet<Vertex> vepsilon = graph.vertexSetFactory().make();
		GSet<Vertex> uepsilon = graph.vertexSetFactory().make();
		for (GSet<Vertex> cc : ccs) {
			cleanupCC_(cc, removed, vepsilon, uepsilon);
		}

		// -3- reconnect fathers
		for (Vertex v : removed) {
			reconnectNeighbors_(graph, v);
		}

		// -4- cleanup vertices
		for (Vertex v : removed) {
			graph.removeVertex(v);
		}
	}
	
	/**
	 * override logParams
	 */
	@Override
	public void logParams() {
		super.logParams();
		log("minsize  : " + minSize_);
		log("minfraction : " + minFraction_);
	}

}
