package helix.graph.algo.c3part.reporter;

import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

/**
 * a static class to check datagraph integrity and report inconsistencies
 */

public class GraphChecker extends BaseReporter {
	
	/**
	 * checking levels
	 */
	public static enum Level {
		None, Normal, Full;
	}
	
	/**
	 * check color integrity of a graph element
	 */
	private static boolean checkColor_(Registry r, int minColor, int maxColor) {
		int color = getColor_(r);
		return ((color >= minColor) && (color <= maxColor));
	}

	/**
	 * check color integrity of vertex
	 */
	private static boolean checkColor_(Vertex v, int maxColor) {
		boolean ok = checkColor_(v, 1, maxColor);
		if (! ok) {
			severe("invalid color for vertex " + v + " (" + getColors_(v) + ")");
		}
		return ok;
	}

	/**
	 * check color integrity of edge
	 */
	private static boolean checkColor_(Edge e, int maxColor) {
		boolean ok = checkColor_(e, 0, maxColor);
		if (! ok) {
			warning("invalid color for edge " + e + " (" + getColors_(e) + ")");
			return false;
		}
		// more check for edges
		int eColor = getColors_(e).singleColor();
		int v1Color = getColors_(e.v1()).singleColor();
		int v2Color = getColors_(e.v2()).singleColor();
		if (eColor == 0) { // epsilon link
			ok = (v1Color != v2Color);
		}
		else {             // internal link
			ok = (v1Color == v2Color);
		}
		if (! ok) {
			warning("invalid end colors for edge " + e  + " (" + getColors_(e) + ") "
					+  e.v1()  + " (" + v1Color + ") "
					+  e.v2()  + " (" + v2Color + ")");
		}
		return ok;
	}
	
	/**
	 * check color integrity of datagraph
	 */
	protected static boolean checkColors(Graph g, Level level) {
		if (level == Level.None)
			return true;
		int nerr = 0;
		int maxColor = RColorSet.nbColors(g) - 1;
		for (Vertex v : g.vertices()) {
			nerr += (checkColor_(v, maxColor) ? 0 : 1);
		}
		if (nerr != 0)
			return false;
		//
		// AV : we actually never use the edge color of the datagraph
		//      so, the following will just raise warnings
		for (Vertex v : g.vertices()) {
			for (Edge e : v.edges()) {
				if (RColorSet.hasColors(e)) { // user specified colors, let's check them
					nerr += (checkColor_(e, maxColor) ? 0 : 1);
				}
			}
		}
		return true;
	}
	
	/*
	 * API
	 */
	
	/**
	 * full check
	 */
	public static boolean checkGraph(Graph g, Level level) {
		return checkColors(g, level);
	}
	
	public static boolean checkGraph(Graph g) {
		return checkGraph(g, Level.Normal);
	}
	
}
