package helix.graph.algo.c3part.reporter;

/**
 * report statistics information about datagraph 
 */

import helix.graph.algo.CCVisitor;
import helix.graph.algo.c3part.visitor.CCByColorsVisitor;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

public class GraphStatistics extends BaseReporter {

	/**
	 * count the number of epsilon links between each datasubgraph
	 * return the matrix of occurences (diagonal contains the
	 * number of internal links) 
	 */
	public static final int[][] epsilonMatrix(Graph g) {
		int nbColor = RColorSet.nbColors(g);
		int matrix[][] = new int[nbColor][nbColor];
		
		for (Vertex v : g.vertices()) {
			int vcol = getColor_(v);
			if (vcol < 0)
				continue;
			for (Vertex u : v.neighbors()) {
				int ucol = getColor_(u);
				if (ucol < 0)
					continue;
				matrix[vcol][ucol]++;
				matrix[ucol][vcol]++;
			}
		}
		for (int i = 0 ; i < nbColor ; i++) {
			for (int j = 0 ; j < nbColor ; j++)
				matrix[i][j] /= 2;
		}
		return matrix;
	}

	/**
	 * compute number of vertices per color
	 */

	public static final int[] verticesPerColor(Graph g) {

		int row[] = new int[RColorSet.nbColors(g)];
		for (Vertex v : g.vertices()) {
			  int color = getColor_(v);
			  if (color >= 0)
				  row[color]++;
		}
		return row;
	}
	
	/**
	 * compute number of connected components per color
	 */

	public static final int[] componentsPerColor(Graph g) {
		int row[] = new int[RColorSet.nbColors(g)];

		CCVisitor visitor = new CCByColorsVisitor(g);
		Iterable<GSet<Vertex>> ccs = visitor.traverse().getComponents();
		
		for (GSet<Vertex> cc : ccs) {
			Vertex v = cc.iterator().next();
			int color = getColor_(v);
			if (color >= 0)
				row[color]++;
		}
		return row;
	}

	/*
	 * API
	 */
	
	/**
	 * report graph statistics
	 */
	public static void logStatistics(Graph g) {

		if (! isVerbose())
			return;

		log("--------------------------------- Graph statistics");
		
		int nbColors = RColorSet.nbColors(g);

		log("nbVertices: " + g.nbVertices());
		log("nbEdges:    " + g.nbEdges());
		log("nbColors:   " + nbColors);

		StringBuffer buf = new StringBuffer();
		
		int row[] = verticesPerColor(g);

		buf.append("nbVerticesPerColor: ");
		for (int i = 1 ; i < nbColors ; i++) {
			buf.append(String.format("%6d", row[i]) + " ");
		}
		log(buf.toString());

		buf.setLength(0);
		row = componentsPerColor(g);

		buf.append("nbCCPerColor:       ");
		for (int i = 1 ; i < nbColors ; i++) {
			buf.append(String.format("%6d", row[i]) + " ");
		}
		log(buf.toString());

		int matrix[][] = epsilonMatrix(g);

		log("epsilonMatrix:");
		for (int i = 1 ; i < nbColors ; i++) {
			buf.setLength(0);
			buf.append("epsmat " + i + " : ");
			for (int j = 1 ; j < nbColors ; j++) {
				buf.append(String.format("%6d", matrix[i][j]) + " ");
			}
			log(buf.toString());
		}
	}
	
}
