package helix.graph.io;

import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.IllegalValueException;
import helix.graph.model.Registry;
import helix.graph.model.RegistryValue;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RBoolean;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RFloat;
import helix.graph.model.impl.RInteger;
import helix.graph.model.impl.RString;

import java.io.IOException;
import java.io.Reader;
import java.util.Hashtable;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *  a Dimacs format reader with additional capabilities
 *  to read attributes attached to Graph, Vertices or Edges.<br>
 *  Attributes are specified as :<br>
 *  <code>
 *  key=value
 *  with key = alpha+
 *       value = any+ | "qany+"
 *       
 *       alpha = [a-z] | [A-Z] | [0-9] | '_'
 *       any = [^ |"]
 *       qany = [^"]
 *  </code>
 * note: Please note that the '-' (dash) character is not allowed in key and is only
 * allowed in values if it is quoted.<br>
 * note: You cannot backquote the '"' in values... sorry<br>
 * <br>
 * examples:
 *  <code>
 * -- for graph
 * p Label=MyGraph
 * -- for Vertex
 * n Label=one Colors=1011 Genes="g1 g2 g3 g4"
 * -- for Edges
 * e 1 4 Label="one to four" Colors=1011
 * </code>
 * <br>
 * 
 * When found, attributes are registered in Vertex or Edge Registry
 *  with the corresponding key. The reader try to choose the most appropriate
 *  type for RegistryValue (i.e. will use RInteger for value 123) but this
 *  has no special effect since all ImmutableRegistryValue's are interconvertible
 *  (including for the <code>equals()</code> comparison).<br>
 *  
 *  in addition, for the following <b>special keys</b>, the
 *  value is further transformed :
 *  <pre><code>
 *  key          RegistryValue type
 *  ----------------------------------
 *  Colors       ColorSet
 *  Color        integer
 *  Label        String
 *  Comment      String
 *  
 *  // for genes
 *  Rank         Integer
 *  Orientation  String   (converted to "+" or "-")
 *  
 *  </code></pre>
 *  note: the special keys are recognized case-insensitive.<br>
 *  note: the IsDirect key is readOnly and is a shortcut for <code>Orientation == "+"</code><br>
 *  
 */

public class DimacsExtendedReader extends DimacsReader {

	/**
	 * visiting flag
	 */
	private static final String visitedFlag = "_dimacs_visited_";
	
	/**
	 * the key=value pattern
	 */
	protected static Pattern labelPattern_ = Pattern.compile("(\\w+)=(?:([^\"|\\s]+)|\"((?:[^\"])+)\")\\s*");

	/*
	 * internal utilities
	 */
	private static final Integer getInteger_(String s) {
		try {
			return Integer.decode(s);
		} catch (NumberFormatException nfe) {}
		return null;
	}

	private static final Float getFloat_(String s) {
		try {
			return Float.valueOf(s);
		} catch (NumberFormatException nfe) {}
		return null;
	}

    private static final String getOrientation_(String s) {
    	if (s.equalsIgnoreCase("D") || s.equals("+"))
    		return "+";
    	return "-";
    }
	
	/**
	 * parse the raw label string and put (key,String:value) pairs
	 * into HashTable
	 */
	protected static Hashtable<String, String> parseLabelString_(String str) 
	  throws GraphIOException {
		Hashtable<String, String> result = new Hashtable<String, String>();
		Matcher m = labelPattern_.matcher(str.trim());
		while (m.find()) {
		  if (m.groupCount() < 3)
			  throw new GraphIOException("bad extended label: " + str);
		  String key = m.group(1);
		  String value = (m.group(2)) != null ? m.group(2) : m.group(3);
		  result.put(key, value);
		}
		return result;
	}

	/**
	 * convert (key,String:value) pairs to (key, RegistryValue) pairs
	 * also takes care of special keys
	 */
	protected static Hashtable<String, RegistryValue> convertToRegistry_(Hashtable<String, String> table)
	  throws IllegalValueException {

		Hashtable<String, RegistryValue> result = new Hashtable<String, RegistryValue>();
		
		for (String key : table.keySet()) {
			
			String value = table.get(key);
			//
			// first try special keys
			//
			if (key.equalsIgnoreCase(Registry.Label)) {
				result.put(Registry.Label, new RString(value));
				continue;
			}
			if (key.equalsIgnoreCase(Registry.Colors)) {
				result.put(Registry.Colors, new RColorSet(value));
				continue;
			}
			if (key.equalsIgnoreCase(Registry.Color)) {
				RColorSet colors = new RColorSet();
				colors.setColor(getInteger_(value));
				result.put(Registry.Colors, colors);
				continue;
			}
			if (key.equalsIgnoreCase(Registry.Comment)) {
				result.put(Registry.Comment, new RString(value));
				continue;
			}
			if (key.equalsIgnoreCase(Registry.Rank)) {
				result.put(Registry.Rank, new RInteger(value));
				continue;
			}
			if (key.equalsIgnoreCase(Registry.Orientation)) {
				String orient = getOrientation_(value);
				result.put(Registry.Orientation, new RString(orient));
				result.put(Registry.IsDirect, new RBoolean(orient.equals("+")));
				continue;
			}
			//
			// use key as is
			// just try to find out the best representation
			//
			Integer ival = getInteger_(value);
			if (ival != null) {
				result.put(key, new RInteger(ival));
				continue;
			}
			Float fval = getFloat_(value);
			if (fval != null) {
				result.put(key, new RFloat(fval));
				continue;
			}
			//
			// tant pis...
			//
			result.put(key, new RString(value));
		}
		return result;
	}

	/**
	 * set attributes according to label
	 */
	protected static void setAttributes_(Registry elt)
	  throws IllegalValueException, GraphIOException {

		if (! elt.has(Registry.Label))
			return;
		
		String label = elt.get(Registry.Label).toString();
		
		Hashtable<String, String> stringTable = parseLabelString_(label);
		Hashtable<String, RegistryValue> valueTable = convertToRegistry_(stringTable);
		
		if (stringTable.size() != 0)
			elt.remove(Registry.Label); // cleanup previous label
			
		for (String key : valueTable.keySet()) {
			elt.set(key, valueTable.get(key));
		}
	}

	/*
	 * API
	 */
	
	public DimacsExtendedReader(Reader reader, GraphFactory factory) {
		super(reader, factory);
	}

	public DimacsExtendedReader(Reader reader, GraphFactory factory, Logger logger) {
		super(reader, factory, logger);
	}

	@Override
	public Graph read() throws IOException {
		
		// read graph using DimacsReader
		
		Graph graph = super.read();

		// now parse the Registry.Label to add attributes if any

		// first to the graph itself
		
		setAttributes_(graph);
		
		// then to vertices and edges

		for (Vertex v : graph.vertices()) {
			setAttributes_(v);
			for (Edge e : v.edges()) {
				if (! e.has(visitedFlag)) { // don't parse it twice
					setAttributes_(e);
					e.set(visitedFlag);
				}
			}
		}

		// remove the temporary visitedFlag
		for (Vertex v : graph.vertices()) {
			for (Edge e : v.edges()) {
				e.remove(visitedFlag);
			}
		}
		
		//
		// deal with colored graphs
		// set the colors of graph registry according to
		// the colors of vertices and edges
		//
		
		RColorSet graphColors = new RColorSet();

		for (Vertex v : graph.vertices()) {
			if (RColorSet.hasColors(v))
				graphColors.orColors(RColorSet.getColors(v));
			for (Edge e : v.edges()) {
				if (RColorSet.hasColors(e))
					graphColors.orColors(RColorSet.getColors(e));
			}
		}

		//
		// fill missing colors
		//
		for (int i = 0 ; i < graphColors.length() ; i++)
			graphColors.setColor(i);
		
		graph.set(Registry.Colors, graphColors);
		
		return graph;
	}

}
