package helix.graph.io;

import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;

import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;

/**
 * Extended Dimacs format :<br>
 * <code>
 * p [label]                 :  problem line : label is from graph registry Label<br>
 * n [label] +               :  vertex lines : label is from vertex registry Label<br>
 * e index1 index2 [label]   :  edge line    : indexi are 1-based index refereing to<br>
 *                                             vertices (see below). 
 *                                             label is from edge registry Label<br>
 * </code>
 * <br>
 */
public class DimacsExtendedWriter extends AbstractGraphWriter {

	/**
	 * internal class to store standard keys, output label and default values
	 */
	protected class Entry {
		String key, label, value;
		public Entry(String key, String label, String value) {
			this.key = key;
			this.label = label;
			this.value = value;
		}
		public Entry(String key, String label) {
			this(key, label, null);
		}
	}
	
	protected static final String EOL = System.getProperty("line.separator");
	
	protected static final String Index = "_DimacsWriterIndex_";

	protected final Entry[] StandardVertexKeys_ = {
							new Entry(Registry.Label, "label"),
							new Entry(Registry.Rank, "rank"),
							new Entry(Registry.Colors, "colors"),
							new Entry(Registry.Orientation, "orientation"),
							new Entry(Registry.Comment, "comment") };

	protected final Entry[] StandardEdgeKeys_ = {
							new Entry(Registry.Label, "label"),
							new Entry(Registry.Colors, "colors"),
							new Entry(Registry.Comment, "comment") };

	protected HashSet<String> vertexKeysSet_;
	protected HashSet<String> edgeKeysSet_;
	
	/**
	 */
	protected HashSet<String> recordKeys_(Entry[] entries) {
		HashSet<String> set = new HashSet<String>();
		for (Entry entry : entries) {
			set.add(entry.key);
		}
		return set;
	}

	/**
	 */
	protected void writeRegistry_(Registry registry, String key, String label, String defaut) 
	    throws IOException {
		boolean quote = false;
		String value= registry.has(key) ? registry.get(key).toString() : defaut;
		if (value == null)
			return;
		if (label != null) {
			writer_.write(label + "=");
			quote = (value.indexOf(' ') >= 0);
		}
		writer_.write((quote ? "\"" : "") + value + (quote ? "\"" : "") + " ");
	}
	
	/**
	 */
	protected void writeStandardRegistry_(Registry registry, Entry[] standardEntries)
    	throws IOException {
		for (Entry entry : standardEntries) {
			writeRegistry_(registry, entry.key, entry.label, entry.value);
		}
	}

	/**
	 * 
	 */
	protected void writeUserRegistry_(Registry registry, HashSet<String> exclude)
	throws IOException {
		for (String key : registry.keys()) {
			if (exclude.contains(key))
				continue;
			if (key.startsWith("_"))
				continue;
			writeRegistry_(registry, key, key, null);
		}
	}

	/**
	 */
	protected void writeKey_(String key)
	    throws IOException {
		writer_.write(key + " ");
	}

	/**
	 */
	protected void writeEol_()
    	throws IOException {
		writer_.write(EOL);
	}
	
	/*
	 * API
	 */
	/**
	 * constructor
	 */
	public DimacsExtendedWriter(Writer writer) {
		super(writer);
		vertexKeysSet_ = recordKeys_(StandardVertexKeys_);
		edgeKeysSet_   = recordKeys_(StandardEdgeKeys_);
	}

	/**
	 * writer
	 */
	//@Implement
	public void write(Graph graph) throws IOException {
		
		// p line
		writeKey_("p");
		writeRegistry_(graph, Registry.Label, null, "grf");
		writeEol_();

		// n lines
		int vertexNum = 0;
		for (Vertex v : graph.vertices()) {
			writeKey_("n");
			writeStandardRegistry_(v, StandardVertexKeys_);
			writeUserRegistry_(v, vertexKeysSet_);
			writeEol_();
			v.set(Index, ++vertexNum);
		}
		
		// e lines
		for (Vertex v : graph.vertices()) {
			for (Edge e : v.edges()) {
				if (v.equals(e.v1())) continue; // write edge only once
				int i1 = e.v1().get(Index).toInt();
				int i2 = e.v2().get(Index).toInt();
				writeKey_("e");
				writer_.write(i1 + " " + i2 + " ");
				writeStandardRegistry_(e, StandardEdgeKeys_);
				writeUserRegistry_(e, edgeKeysSet_);
				writeEol_();
			}
		}
		
		
		// cleanup vertices
		for (Vertex v : graph.vertices()) {
			v.remove(Index);
		}
		
	}

}
