package helix.graph.io;

import helix.graph.model.GSet;
import helix.graph.model.Registry;
import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.Vertex;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Extended Dimacs format :<br>
 * <code>
 * [c [comment]]               :  comment line : may appear anywhere<br>
 * [p [label]]                 :  problem line : label is kept in graph registry as Label<br>
 * [n [label]]+                :  vertex line  : label is kept in vertex registry as Label<br>
 * e index1 index2 [label]     :  edge line    : indexi are 1-based index refereing to<br>
 *                                               vertices (see below). label is kept in edge registry as Label<br>
 * </code>
 * <br>
 * 
 * note: the 'n' lines are optional (if you don't need to store anything in vertices)
 * Vertices will be created as needed when reading the 'e index1 index2' lines.<br>
 * However, for efficiency, it is advisable to put indices in decreasing order
 * (i.e. higher values coming first).<br>
 * If the 'n' lines are present, then the indices in 'e' lines refer to the order where
 * vertices have been read (starting at 1 for the first vertex).<br>
 * 
 * @author viari 
 *
 */
public class DimacsReader extends AbstractGraphReader {

	protected static final String Index = "_DimacsReaderIndex_";

	/**
	 * the edge pattern
	 */
	protected static Pattern edgePattern_ = Pattern.compile("^e\\s+(\\d+)\\s+(\\d+)\\s*(.*)");

	/**
	 * default constructor
	 */
	public DimacsReader(Reader reader, GraphFactory factory) {
		super(reader, factory);
	}
	
	/**
	 * constructor with logging addition
	 */
	public DimacsReader(Reader reader, GraphFactory factory, Logger logger) {
		super(reader, factory, logger);
	}
	
	/*
	 * API
	 */

	@Override
	public Graph read() throws IOException {
		BufferedReader bufReader = new BufferedReader(reader_);
		String line;

		Graph graph = factory_.make();
		
		while ((line = bufReader.readLine()) != null) {
			line = line.trim();
			String subLine = line.substring(1).trim();
			
			// comment line
			if (line.length() == 0)
				continue;
			if (line.startsWith("c"))
				continue;
			
			// problem line
			if (line.startsWith("p")) {
				graph.set(Registry.Label, subLine);
				continue;
			}

			// optional node lines
			if (line.startsWith("n")) {
				Vertex v = graph.addVertex();
				v.set(Index, graph.nbVertices());
				v.set(Registry.Label, subLine);
				continue;
			}

			// edge lines
			if (line.startsWith("e")) {
				Matcher m = edgePattern_.matcher(line.trim());
				if (! (m.matches() && (m.groupCount() == 3))) {
					severe_("invalid edge line : " + line);
				}
				int i1 = Integer.parseInt(m.group(1));
				int i2 = Integer.parseInt(m.group(2));
				if ((i1 <= 0) || (i2 <= 0))
					severe_("bad vertex index (should be 1-based) : " + line);

				int max = (i1 > i2 ? i1 : i2);
				boolean modify = (max > graph.nbVertices());
				while (max > graph.nbVertices()) {
					Vertex v = graph.addVertex();
					v.set(Index, graph.nbVertices());
				}
				if (modify)
					graph.findVertex(GSet.RehashKey, null); // will force rehash
				
				Vertex v1 = graph.findVertex(Index, i1);
				Vertex v2 = graph.findVertex(Index, i2);
				if (v1 == v2) {
					warning_("self loop edge : " + line);
					continue;
				}
				
				if (graph.hasEdge(v1, v2)) {
					warning_("duplicate edge : " + line);
					continue;
				}
				
				Edge e = graph.addEdge(v1, v2);
				e.set(Registry.Label, m.group(3));
				continue;
			}
			
			// ???
			warning_("invalid line : " + line);
		}
		
		// remove temporary Index
		for (Vertex v : graph.vertices()) {
			v.remove(Index);
		}
		
		return graph;
	}

}
