package helix.graph.io;

import org.junit.Test;

import static org.junit.Assert.*;

import helix.graph.model.Edge;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.memory.MemGraphFactory;

import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;

public class TestDimacs {

	@Test
	public void testReadSimple() throws IOException {
		String s = "p test\ne 1 2\n e 1 3";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsReader(sreader, factory);
		Graph g = reader.read();
		assertEquals("graph size", 3, g.nbVertices());
	}

	@Test
	public void testReadWithVertices() throws IOException {
		String s = "p test\nn one\nn two\nn three\ne 1 2\n e 1 3";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsReader(sreader, factory);
		Graph g = reader.read();
		assertEquals("graph size", 3, g.nbVertices());
		Vertex v1 = g.findVertex(Registry.Label, "one");
		assertTrue("vertex one", (v1 != null));
		Vertex v2 = g.findVertex(Registry.Label, "two");
		assertTrue("vertex one", (v2 != null));
		Vertex v3 = g.findVertex(Registry.Label, "three");
		assertTrue("vertex one", (v3 != null));
		assertTrue("edge 1 2", g.hasEdge(v1, v2));
		assertTrue("edge 1 3", g.hasEdge(v1, v3));
		assertFalse("edge 2 3", g.hasEdge(v2, v3));
	}

	@Test
	public void testReadWithMoreVertices() throws IOException {
		String s = "p test\nn one\nn two\nn three\ne 1 2\n e 1 4";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsReader(sreader, factory);
		Graph g = reader.read();
		assertEquals("graph size", 4, g.nbVertices());
	}

	@Test
	public void testReadExtended() throws IOException {
		String s = "p test\nn label=one colors=10\nn label=two colors=1\nn label=\"third element\" \ne 1 2 label=\"one-to-two\" colors=10\n e 1 3 colors=1";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(sreader, factory);
		Graph g = reader.read();
		assertEquals("graph size", 3, g.nbVertices());

		assertEquals("graph nbcolors", 2, RColorSet.nbColors(g));

		Vertex v1 = g.findVertex(Registry.Label, "one");
		assertFalse("v1 color 0", RColorSet.hasColor(v1, 0));
		assertTrue("v1 color 1", RColorSet.hasColor(v1, 1));
		Vertex v2 = g.findVertex(Registry.Label, "two");
		assertTrue("v2 color 1", RColorSet.hasColor(v2, 0));
		assertFalse("v2 color 0", RColorSet.hasColor(v2, 1));
		Vertex v3 = g.findVertex(Registry.Label, "third element");
		assertFalse("v3 color 0", RColorSet.hasColor(v3, 1));
		assertFalse("v3 color 1", RColorSet.hasColor(v3, 1));
		
		Edge e = g.getEdge(v1, v2);
		assertFalse("v1-v2 color 0", RColorSet.hasColor(e, 0));
		assertTrue("v1-v2 color 1", RColorSet.hasColor(e, 1));
		assertEquals("v1-v2 label", "one-to-two", e.get(Registry.Label).toString());
	}
	
	@Test
	public void testReadExtendedComment() throws IOException {
		String s = "p test\nn label=one.1 comment=\"noeud numero un\"\nn label=\"two.2 second\" comment=\"deux\"\ne 1 2 label=\"one-to-two\" colors=10 comment=\"edge\"\n";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(sreader, factory);
		Graph g = reader.read();
		assertEquals("graph size", 2, g.nbVertices());
		assertEquals("graph nbcolors", 2, RColorSet.nbColors(g));

		Vertex v1 = g.findVertex(Registry.Label, "one.1");
		assertEquals("v1 comment", "noeud numero un", v1.get(Registry.Comment).toString());
		Vertex v2 = g.findVertex(Registry.Comment, "deux");
		assertEquals("v2 label", "two.2 second", v2.get(Registry.Label).toString());
		assertTrue("edge 1 2", g.hasEdge(v1, v2));
	}

	@Test
	public void testReadExtendedAny() throws IOException {
		String s = "p test\nn label=one.1 rank=1\nn label=\"two 2\" position=2";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(sreader, factory);
		Graph g = reader.read();

		Vertex v1 = g.findVertex(Registry.Label, "one.1");
		assertEquals("v1 rank", 1, v1.get(Registry.Rank).toInt());
		Vertex v2 = g.findVertex(Registry.Label, "two 2");
		assertEquals("v2 rank", 2, v2.get("position").toInt());
	}
	
	@Test
	public void testReadFile() throws IOException {

		FileReader freader = new FileReader("samples/sample.poly");
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(freader, factory);
		Graph g = reader.read();
		
		assertEquals("graph size", 1170, g.nbVertices());
		Vertex v1 = g.findVertex(Registry.Label, "MG013.folD");
		assertTrue("v1 color", RColorSet.hasColor(v1, 1));
		assertFalse("v1 color", RColorSet.hasColor(v1, 2));
		assertEquals("v1 orientation", "-", v1.get(Registry.Orientation).toString());
		assertTrue("v1 direction", ! v1.get(Registry.IsDirect).toBoolean());
	}
	
	@Test
	public void testWriteSimple() throws IOException {
		String s = "p test\nn label=one colors=10 myslot=3.14\nn label=\"second vertex\" colors=1\nn colors=111 \ne 1 2 label=\"one-to-two\" colors=10\n e 1 3 colors=1";
		StringReader sreader = new StringReader(s);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(sreader, factory);
		Graph g = reader.read();
		StringWriter swriter = new StringWriter();
		GraphWriter writer = new DimacsExtendedWriter(swriter);
		writer.write(g);
		String out = swriter.getBuffer().toString();
		System.out.println(out);
		sreader = new StringReader(out);
		reader = new DimacsExtendedReader(sreader, factory);
		g = reader.read();
		assertEquals("graph size", 3, g.nbVertices());
	}
	
	

	// AV: TODO exception throw on bad labels
}
