package helix.graph.model.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import helix.graph.model.DuplicateValueException;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.NoSuchVertexException;
import helix.graph.model.RegistryValue;
import helix.graph.model.SameVertexException;
import helix.graph.model.Vertex;

/*
 * Abstract Graph implementation
 * provides basic operations in order to minimize your implementation
 */

public abstract class AbstractGraph extends DelegateRegistry implements Graph {

	protected GSet<Vertex> vertices_;

	protected Vertex checkVertex(Vertex v)  throws NoSuchVertexException {
		if (! vertices_.contains(v))
			throw new NoSuchVertexException(v.toString());
		return v;
	}
	
	/*
	 * default Graph API
	 */
	
	public AbstractGraph() {
		vertices_ = this.vertexSetFactory().make();
	}
	
	public int nbVertices() {
		return vertices_.size();
	}

	/**
	 * this implementation is O(n).
	 * you may override it if you can provide a better way
	 * to do this
	 */
	public int nbEdges() {
		int n = 0;
		for (Vertex v : vertices_) {
			n += v.degree();
		}
		return (n / 2);
	}
	
	public Iterable<Vertex> vertices() {
		return vertices_;
	}

	public boolean hasVertex(Vertex v) {
		return vertices_.contains(v);
	}
	
	public void disconnectVertex(Vertex v) throws NoSuchVertexException {
		checkVertex(v);
		// we need to copy the edges to remove in order to avoid
		// concurrent access
		for (Edge e : this.edgeSetFactory().make(v.edges())) {
			removeEdge(e);
		}
	}

	public void removeVertex(Vertex v) throws NoSuchVertexException {
		// disconnect edges
		disconnectVertex(v);
		// remove vertex from set
		vertices_.remove(v);
	}

	//@return null if no such edge
	public Edge getEdge(Vertex v1, Vertex v2) throws NoSuchVertexException {
		checkVertex(v1);
		checkVertex(v2);
		return v1.getEdgeTo(v2);
	}
	
	public boolean hasEdge(Vertex v1, Vertex v2) {
		return (this.getEdge(v1, v2) != null);
	}

	public Edge assertEdge(Vertex v1, Vertex v2) throws NoSuchVertexException, SameVertexException {
		Edge e = this.getEdge(v1, v2);
		return (e != null ? e : this.addEdge(v1, v2));
	}
	
	public void removeEdge(Edge e) {
		removeEdge(e.v1(), e.v2());
	}

	public Vertex findVertex(String registryKey, RegistryValue registryValue)
	  throws DuplicateValueException {
		return vertices_.findUnique(registryKey, registryValue);
	}

	public Vertex findVertex(String registryKey, Object anyValue)
	  throws DuplicateValueException {
		return vertices_.findUnique(registryKey, anyValue);
	}

	public GSet<Vertex> findVertices(String registryKey, RegistryValue registryValue) {
		return vertices_.findMultiple(registryKey, registryValue);
	}

	public GSet<Vertex> findVertices(String registryKey, Object anyValue) {
		return vertices_.findMultiple(registryKey, anyValue);
	}
	
	public void clear() {
		vertices_.clear();
	}
	
	/*
	 * experimental section
	 */
	
	/**
	 * TODO a documenter serieusement
	 */
	public void addPhysicalVertex(Vertex v) {
		vertices_.add(v);
	}
	
	/**
	 * TODO a documenter serieusement
	 * en particulier : destroy the current graph
	 * accepte un visiteur a classes recouvrantes et non totalement couvrant
	 */
	public Set<Graph> physicalSplit(Iterable<GSet<Vertex>> splits) {

		// these are temporary flags
		String colorFlag_ = "_physicalSplit_color_";
		String sonsFlag_ = "_physicalSplit_sons_";
		
		// -1- put color(s) to vertices in splits
		//
		ArrayList<GSet<Vertex>> splitList = new ArrayList<GSet<Vertex>>();
		int nbColors = 0;
		
		for (GSet<Vertex> set : splits) {
			for (Vertex v : set) {
				RColorSet.getColors(v, colorFlag_).setColor(nbColors);
			}
			splitList.add(set);
			nbColors++;
		}

		// -2- focus on uncolored nodes to remove them
		//
		GSet<Vertex> focusVertices = this.vertexSetFactory().make();
		for (Vertex v : this.vertices_) {
			if (! v.has(colorFlag_))
				focusVertices.add(v);
		}
		for (Vertex v : focusVertices) {
			this.removeVertex(v);
		}		
		
		// -3- focus on multicolor nodes to duplicate them
		//     this is the most tedious part. note there is
		//	   no multicolor node if splits forms a partition
		//	   of the graph vertices.
		//

		focusVertices.clear();
		
		// -3.1- collect multicolor vertices
		//
		for (Vertex v : this.vertices_) {
			if (RColorSet.nbColors(v, colorFlag_) > 1)
				focusVertices.add(v);
		}
		
		// -3.2- duplicate them and restore edges to vertices of same
		//
		RColorSet common = new RColorSet();
		
		for (Vertex v : focusVertices) {
			RColorSet vcolors = RColorSet.getColors(v, colorFlag_);
			RVertexArray vsons = new RVertexArray(nbColors);
			v.set(sonsFlag_, vsons);
			// duplicate in 'color' number of copies 
			for (int color : vcolors) {
				Vertex nv = this.addVertex();
				nv.copy(v);
				RColorSet.getColors(nv, colorFlag_).setColor(color);
				vsons.setVertex(color, nv);
				splitList.get(color).remove(v); // remove old vertex from proper split
				splitList.get(color).add(nv);   // add new vertex in proper split
			}
			// restore edges
			for (Edge e : v.edges()) {
				Vertex u = e.getVertexTo(v);
				RColorSet ucolors = RColorSet.getColors(u, colorFlag_);
				if ((ucolors.countColors() == 1)) { // -----  monochrome link
					int ucolor = ucolors.singleColor();
					if (vcolors.hasColor(ucolor)) {
						Edge ne = this.addEdge(u, vsons.getVertex(ucolor));
						ne.copy(e);
					}
				}
				else {						   		// -----  multicolor link
					if (! u.has(sonsFlag_))			// not yet duplicated
						continue;					// don't care, we'll do it later
					RVertexArray usons = (RVertexArray) u.get(sonsFlag_);
					common.setColors(vcolors);
					common.andColors(ucolors);
					for (int color : common) {
						Edge ne = this.addEdge(usons.getVertex(color), vsons.getVertex(color));
						ne.copy(e);
					}
				}
			}
		}
		
		// -3.3- finally remove originals multicolor nodes
		//
		for (Vertex v : focusVertices) {
			v.remove(colorFlag_);
			v.remove(sonsFlag_);
			this.removeVertex(v);
		}

		// -4- remove inter split edges
		//
		
		// -4.1- determine color of largest split
		//
		int colormax = 0;
		int maxsize = (nbColors > 0 ? splitList.get(colormax).size() : 0);
		for (int color = 1 ; color < nbColors ; color++) {
			int size = splitList.get(color).size();
			if (size > maxsize) {
				maxsize = size;
				colormax = color;
			}
		}
		
		// -4.2- loop on all splits except largest one
		//
		GSet<Edge> removedEdges = this.edgeSetFactory().make();
		for (int color = 0 ; color < nbColors ; color++) {
			if (color == colormax)
				continue;
			for (Vertex v : splitList.get(color)) {
				for (Edge e : v.edges()) {
					Vertex u = e.getVertexTo(v);
					if (RColorSet.getColors(u, colorFlag_).singleColor() != color) {
						removedEdges.add(e);
					}
				}
			}
		}
		for (Edge e : removedEdges) {
			this.removeEdge(e);
		}
		
		// -5- create subgraphs
		//
		HashSet<Graph> result = new HashSet<Graph>();
		for (GSet<Vertex> set : splits) {
			AbstractGraph g = (AbstractGraph) this.factory().make();
			g.vertices_.addAll(set);
			g.copy(this); // copy graph registry
			result.add(g);
		}

		// -6- final cleanup : remove flags and clear graph
		//
		for (Vertex v : this.vertices_) {
			v.remove(colorFlag_);
		}
		
		// clear me (physically)
		this.vertices_.clear();
		
		return result;
	}
	

}
