package helix.graph.model.impl;

import helix.graph.model.IllegalValueException;
import helix.graph.model.Registry;
import helix.graph.model.RegistryValue;
import helix.graph.util.BitSetIterator;

import java.util.BitSet;
import java.util.Iterator;

/**
 * Implements set of colors to store into Registry 
 *
 */

public class RColorSet extends AbstractRegistryValue implements Iterable<Integer> {

	/**
	 * internal BitSet to store colors
	 */
	protected BitSet bitset_;

	private BitSet tmp_ = new BitSet(); // internal for speedup

	/**
	 * convert binary String (e.g. "10001") to BitSet
	 * note: the first (lowest) bit is on the right
	 */
	protected BitSet stringToBitset_(String str) throws IllegalValueException {
		BitSet bset = new BitSet();
		int len = str.length();
		for (int i = 0 ; i < len ; i++) {
		  char c = str.charAt(len-i-1); 
		  if (c == '1')
			bset.set(i);
		  else if (c != '0')
			throw new IllegalValueException("Illegal bitset: " + str);
		}
		return bset;
	}
	
	/**
	 * convert BitSet to binary String (e.g. "10001")
	 * note: the first (lowest) bit is on the right
	 */
	protected String bitSetToString_(BitSet bset) {
		int len = bset.length();
		if (len == 0) 
			return "0";
		
		StringBuffer sb = new StringBuffer();
		for (int i = 0 ; i < len ; i++) {
		  sb.append(bset.get(len-i-1) ? "1" : "0");
		}
		return sb.toString();
	}
	
	/*
	 * Registrable API
	 */

	public RegistryValue klone() {
		return new RColorSet(this);
	}
	
	//@Override
	public boolean equals(Object o) {
		return (  (o instanceof RColorSet)
			    ? this.bitset_.equals(((RColorSet)o).bitset_)
				: false);
	}

	//@Override
	public int hashCode() {
		return this.bitset_.hashCode();
	}


	//@Override
	public String toString() {
		return bitSetToString_(this.bitset_);
	}

	/*
	 * Iterable API
	 */
	
	/**
	 * return an iterable over colors
	 */
	public Iterator<Integer> iterator() {
		return new BitSetIterator(bitset_);
	}
	
	/*
	 * RcolorSet API
	 */
	
	/**
	 * constructor : make an empty RColorSet
	 */
	public RColorSet() {
		this.bitset_ = new BitSet();
	}
	
	/**
	 * constructor : make a RColorSet by parsing the binary
	 * string argument (e.g. "10001")
	 * note: the first (lowest) color is on the right
	 */
	public RColorSet(String s) throws IllegalValueException {
		this.bitset_ = stringToBitset_(s);
	}

	/**
	 * constructor : make a RColorSet by copying another RColorSet
	 */
	public RColorSet(RColorSet cset) {
		this.bitset_ = new BitSet();
		this.bitset_.or(cset.bitset_);
	}

	/**
	 * constructor : make a RColorSet by using a specified set of integers
	 */
	public RColorSet(Iterable<Integer> colors) {
		this.bitset_ = new BitSet();
		for (int color : colors) {
			this.setColor(color);
		}
	}

	/**
	 * test if this ColorSet has the specified color
	 */
	public boolean hasColor(int color) {
		return this.bitset_.get(color);
	}

	/**
	 * set/unset color in this ColorSet
	 */
	public void setColor(int color, boolean b) {
		this.bitset_.set(color,b);
	}

	/**
	 * set color in this ColorSet
	 * same as : setColor(i, true)
	 */
	public void setColor(int color) {
		this.bitset_.set(color);
	}

	/**
	 * unset color in this ColorSet
	 * same as : setColor(i, false)
	 */
	public void clearColor(int color) {
		this.bitset_.clear(color);
	}
	
	/**
	 * clear all colors in this ColorSet
	 */
	public void clearColors() {
		this.bitset_.clear();
	}
	
	/**
	 * set all colors in this ColorSet 
	 * at same values as in argument
	 * 
	 * same as <code>this.clearColors(); this.orColors(cset);</code>
	 */
	public void setColors(RColorSet cset) {
		this.clearColors();
		this.orColors(cset);
	}
	
	/**
	 * compute the intersection ('and') between this ColorSet and the
	 * ColorSet argument. i.e. keep only colors that present in both
	 * ColorSets.
	 */
	public void andColors(RColorSet cset) {
		this.bitset_.and(cset.bitset_);
	}

	/**
	 * compute the union ('or') between this ColorSet and the
	 * ColorSet argument. i.e. keep colors that present in either
	 * or both ColorSets.
	 */
	public void orColors(RColorSet cset) {
		this.bitset_.or(cset.bitset_);
	}

	/**
	 * compute the 'xor' between this ColorSet and the
	 * ColorSet argument. i.e. keep colors that present in either
	 * (but not both) ColorSets.
	 */
	public void xorColors(RColorSet cset) {
		this.bitset_.xor(cset.bitset_);
	}

	/**
	 * shift bits to left or right in this ColorSet by specified amount.
	 * shift > 0 : shift bits to left
	 * shift < 0 : shift bits to right
	 */
	public void shift(int shift) {
		if (shift == 0)
			return;
		int size = bitset_.size() + shift;
		BitSet newbits = new BitSet(size > 0 ? size : 1);
		int next = -1;
		while ((next = bitset_.nextSetBit(next+1)) >= 0) {
			if ((next + shift) >= 0)
				newbits.set(next+shift);
		}
		this.bitset_ = newbits;
	}
	
	/**
	 * count the number of colors in this ColorSet
	 */
	public int countColors() {
		  return this.bitset_.cardinality();
	}
	
	/**
	 * return lowest color in this colorSet
	 * (or -1 if colorSet is empty)
	 */
	public int minColor() {
		return bitset_.nextSetBit(0);
	}

	/**
	 * return highest color in this colorSet
	 * (or -1 if colorSet is empty)
	 */
	public int maxColor() {
		int next, prev = -1;
		while ((next = bitset_.nextSetBit(prev+1)) >= 0) {
			prev = next;
		}
		return prev;
	}
	
	/**
	 * get the "logical size" of this ColorSet : the highest color of the set plus one.
	 * Returns zero if the BitSet contains no set bits
	 */
	public int length() {
		return this.bitset_.length();
	}

	/**
	 * tell if this ColorSet is empty
	 * same as (cset.countColors == 0)
	 */
	public boolean isEmpty() {
		  return this.bitset_.isEmpty();
	}

	/**
	 * tell if this ColorSet includes all colors
	 * in the ColorSet argument.
	 */
	public boolean includeColors(RColorSet cset) {
		tmp_.clear();
		tmp_.or(this.bitset_);
		tmp_.or(cset.bitset_);
		return tmp_.equals(this.bitset_);
	}

	/**
	 * tell if this ColorSet intersects the ColorSet argument
	 * i.e. if they have at least one color in common
	 */
	public boolean intersectColor(RColorSet cset) {
		return this.bitset_.intersects(cset.bitset_);
	}
	
	/**
	 * if this ColorSet has a single color return
	 * this color else throws an IllegalValueException
	 */
	public int singleColor() throws IllegalValueException {
		if (countColors() != 1)
			throw new IllegalValueException("count colors yields " + countColors());
		return bitset_.nextSetBit(0);
	}
	
	/*
	 * Helper methods (static)
	 */

	/**
	 * Lookup the RColorSet associated to key in registry and
	 * create an empty RColorSet if it does not exist.
	 * This helper function is equivalent to :
	 * <pre><code>
	 * if (r.has(key))
	 *    return (RColorSet) r.get(key);
	 * else
	 *    return (RColorSet) r.set(key, new RColorSet());
	 * </code></pre>
	 */
	public static final RColorSet getColors(Registry r, String key) {
		RColorSet cset = (RColorSet) r.get(key);
		if (cset == null) { 
			cset = new RColorSet();
			r.set(key,cset);
		}
		return cset;
	}

	/**
	 * same as getColors(r, Registry.Colors)
	 */
	public static final RColorSet getColors(Registry r) {
		return getColors(r, Registry.Colors);
	}

	/**
	 * tells if registry r has a RcolorSet on key and if this
	 * RcolorSet is not empty
	 */
	public static final boolean hasColors(Registry r, String key) {
		RColorSet cset = (RColorSet) r.get(key);
		return (cset != null) && (! cset.isEmpty());
	}

	/**
	 * same as hasColors(r, Registry.Colors)
	 */
	public static final boolean hasColors(Registry r) {
		return hasColors(r, Registry.Colors);
	}
	
	/**
	 * tells if registry r has a RcolorSet on key and if this
	 * RColorSet has color i
	 */
	public static final boolean hasColor(Registry r, String key, int i) {
		RColorSet cset = (RColorSet) r.get(key);
		return (cset != null) && cset.hasColor(i);
	}

	/**
	 * same as hasColor(r, Registry.Colors)
	 */
	public static final boolean hasColor(Registry r, int i) {
		return hasColor(r, Registry.Colors, i);
	}
	
	/**
	 * get the number of colors in RColorSet associated o key in registry r.
	 * this number is 0 if there is no RColorSet or if the RColorSet is empty
	 */
	public static final int nbColors(Registry r, String key) {
		RColorSet cset = (RColorSet) r.get(key);
		return ((cset == null) ? 0 : cset.countColors());
	}

	/**
	 * same as nbColors(r, Registry.Colors)
	 */
	public static final int nbColors(Registry r) {
		return nbColors(r, Registry.Colors);
	}

}
