package helix.graph.model.impl;

import helix.graph.model.IllegalValueException;
import helix.graph.model.RegistryValue;
import helix.graph.model.Vertex;
import helix.graph.model.impl.memory.MemGraph;

import org.junit.*;

import static org.junit.Assert.*;

public class TestRegistryValue {

	@Test
	public void testRegistryValue() {
		RegistryValue r0 = new RString("1");
		assertEquals("registryValue string 1", "1", r0.toString());
		assertEquals("registryValue int 1", 1, r0.toInt());
		assertEquals("registryValue float 1", 1f, r0.toFloat());
		assertFalse("registryValue boolean false", r0.toBoolean());
		RegistryValue r1 = new RString("1234");
		RegistryValue r2 = new RInteger(1234);
		assertEquals("registryValue hashCode", r1.hashCode(), r2.hashCode());
		assertTrue("registryValue equals r2", r1.equals(r2));
		assertFalse("registryValue not equals 1234", r1.equals("1234"));
		RegistryValue r3 = new RBoolean(false);
		assertFalse("registryValue not equals r3", r1.equals(r3));
		RegistryValue r4 = new RFloat(1234f);
		assertFalse("registryValue not equals r4", r2.equals(r4));
		assertFalse("registryValue not equals r2", r4.equals(r2));
	}

	@Test(expected= IllegalValueException.class)
	public void testThrow1() {
		new RString(null);
	}

	@Test(expected= IllegalValueException.class)
	public void testThrow2() {
		new RInteger((String) null);
	}

	@Test(expected= IllegalValueException.class)
	public void testThrow3() {
		new RInteger("abc");
	}

	@Test
	public void testThrow4() {
		RegistryValue r1 = new RBoolean("abc");
		assertFalse("registryValue boolean false", r1.toBoolean());
		RegistryValue r2 = new RBoolean("true");
		assertTrue("registryValue boolean true", r2.toBoolean());
	}

	@Test(expected= IllegalValueException.class)
	public void testThrow5() {
		new RColorSet("abc");
	}

	@Test
	public void testColorSet() {
		RColorSet r1 = new RColorSet();
		r1.setColor(1);
		r1.setColor(2);
		r1.setColor(3);
		r1.setColor(4);
		r1.clearColor(3);
		assertEquals("RColorSet count", 3, r1.countColors());
		assertTrue("RcolorSet bit0 set ", r1.hasColor(1));
		assertFalse("RcolorSet bit3 unset", r1.hasColor(3));
		assertEquals("RColorSet value int", 10110, r1.toInt());
		assertEquals("RColorSet value string", "10110", r1.toString());

		RColorSet r2 = new RColorSet("10110");
		assertTrue("RcolorSet r1==r2", r1.equals(r2));
		
		RegistryValue r3 = new RString("10110");
		assertFalse("RcolorSet r1!=r3", r1.equals(r3));
		
		r1.shift(0);
		assertEquals("RColorSet shift 0", "10110", r1.toString());
		r1.shift(2);
		assertEquals("RColorSet shift 0", "1011000", r1.toString());
		r1.shift(-3);
		assertEquals("RColorSet shift 0", "1011", r1.toString());
		
		r1.clearColors();
		assertEquals("RColorSet mincolor", -1, r1.minColor());
		assertEquals("RColorSet maxcolor", -1, r1.maxColor());
		
		r1.setColor(2);
		r1.setColor(3);
		r1.setColor(4);

		assertEquals("RColorSet mincolor", 2, r1.minColor());
		assertEquals("RColorSet maxcolor", 4, r1.maxColor());

		r1.setColor(0);

		assertEquals("RColorSet mincolor", 0, r1.minColor());

		r1.clearColors();
		r1.setColor(1);
		r1.setColor(2);
		r1.setColor(3);

		r2.clearColors();
		r2.setColor(2);
		r2.setColor(3);
		
		assertTrue("RColorSet include strict", r1.includeColors(r2));

		r2.setColor(1);
		assertTrue("RColorSet include equal", r1.includeColors(r2));

		r2.setColor(4);
		assertFalse("RColorSet not include", r1.includeColors(r2));
	}

	@Test
	public void testKloneImmutable() {
		RegistryValue r0, r1;
		r0 = new RString("1");
		r1 = r0.klone();
		assertTrue("Klone Immutable [string]", r0 == r1);
		r0 = new RInteger(1);
		r1 = r0.klone();
		assertTrue("Klone Immutable [integer]", r0 == r1);
	}

	@Test
	public void testKloneColor() {
		RColorSet r2 = new RColorSet("1");
		assertTrue("RcolorSet bit0 set ", r2.hasColor(0));
		RColorSet r3 = (RColorSet) r2.klone();
		assertTrue("Klone colorset", r2 != r3);
		assertTrue("RcolorSet bit0 set ", r3.hasColor(0));
		r3.setColor(4);
		assertTrue("RcolorSet bit4 set ", r3.hasColor(4));
		assertFalse("RcolorSet bit4 unset ", r2.hasColor(4));
	}

	@Test
	public void testRVertex() {
	   	MemGraph g = new MemGraph();
        Vertex v1 = g.addVertex();
        Vertex v2 = g.addVertex();
        RVertex r = new RVertex(v2);
        v1.set("vertex", r);
        RVertex rr = (RVertex) v1.get("vertex");
		assertEquals("rvertex", rr.get(), v2);
        RVertex rc = (RVertex) rr.klone();
        rc.set(v1);
		assertEquals("rvertex", rc.get(), v1);
		assertEquals("rvertex", rr.get(), v2);
	}
	
	@Test
	public void testRVertexList() {
	   	MemGraph g = new MemGraph();
        Vertex v1 = g.addVertex();
        Vertex v2 = g.addVertex();
		RVertexList t = new RVertexList();
		t.addVertex(v1);
		t.addVertex(v2);
		t.addVertex(v1);
		assertEquals("tuple size", 3, t.size());
	}
}
