package helix.graph.model.impl.memory;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.algo.Cloner;

public class TestCloner {

	private Graph g;
	private Vertex v1, v2, v3, v4, v5;

	/**
	 * Sets up the test fixture. 
	 * (Called before *every* test case method.)
	 * 
	 * this graph is :
	 * 
	 *         v1
	 *        / | \
 	 *     v2--v3  v4
 	 *      |
 	 *     v5
	 * 
	 * 
	 * 
	 */
	@Before
	public void initGraph() {
		g = new MemGraph();
		v1 = g.addVertex();
		v2 = g.addVertex();
		v3 = g.addVertex();
		v4 = g.addVertex();
		v5 = g.addVertex();
		v1.set("label", "v1");
		v2.set("label", "v2");
		v3.set("label", "v3");
		v4.set("label", "v4");
		v5.set("label", "v5");
		g.addEdge(v1, v2);
		g.addEdge(v2, v3);
		g.addEdge(v1, v3);
		g.addEdge(v1, v4);
		g.addEdge(v2, v5);
	}

	/**
	 * Tears down the test fixture. 
	 * (Called after *every* test case method.)
	 */
	@After
	public void tearDown() {
	}

	/*
	 * Graph tests
	 */

	@Test
	public void testDuplicateVertex() {
		Vertex nv1 = Cloner.duplicateVertex(g, v1);
		assertEquals("duplicate size", 6, g.nbVertices());
		assertEquals("duplicate label", "v1", nv1.get("label").toString());
		assertFalse("duplicate nv1 v1", g.hasEdge(nv1, v1));
		assertTrue("duplicate nv1 v2", g.hasEdge(nv1, v2));
		assertTrue("duplicate nv1 v3", g.hasEdge(nv1, v3));
		assertTrue("duplicate nv1 v4", g.hasEdge(nv1, v4));
		assertFalse("duplicate nv1 v5", g.hasEdge(nv1, v5));
	}

	@Test
	public void testDuplicateVertices() {
		GSet<Vertex> set = new MemGSet<Vertex>();
		set.add(v1);
		set.add(v2);
		GSet<Vertex> dup = Cloner.duplicateVertices(g, set);
		assertEquals("duplicate size", 2, dup.size());
		Vertex w1 = dup.findUnique("label", "v1");
		Vertex w2 = dup.findUnique("label", "v2");
		assertTrue("duplicate w1 w2", g.hasEdge(w1, w2));
		assertTrue("duplicate w1 v3", g.hasEdge(w1, v3));
		assertTrue("duplicate w1 v4", g.hasEdge(w1, v4));
		assertTrue("duplicate w2 v3", g.hasEdge(w2, v3));
		assertTrue("duplicate w2 v5", g.hasEdge(w2, v5));
		assertFalse("duplicate w1 v1", g.hasEdge(w1, v1));
		assertFalse("duplicate w1 v2", g.hasEdge(w1, v2));
		assertFalse("duplicate w1 v5", g.hasEdge(w1, v5));
		assertFalse("duplicate w2 v4", g.hasEdge(w2, v4));
	}

	@Test
	public void testClone() {
		Graph gg = Cloner.cloneGraph(g);
		assertEquals("clone size", 5, gg.nbVertices());
		v1 = gg.findVertex("label", "v1");
		v2 = gg.findVertex("label", "v2");
		v3 = gg.findVertex("label", "v3");
		v4 = gg.findVertex("label", "v4");
		v5 = gg.findVertex("label", "v5");
		assertTrue("clone v1 v2", gg.hasEdge(v1, v2));
		assertTrue("clone v1 v3", gg.hasEdge(v1, v3));
		assertTrue("clone v1 v4", gg.hasEdge(v1, v4));
		assertTrue("clone v2 v3", gg.hasEdge(v2, v3));
		assertTrue("clone v2 v5", gg.hasEdge(v2, v5));
		assertFalse("clone v1 v5", gg.hasEdge(v1, v5));
		assertFalse("clone v2 v4", gg.hasEdge(v2, v4));
	}

	@Test
	public void testSubGraph() {
		GSet<Vertex> set = new MemGSet<Vertex>();
		set.add(v1);
		set.add(v4);
		set.add(v5);
		Graph gg = Cloner.subGraph(g, set);
		assertEquals("subgraph size", set.size(), gg.nbVertices());
		v1 = gg.findVertex("label", "v1");
		v4 = gg.findVertex("label", "v4");
		v5 = gg.findVertex("label", "v5");
		assertTrue("clone v1 v4", gg.hasEdge(v1, v4));
		assertFalse("clone v1 v5", gg.hasEdge(v1, v5));
		assertFalse("clone v4 v5", gg.hasEdge(v4, v5));
	}

	@Test
	public void testUnion() {
		Graph gg = Cloner.cloneGraph(g);
		gg.findVertex("label", "v1").set("label", "w1");
		gg.findVertex("label", "v2").set("label", "w2");
		Graph union = Cloner.union(g, gg);
		assertEquals("union size", 10, union.nbVertices());
		v1 = union.findVertices("label", "v1").findUnique("label", "v1");
		v2 = union.findVertices("label", "v2").findUnique("label", "v2");
		assertEquals("v1 degree", 3, v1.degree());
		assertEquals("v2 degree", 3, v2.degree());
		assertTrue("union v1 v2", union.hasEdge(v1, v2));
	}
	
}
