 package helix.graph.model.impl.memory;

import java.io.FileReader;
import java.io.IOException;

import org.junit.*;

import static org.junit.Assert.*;

import helix.graph.algo.BasicGenerator;
import helix.graph.algo.CircleGenerator;
import helix.graph.algo.StarGenerator;
import helix.graph.io.DimacsExtendedReader;
import helix.graph.io.GraphReader;
import helix.graph.model.DuplicateEdgeException;
import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.NoSuchVertexException;
import helix.graph.model.SameVertexException;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.util.MemoryBench;

public class TestMemGraph {

    private Graph g;
    private Vertex v1, v2;

    /**
     * Sets up the test fixture. 
     * (Called before *every* test case method.)
     */
    @Before
    public void initGraph() {
    	g = new MemGraph();
        v1 = g.addVertex();
        v2 = g.addVertex();
    }
    
    /**
     * Tears down the test fixture. 
     * (Called after *every* test case method.)
     */
    @After
    public void tearDown() {
    	g = null;
    	v1 = v2 = null;
    }
    
    /*
     * Graph tests
     */

    @Test
    public void testGraph() {
        assertNotNull("graph", g);
        assertNotNull("v1", v1);
        assertNotNull("v2", v2);
        assertEquals("graph vertices", 2, g.nbVertices());
    }

    @Test
    public void testAddEdge() {
        Edge e = g.addEdge(v1, v2);
        Edge ee = g.getEdge(v1, v2);
        assertNotNull("edge", ee);
        assertEquals("same edge", e, ee);
        ee = g.getEdge(v2, v1);
        assertEquals("same edge", e, ee);
        ee = v1.getEdgeTo(v2);
        assertEquals("same edge", e, ee);
        ee = v2.getEdgeTo(v1);
        assertEquals("same edge", e, ee);
    }
    
    @Test(expected=DuplicateEdgeException.class)
    public void testDuplicateEdgeException() {
        g.addEdge(v1, v2);
    	g.addEdge(v2, v1);
    }

    @Test(expected=SameVertexException.class)
    public void testSameVertexException() {
        g.addEdge(v1, v1);
    }
 
    @Test(expected=NoSuchVertexException.class)
    public void testNoSuchVertexException() {
    	MemGraph gg = new MemGraph();
    	Vertex vv = gg.addVertex();
        g.addEdge(v1, vv);
    }

    @Test
    public void testRemoveEdge() {
        Edge e = g.addEdge(v1, v2);
        g.removeEdge(e);
        Edge ee = g.getEdge(v1, v2);
        assertNull("edge", ee);
    }

    @Test
    public void testRemoveVertex() {
        g.addEdge(v1, v2);
        g.removeVertex(v1);
        assertEquals("graph vertices", 1, g.nbVertices());
    }
    
    @Test(expected=NoSuchVertexException.class)
    public void testNoSuchVertex() {
        g.addEdge(v1, v2);
        g.removeVertex(v1);
        assertEquals("graph vertices", 1, g.nbVertices());
        g.getEdge(v1, v2);
    }

    @Test
    public void testDisconnectVertex() {
        g.addEdge(v1, v2);
        g.disconnectVertex(v1);
        Edge ee = g.getEdge(v1, v2);
        assertNull("edge", ee);
    }

	@SuppressWarnings("unused")
    @Test
    public void testVertexIterator() {
    	int count = 0;
    	for (Vertex v : g.vertices()) {
    		count++;
    	}
        assertEquals("vertex count", 2, count);
    }

    /*
     * Vertex tests
     */

    @Test
    public void testVertexDegree() {
        g.addEdge(v1, v2);
        Vertex v3 = g.addVertex();
        assertEquals("graph vertices", 3, g.nbVertices());
        g.addEdge(v1, v3);
        assertEquals("degree v1", 2, v1.degree());
    }

    @Test
    public void testVertexEdgeTo() {
        g.addEdge(v1, v2);
        Vertex v3 = g.addVertex();
        Edge e = v1.getEdgeTo(v2);
        assertNotNull("edge", e);
        e = v1.getEdgeTo(v3);
        assertNull("edge", e);
        g.addEdge(v1, v3);
        e = v1.getEdgeTo(v3);
        assertNotNull("edge", e);
    }

    @Test
    public void testVertexEdges() {
        g.addEdge(v1, v2);
        Vertex v3 = g.addVertex();
        g.addEdge(v1, v3);
    	
        int count = 0;
        for (Edge e : v1.edges()) {
        	count++;
        	Vertex v = e.getVertexTo(v1);
        	Vertex vv = e.getVertexTo(v);
        	assertEquals("v1", vv, v1);
        }
        assertEquals("degree v1", 2, count);
    }

    @Test
    public void testVertexNeighbors() {
        g.addEdge(v1, v2);
        Vertex v3 = g.addVertex();
        g.addEdge(v1, v3);
    	
        
        int count = 0;
        for (Vertex v : v1.neighbors()) {
        	count++;
        	assertTrue("edge v1 v", g.hasEdge(v1, v));
        }
        assertEquals("degree v1", 2, count);
    }
    
    
    /*
     * Edge tests
     */
    
    @Test
    public void testEdgeHasVertex() {
    	Edge e = g.addEdge(v1, v2);
    	assertTrue("has vertex", e.hasVertex(v1));
    	assertTrue("has vertex", e.hasVertex(v2));
    }

    @Test
    public void testEdgeGetVertexTo() {
    	Edge e = g.addEdge(v1, v2);
    	Vertex v = e.getVertexTo(v1);
    	assertEquals("v2", v, v2);
    	v = e.getVertexTo(v2);
    	assertEquals("v1", v, v1);
    }

    @Test
    public void testEdgeToArray() {
    	Edge e = g.addEdge(v1, v2);
    	Vertex[] a = e.toArray();
    	assertEquals("array size", 2, a.length);
    	int idx = a[0].equals(v1) ? 0 : 1;
    	assertEquals("v1", a[idx], v1);
    	idx = (idx + 1) % 2;
    	assertEquals("v2", a[idx], v2);
    }

    @Test
    public void testEdgeEquals() {
    	Edge e1 = g.addEdge(v1, v2);
    	g.removeEdge(e1);
    	Edge e2 = g.addEdge(v2, v1);
    	g.removeEdge(e2);
    	assertEquals("edge hashcode", e1.hashCode(), e2.hashCode());
    	assertTrue("edge equals", e1.equals(e2));
    }

    /*
     * Vertex/Edge/Graph Registry tests
     */

	@SuppressWarnings("unused")
    @Test
    public void testRegistrySimple() {
    	g.set("title", "standard graph");
    	assertEquals("graph registry", g.get("title").toString(), "standard graph");
    	
    	v1.set("age", 25);
    	assertEquals("graph registry", 25, v1.get("age").toInt());
    	assertEquals("graph registry", 25f, v1.get("age").toFloat());
    	assertEquals("graph registry", "25", v1.get("age").toString());
    	assertTrue(  "graph registry", v1.get("age").toBoolean());

    	Edge e = g.addEdge(v1, v2);
    	e.set("color", "blue");
    	assertEquals("edge registry", "blue", e.get("color").toString());
    	int count = 0;
    	for (String key : e.keys()) {
    		count++;
    	}
    	assertEquals("edge registry", 1, count);
    }

    @Test
    public void testRegistryColor() {
		RColorSet color = new RColorSet();
    	color.setColor(1);
    	color.setColor(3);
    	v1.set("color", color);
    	v1.set("age", 25);
    	color = (RColorSet) v1.get("color");
       	assertTrue("vertex registry", color.hasColor(1));
       	assertFalse("vertex registry", color.hasColor(2));
       	assertTrue("vertex registry", color.hasColor(3));
       	v2.copy(v1);
       	color.setColor(2);
       	RColorSet color2 = (RColorSet) v2.get("color");
       	assertTrue("vertex registry", color.hasColor(2));
       	assertFalse("vertex registry", color2.hasColor(2));
    	assertEquals("vertex registry", 25, v2.get("age").toInt());
    }

    @Test
    public void testFind() {
    	v1.set("color", "red");
    	v2.set("color", "blue");
    	Vertex u1;
    	u1 = g.findVertex("color", "red");
    	assertEquals("findVertex red", u1, v1);
    	u1 = g.findVertex("color", "green");
    	assertEquals("findVertex green", u1, null);
    	Vertex v3 = g.addVertex();
    	v3.set("color", "red");
    	GSet<Vertex> s = g.findVertices("color", "red");
    	assertTrue("findVertices v1", s.contains(v1));
    	assertFalse("findVertices v2", s.contains(v2));
    	assertTrue("findVertices v3", s.contains(v3));
    }

    /*
     * Memory usage tests
     */
    
    private final int nbEdges_(Graph g) {
    	int n = 0;
    	for (Vertex v : g.vertices()) {
    		n += v.degree();
    	}
    	return (n / 2);
    }

	@Test
	public void testMemUsage() throws IOException {
		
		int nv = 100;
		
		// basic graph : no edge
		BasicGenerator gen = new BasicGenerator(nv);
		GraphFactory factory = MemGraphFactory.instance();
		
		long m0 = MemoryBench.usedMemory();
		Graph g = gen.generate(factory);
		long dm = MemoryBench.usedMemory() - m0;
		float bpv = dm / nv;
		System.out.println("BasicGraph: byte per empty vertex " + bpv);
		
		// registry overhead
		m0 = MemoryBench.usedMemory();
		for (Vertex v : g.vertices()) {
			v.set("value", 1);
			v.remove("value");
		}
		dm = MemoryBench.usedMemory() - m0;
		float bpi = dm / nv;
		System.out.println("BasicGraph: byte per registry overhead " + bpi);
		
		// registry
		m0 = MemoryBench.usedMemory();
		int nr = 10;
		for (Vertex v : g.vertices()) {
			for (int i = 0 ; i < nr ; i++)
			    v.set("value_" + i, i);
		}
		dm = MemoryBench.usedMemory() - m0;
		float bpj = dm / (nv * nr);
		System.out.println("BasicGraph: byte per registry value " + bpj);

		// edges (chain)
		g = null;
		CircleGenerator cgen = new CircleGenerator(nv);
		m0 = MemoryBench.usedMemory();
		g = cgen.generate(factory);
		int ne = nbEdges_(g);
		dm = MemoryBench.usedMemory() - m0;
		float bpe = (dm - (nv * bpv)) / ne;
		System.out.println("CircleGraph: byte per empty edge " + bpe);

		// edges (star)
		g = null;
		StarGenerator sgen = new StarGenerator(nv);
		m0 = MemoryBench.usedMemory();
		g = sgen.generate(factory);
		ne = nbEdges_(g);
		dm = MemoryBench.usedMemory() - m0;
		bpe = (dm - (nv * bpv)) / ne;
		System.out.println("CircleGraph: byte per empty edge " + bpe);
		
		// real graph
		g = null;
		FileReader freader = new FileReader("samples/sample.poly");
		GraphReader reader = new DimacsExtendedReader(freader, factory);
		m0 = MemoryBench.usedMemory();
		g = reader.read();
		nv = g.nbVertices();
		ne = nbEdges_(g);
		dm = MemoryBench.usedMemory() - m0;
		float bpt = dm / (ne+nv);
		System.out.println("SampleGraph: nv=" + nv + " ne=" + ne + " size_(Mb)=" + (dm/1e6) + " byte_per_element=" + bpt);
	}
    

    /*
     * Graph visiting tests
     */

    
    /*
     * Complexity tests
     */
    
}
