package helix.graph.program.isofun;

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.reporter.BaseReporter;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RVertexList;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;
import java.util.Arrays;

/**
 * a reporter to compute and print gene contiguity in syntons.
 */

public class ContiguityReporter extends BaseReporter {

	// sort int[] by increasing absolute value of int[0]
	protected static class ArrayComparator implements Comparator<int[]> {
		public int compare(int[] arg0, int[] arg1) {
			return (Math.abs(arg0[0]) - Math.abs(arg1[0]));
		}
	}

	// a static instance of this comparator
	protected static ArrayComparator compare_ = new ArrayComparator();
	
	/*
	 * utilities
	 */

	/**
	 * return sign of argument
	 */
	protected static final int sign_(int x) {
		return (x >= 0 ? 1 : -1);
	}

	/**
	 * get direction of gene and return true if direct
	 */
	protected static final boolean isDirect_(Vertex v) {
		return (v.has(Registry.IsDirect) && v.get(Registry.IsDirect).toBoolean());
	}

	/**
	 * get sign of gene
	 */
	protected static final int getSign_(Vertex v) {
		return (isDirect_(v) ? 1 : -1);
	}
	
	/**
	 * get rank of gene
	 */
	protected static final int getPosition_(Vertex v) {
		return (v.has(Registry.Rank) ? v.get(Registry.Rank).toInt() : -1);
	}

	/**
	 * transpose n x m array to m x n array
	 */
	protected static int[][] transpose_(int[][] array) {
		int n = array.length;
		int m = array[0].length;
		int[][] result = new int[m][n];
		
		for (int i = 0 ; i < n ; i++) {
			for (int j = 0 ; j < m ; j++) {
				result[j][i] = array[i][j];
			}
		}
		return result;
	}

	/**
	 * replace values in array by their original ranks<br>
	 * in case of ties, the rank is not incremented<br>
	 * note: the original array is overwritten<br>
	 * e.g. <code>
	 * array = [20 30 10] => array = [1 2 0]
	 * array = [20 30 10 30 50] => array = [1 2 0 2 3]
	 * </code><br>
	 */
	protected static void sortToRanks_(int[] array) {
		int n = array.length;
		int[][] tab = new int[n][2];

		// premier tri par valeurs
		for (int i = 0 ; i < n ; i++) {
			tab[i][0] = array[i];
			tab[i][1] = i;
		}
		Arrays.sort(tab, compare_);
		
		// put ranks and sort again according to original indices
		int rank = -1;
		int previous = tab[0][0] + 1; // just a value != tab[0][0]
		for (int i = 0 ; i < n ; i++) {
			if (tab[i][0] != previous) {
				rank++;
				previous = tab[i][0];
			}
			tab[i][0] = tab[i][1]; // original index for resorting
			tab[i][1] = rank;
		}
		Arrays.sort(tab, compare_);

		// put result (ranks) into original array
		for (int i = 0 ; i < n ; i++) {
			array[i] = tab[i][1];
		}
	}

	/**
	 * get an array of signed positions (i.e.signed ranks) for all colors in V
	 */
	protected static int[] getPositions_(Vertex V) {
		RVertexList tuple = C3PartUtil.getNodes(V);
		int[] positions = new int[tuple.size()];
		
		int color = 0;
		for (Vertex v : tuple) {
			positions[color++] = getSign_(v) * getPosition_(v);
		}
		
		return positions;
	}
	
	/**
	 * get number of ties in array starting at 'start' (excluded)
	 */
	protected static int getNbTies_(int[][] positions, int start) {
		int nties = 0;
		for (int i = start + 1 ; i < positions.length ; i++) {
			if (positions[i][0] != positions[start][0])
				break;
			nties++;
		}
		return nties;
	}
	
	/**
	 * determine if the pair (i,j) is contiguous for all colors
	 * we use the ranks array to speedup the test
	 */
	protected static boolean isContigous_(int[][] ranks, int i, int j) {
		int ncolors = ranks.length;
		for (int color = 0 ; color < ncolors ; color++) {
			int delta = Math.abs(ranks[color][i] - ranks[color][j]);
			if (delta > 1)
				return false;
		}
		return true;
	}
	
	/**
	 * get pattern of gaps
	 */
	protected static List<Integer> gapPattern_(int[][] positions, int i, int j) {
		int ncolors = positions[0].length;
		ArrayList<Integer> pattern = new ArrayList<Integer>(ncolors);
		
		// fill pattern with gap values
		for (int color = 0 ; color < ncolors ; color++) {
			int signi = sign_(positions[i][0] * positions[i][color]);
			int signj = sign_(positions[j][0] * positions[j][color]);
			int gap  = (Math.abs(positions[j][color]) - Math.abs(positions[i][color]));
			// rule for gap sign :
			// if same_sign s : sign =  s * sign(gap)
			// else             sign = -1
			if (signi == signj)
				gap *= signi;
			else
				gap = - Math.abs(gap);
			pattern.add(gap);
		}

		// get value of the first non null element
		int first = 0;
		for (int color = 0 ; color < ncolors ; color++) {
			if (pattern.get(color) != 0) {
				first = pattern.get(color);
				break;
			}
		}
		
		// reverse sign if the first non null element is < 0
		if (first < 0) {
			for (int color = 0 ; color < ncolors ; color++) {
				pattern.set(color, - pattern.get(color));
			}
		}
		
		return pattern;
	}
	
	/**
	 * store gap pattern into patterns table
	 * key is pattern
	 * value is count
	 */
	protected static void storePattern_(int[][] positions, int i, int j, Hashtable<List<Integer>, Integer> patterns) {
		
		List<Integer> pattern = gapPattern_(positions, i, j);
		int count = (patterns.containsKey(pattern) ? patterns.get(pattern) : 0);
		patterns.put(pattern, count+1);
		
	}

	/**
	 * printout patterns
	 */
	protected static void printPatterns_(int[][] patterns) {
		int npat = patterns.length;
		if (npat == 0) return;
		int ncolor = patterns[0].length;
		for (int i = npat-1 ; i >= 0 ; i--) {
			print("g [" + patterns[i][0] + "] ");
			for (int j = 1 ; j < ncolor ; j++) {
				print(patterns[i][j] + " ");
			}
			println("");
		}
	}
	
	/*
	 * API
	 */
	
	/**
	 * print synton contiguity
	 */
	public static void printContiguity(Graph synton) {

		int nbNodes = synton.nbVertices();
		
		int[][] positions = new int[nbNodes][];

		
		// get gene positions per color
		//
		int inode = 0;
		for (Vertex V : synton.vertices()) {
			positions[inode++] = getPositions_(V);
		}

		// sort positions along color 0 (i.e. first genome)
		//
		Arrays.sort(positions, compare_);
		
		// transpose, and keep ranks 
		// this is a speedup for function isContigous_
		//
		int[][] ranks = transpose_(positions);
		for (int i = 0 ; i < ranks.length ; i++) {
			sortToRanks_(ranks[i]);
		}

		// compute gaps patterns 
		//
		Hashtable<List<Integer>, Integer> patterns = new Hashtable<List<Integer>, Integer>();
		
		for (int i = 0 ; i < nbNodes - 1 ; i++) {
			
			int nties = getNbTies_(positions, i);            // # ties from i for color 0
			
			int jmax = Math.min(i + nties + 1, nbNodes - 1); // don't overflow please
			
			for (int j = i+1 ; j <= jmax ; j++) {  // we are checking pair i, j
				
				if (isContigous_(ranks, i, j)) {   // pair is contigous on all colors
					
					storePattern_(positions, i, j, patterns);
					
				}
			}
		}

		// sort patterns according to number of occurences
		//
		int[][] pattab = new int[patterns.size()][ranks.length + 1];
		int ipat = 0;
		for (List<Integer> lst : patterns.keySet()) {
			pattab[ipat][0] = patterns.get(lst);
			for (int i = 0 ; i < lst.size() ; i++) {
				pattab[ipat][i+1] = lst.get(i);
			}
			ipat++;
		}
		Arrays.sort(pattab, compare_);
		
		// printout patterns
		//
		printPatterns_(pattab);
	}

}
