package helix.graph.program.isofun;

import java.util.HashSet;
import java.util.Set;

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.reporter.BaseReporter;
import helix.graph.model.Graph;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;

/**
 * a postprocessor to compute and print epsilon degeneracy
 * of syntons
 */

public class DegeneracyReporter extends BaseReporter {

	protected static class Statistics {
		protected int count;
		protected float min, max, mean;
		protected Statistics() {};
	}
	
	/**
	 * count number of epsilon links for vertex
	 */
	
	private static final int countEpsilon_(Vertex v) {
		int count = 0;
		int color = getColor_(v);
		if (color < 0)
			return count;
		for (Vertex u : v.neighbors()) {
			if (getColor_(u) != color)
				count++;
		}
		return count;
	}
	
	/**
	 * get epsilon statistics for a set of vertices
	 */
	private static final Statistics countEpsilon_(Iterable<Vertex> set, int nbColors) {
		Statistics stat = new Statistics();
		float norm = (float) (nbColors <= 1 ? 1 : nbColors - 1);
		for (Vertex v : set) {
			float degen = (float) countEpsilon_(v) / norm;
			if (stat.count == 0)
				stat.min = stat.max = degen;
			if (degen < stat.min)
				stat.min = degen;
			if (degen > stat.max)
				stat.max = degen;
			stat.mean += degen;
			stat.count++;
		}
		
		stat.mean /= (float) (stat.count == 0 ? 1 : stat.count);

		return stat;
	}
	
	/*
	 * 
	 */
	private static final void log_(String header, Statistics stat) {
		log(header + " min= " + stat.min + " max= " + stat.max + " mean= " + stat.mean);
	}
	
	/*
	 * API
	 */

	/**
	 * compute and print epsilon degeneracy information
	 * for multigraph results (i.e. syntons)
	 */
	public static void logDegeneracy(Set<Graph> result) {

		int  nbColors = 0;
		
		// collect datagraph nodes
		
		HashSet<Vertex> nodes = new HashSet<Vertex>();
		for (Graph g : result) {
			if (nbColors <= 0)
				nbColors = RColorSet.nbColors(g);
			for (Vertex V : g.vertices()) {
				for (Vertex v : C3PartUtil.getNodes(V)) {
					nodes.add(v);
				}
			}
		}
		
		// count degeneracy
		
		Statistics stat = countEpsilon_(nodes, nbColors);
		
		log_("out-degeneracy", stat);
	}

	/**
	 * compute and print epsilon degeneracy information
	 * for input datagraph
	 */

	public static void logDegeneracy(Graph datagraph) {
		int  nbColors = RColorSet.nbColors(datagraph);
		
		Statistics stat = countEpsilon_(datagraph.vertices(), nbColors);
		
		log_("in-degeneracy", stat);
	}
}
