package helix.graph.program.isofun;

import helix.graph.algo.c3part.C3PartAlgo;
import helix.graph.algo.c3part.C3PartBFS;
import helix.graph.algo.c3part.C3PartDFS;
import helix.graph.algo.c3part.C3PartDFSQ;
import helix.graph.algo.c3part.C3PartOTF;
import helix.graph.algo.c3part.C3PartOTFBasic;
import helix.graph.algo.c3part.C3PartOTFQLocal;
import helix.graph.algo.c3part.C3PartOTFQGlobal;
import helix.graph.algo.c3part.C3PartOptions;
import helix.graph.model.Graph;
import helix.graph.util.CommandLineException;
import helix.graph.util.LogUtil;

import java.util.ArrayList;
import java.util.List;

public class IsofunOptions extends C3PartOptions {

	/*
	 * know algorithms
	 */
	protected static String ALGO_BFS       = "bfs";
	protected static String ALGO_DFS       = "dfs";
	protected static String ALGO_OTF       = "otf";
	protected static String ALGO_OTF_BASIC = "otfb";
	protected static String ALGO_OTFQL     = "otfql";
	protected static String ALGO_OTFQG     = "otfqg";
	protected static String ALGO_DFSQ      = "dfsq";
	
	/*
	 * change C3PartOptions default values
	 */
	protected static int DFT_MINSIZE = 2; 	     // AV: default should be 2 for OTF
	protected static int DFT_MINELTSIZE = 2; 	 // AV: default should be 2 for OTF
	
	/*
	 * new options default values
	 */
	protected static String DFT_ALGO_NAME = ALGO_OTF;
	protected static boolean DFT_CHECK = true;
	protected static boolean DFT_CLEANUP = true;
	protected static boolean DFT_TANDEM = false;
	protected static boolean DFT_REPORTGAPS = false;
	private static boolean DFT_REPORTOGS = false;
	protected static boolean DFT_MEMDOG = false;
	
	protected static int DFT_TANDEM_MINSIZE = -1; // means auto
	protected static int DFT_TANDEM_MINFRAC = 50;
	
	public String algoName = DFT_ALGO_NAME;
	public boolean check = DFT_CHECK;
	public boolean cleanup = DFT_CLEANUP;
	public boolean tandem = DFT_TANDEM;
	public int tandemMinSize = DFT_TANDEM_MINSIZE;
	public int tandemMinFrac = DFT_TANDEM_MINFRAC;
	public boolean reportGaps = DFT_REPORTGAPS;
	public boolean memdog = DFT_MEMDOG;
	public List<Integer> deltaGap = null;
	public List<Integer> deltaEps = null;
	public boolean reportOGs = DFT_REPORTOGS;
	
	public IsofunOptions(String[] args) {
		
		super(args);
		
		minsize  = DFT_MINSIZE;
		mineltsize  = DFT_MINELTSIZE;
		
		// note: use only uppercase for short options
		//       except M U P S that are already used by C3PartOptions

		// change default
		getOption("minsize").setDescription("[<?>=integer] ccc min size (default " +  DFT_MINSIZE + ")");
		getOption("mineltsize").setDescription("[<?>=integer] ccc min element size (default " +  DFT_MINELTSIZE + ")");

		// new options
		addOption("algo", 'A').setRequiredArgument().setDescription("[<?>=bfs|dfs|otf|otfql|otfqg] select algo");
		
		addOption("check", 'K').setRequiredArgument().setDescription("[<?>=on|off] check datagraph before processing (default: " + boolToString(DFT_CHECK) + ")");
		
		addOption("cleanup", 'C').setRequiredArgument().setDescription("[<?>=on|off] cleanup datagraph before processing (default: " + boolToString(DFT_CLEANUP) + ")");

		addOption("reportgaps", 'G').setRequiredArgument().setDescription("[<?>=on|off] report on gene contiguity and gaps in syntons (default: " + boolToString(DFT_REPORTGAPS) + ")");
		addOption("reportogs", 'Z').setRequiredArgument().setDescription("[<?>=on|off] report OGs in syntons (default: " + boolToString(DFT_REPORTOGS) + ")");

		addOption("memdog", 'O').setRequiredArgument().setDescription("[<?>=on|off] try to report actual memory used (default: " + boolToString(DFT_MEMDOG) + ")");

		addOption("deltagap", 'D').setUnlimitedArguments().setDescription("[<?>=integer+] delta for gaps on each color. use 0 for no gap. " +
				                                                          "if you specify a single value, it will apply on all colors (default: no gaps)");

		addOption("deltaeps", 'E').setRequiredArgument().setDescription("[<?>=integer+] delta closure for epsilon (default: 0 (no closure))");

		addOption("tandem", 'T').setUnlimitedArguments().setDescription("[<?>=off|on [minfrac [minsize]]] cleanup tandems before processing (default: " + boolToString(DFT_TANDEM) + ")");
		
		LogUtil.setHeader(logger, "+ [IsoFun] ");
	}

	/*
	 * helper methods
	 */
	
	/**
	 * get proper algorithm from name
	 */
	public static C3PartAlgo getAlgoForName(String name, Graph graph) {
		
		if (ALGO_BFS.compareToIgnoreCase(name) == 0)
			return new C3PartBFS(graph);
		if (ALGO_DFS.compareToIgnoreCase(name) == 0)
			return new C3PartDFS(graph);
		if (ALGO_OTF.compareToIgnoreCase(name) == 0)
			return new C3PartOTF(graph);
		if (ALGO_OTF_BASIC.compareToIgnoreCase(name) == 0)
			return new C3PartOTFBasic(graph);
		if (ALGO_OTFQL.compareToIgnoreCase(name) == 0)
			return new C3PartOTFQLocal(graph);
		if (ALGO_OTFQG.compareToIgnoreCase(name) == 0)
			return new C3PartOTFQGlobal(graph);
		if (ALGO_DFSQ.compareToIgnoreCase(name) == 0)
			return new C3PartDFSQ(graph);

		throw new CommandLineException("unknown algorithm " + name);
	}
	
	/**
	 * override parse to further check user options.
	 * 
	 * note: further checking may also be done in the
	 * <code>setOptions</code> method for each algo.
	 */
	@Override
	public void parse() throws CommandLineException {
		//
		// parse using super
		//
		super.parse();
		
		//
		// deal with some specific cases
		//
		
		//
		// -maxstar option
		//
		if ((! (    ALGO_OTFQL.equalsIgnoreCase(algoName)
			     || ALGO_OTFQG.equalsIgnoreCase(algoName)
			     || ALGO_DFSQ.equalsIgnoreCase(algoName)))
			&& (maxstar != 0)) {
			logger.severe(  "* warning * --maxstar should only be used with otq[l|g] or dfsq algos (will be forced to 0)");
			maxstar = 0;
		}
	}
	

	/**
	 * new stubs
	 */

	public void set_algo(List<String> args) {
		algoName = args.get(0);
	}

	public void set_check(List<String> args) {
		check = (args.size() == 0) ? true : stringToBool(args.get(0));
	}
	
	public void set_cleanup(List<String> args) {
		cleanup = (args.size() == 0) ? true : stringToBool(args.get(0));
	}

	public void set_tandem(List<String> args) {
		tandem = (args.size() == 0) ? true : stringToBool(args.get(0));
		if (tandem && (args.size() > 1)) {
			tandemMinFrac = Integer.parseInt(args.get(1));
			if ((tandemMinFrac < 0) || (tandemMinFrac > 100))
				throw new CommandLineException("tandem minfraction should be in range [0,100]");
		}
		if (tandem && (args.size() > 2)) {
			tandemMinSize = Integer.parseInt(args.get(2));
		}
	}

	public void set_reportgaps(List<String> args) {
		reportGaps = (args.size() == 0) ? true : stringToBool(args.get(0));
	}

	public void set_reportogs(List<String> args) {
		reportOGs = (args.size() == 0) ? true : stringToBool(args.get(0));
	}

	
	public void set_memdog(List<String> args) {
		memdog = (args.size() == 0) ? true : stringToBool(args.get(0));
	}

	public void set_deltagap(List<String> args) {
		deltaGap = new ArrayList<Integer>();
		for (String str : args) {
			int delta = Integer.parseInt(str);
			if (delta < 0)
				throw new CommandLineException("delta should be >= 0");
			deltaGap.add(delta);
		}
	}
	
	public void set_deltaeps(List<String> args) {
		int delta = stringToInt(args.get(0), -1);
		deltaEps = new ArrayList<Integer>();
		deltaEps.add(delta);
	}
	
}
