package helix.graph.program.isofun;

import helix.graph.algo.c3part.C3PartAlgo;
import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.preprocessor.EpsilonCloser;
import helix.graph.algo.c3part.preprocessor.GraphCleaner;
import helix.graph.algo.c3part.preprocessor.ColorCleaner;
import helix.graph.algo.c3part.preprocessor.GraphCloser;
import helix.graph.algo.c3part.preprocessor.DataGraphCloser;
import helix.graph.algo.c3part.preprocessor.TandemCleaner;
import helix.graph.algo.c3part.reporter.CCCReporter;
import helix.graph.algo.c3part.reporter.GraphChecker;
import helix.graph.algo.c3part.reporter.GraphStatistics;
import helix.graph.io.DimacsExtendedReader;
import helix.graph.io.GraphReader;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.GraphFactory;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;
import helix.graph.model.impl.memory.MemGraphFactory;
import helix.graph.util.CommandLineException;
import helix.graph.util.MemoryBench;
import helix.graph.util.MemoryWatcher;
import helix.graph.util.StopWatch;

import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;







public class Main {
	
	
	protected static boolean checkifcc_(Vertex v, RVertexList U, RColorSet currentColors,RColorSet allColors) {
		
		for (Vertex u : U) {
			if (u.hasEdgeTo(v)) {
				if (! currentColors.hasColor(RColorSet.getColors(u).singleColor())) {
					currentColors.setColor(RColorSet.getColors(u).singleColor());
					if (currentColors.includeColors(allColors)) return true; // end of recursion
					if (checkifcc_(u,U,currentColors,allColors)) return true;
				}
			}
		}
		return false;
	}
	
	
	
	protected static boolean comparetoVertex_(Vertex U, Vertex V) {
		RVertexList Unodes = C3PartUtil.getNodes(U);
		RVertexList Vnodes = C3PartUtil.getNodes(V);
		for (Vertex u : Unodes) {
			boolean missing =true;
			for (Vertex v : Vnodes) {
				if (u.equals(v)) 
					{missing = false;break;}
			}
			if (missing) 
				return false;
		}
		return true;
	}
	
	protected static boolean comparetoVertexList_(GSet<Vertex> og, Vertex V) {
		if (og.isEmpty()) return true;
		
		for (Vertex U : og) {
			if (comparetoVertex_(U,V)) return false;
		}
		return true;
	}
	
	public static void main (String[] args) throws IOException {

		// parse user's command line arguments
		
		IsofunOptions options = new IsofunOptions(args);

		try {
			options.parse();
		}
		catch (CommandLineException cle) {
			System.err.println(cle.getMessage());
        	System.exit(1);
		}
		
        // should remain exactly one argument : datagraph file

        List<String> otherArgs = options.getNonOptionArguments();
        
        if (otherArgs.size() != 1) {
        	System.err.println("need a datagraph input file");
        	System.exit(2);
        }
        
		String inputFile = otherArgs.get(0);

		// run a watchdog to pickup memory usage
		
		MemoryWatcher memdog = new MemoryWatcher(options.logger);
		if (options.memdog) {
			memdog.start();
		}

		// setup a stopwatch for counting time
		
		StopWatch watch = new StopWatch(options.logger);
		watch.start();

try {
		// read graph
		
		FileReader freader = new FileReader(inputFile);
		GraphFactory factory = MemGraphFactory.instance();
		GraphReader reader = new DimacsExtendedReader(freader, factory, options.logger);
		Graph graph = reader.read();
		
		options.logger.info("read graph: N=" + graph.nbVertices() + " E=" + graph.nbEdges() + " C=" + RColorSet.nbColors(graph));

		watch.log("read time:");
		watch.reset();

		// get proper algo
		
		C3PartAlgo algo = IsofunOptions.getAlgoForName(options.algoName, graph);
		
		// set options for this algo
		
		algo.setOptions(options);
		
		// preprocess graph (for datagraph only)
		
		if (! options.ismulti) {
			
			if (options.check) {
				if (! GraphChecker.checkGraph(graph))
					System.exit(3);
			}

			if (options.userColors != null) {
				ColorCleaner cleaner = new ColorCleaner(options.userColors);
				cleaner.run(graph, options.logger);
			}
			
			// delta gap closure

			if (options.deltaGap != null) {
				GraphCloser closer = new DataGraphCloser(options.deltaGap);
				closer.run(graph, options.logger);
			}

			// delta epsilon closure

			if (options.deltaEps != null) {
				EpsilonCloser closer = new EpsilonCloser(options.deltaEps.get(0));
				closer.run(graph, options.logger);
			}

			// cleanup
			
			if (options.cleanup) {
				GraphCleaner cleaner = algo.getAggregatorFor().getCleanerFor();
				cleaner.run(graph, options.logger);
			}
						
			if (options.tandem) {
				if (options.tandemMinSize < 0) // auto mode
					options.tandemMinSize = RColorSet.nbColors(graph) - 1;
				GraphCleaner tandem = new TandemCleaner(options.tandemMinSize, options.tandemMinFrac);
				tandem.run(graph, options.logger);
			}
			
			GraphStatistics.setLogger(options.logger);
			GraphStatistics.logStatistics(graph);
		}
		
		watch.log("preprocessing time:");
		watch.reset();
		//
		// run algo
		//
		Set<Graph> partition = algo.run();

		options.logger.info("found " + partition.size() + " CCC");

		// print Syntons
		
		for (Graph synton : partition) {
			// report syntons
			CCCReporter.printSynton(synton, options);
			// CCCReporterDebug.printSynton(synton, options);
			// report gaps
			if (options.reportGaps)
			  ContiguityReporter.printContiguity(synton);
			if (options.reportOGs) { // 3 cases : no delta, [delta] for all colors, or [delta1,delta2, etc]
				if(options.deltaGap != null){
				  if (options.deltaGap.size()==1) { // case 1 : [delta] for all colors
					  ArrayList<Integer> deltaGap = new ArrayList<Integer>();
					  for (int i = 0; i<RColorSet.nbColors(synton);i++){ 
							deltaGap.add(options.deltaGap.get(0));
					  }
					  OGReporter.printOGs(synton,deltaGap);
				  }
				  else OGReporter.printOGs(synton,options.deltaGap); // case 2 : [delta1,delta2, etc]
				}
				else { // case 3 : no delta
					ArrayList<Integer> deltaGap = new ArrayList<Integer>();
					for (int i = 0; i<RColorSet.nbColors(synton);i++){ 
							deltaGap.add(0);
					}
					OGReporter.printOGs(synton,deltaGap);
				}
				
			}
		}
//		
//		GSet<Vertex> spines_ = graph.vertexSetFactory().make();
//
//		for (Graph synton : partition) {
//			for (Vertex V : synton.vertices()){
//				if (comparetoVertexList_(spines_,V))
//						spines_.add(V);
//			}
//		}
//		
//		System.err.println("NOMBRE DE SPINES : "+ spines_.size());
//
//		RColorSet allColors = new RColorSet();
//		allColors.setColors(RColorSet.getColors(graph));
//		allColors.clearColor(0);
//		
//		RColorSet currentColors = new RColorSet();
//		
//		for (Vertex U : spines_){
//			//System.err.println("spine");
//			currentColors.clearColors();
//			for (Vertex v : C3PartUtil.getNodes(U)){
//				//System.err.println("before");
//				if (!checkifcc_(v, C3PartUtil.getNodes(U), currentColors,allColors)) {System.err.println("EPIC FAIL");break;}
//				else break;
//			}
//			//System.err.println("after");
//		}
		
		//  verbose
		if (options.verbose) {
			DegeneracyReporter.setLogger(options.logger);
			DegeneracyReporter.logDegeneracy(partition);
		}
		
		// end
}
catch (Throwable t) {
	t.printStackTrace();
}
		watch.log("algo runtime:");
		MemoryBench.log(options.logger, "memory:");
		
		if (options.memdog) {
			memdog.stopRun();
			memdog.log("end:");
		}
	}
}
