package helix.graph.program.isofun;

import helix.graph.algo.c3part.C3PartUtil;
import helix.graph.algo.c3part.reporter.BaseReporter;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;
import helix.graph.model.impl.RColorSet;
import helix.graph.model.impl.RVertexList;

import java.util.HashSet;
import java.util.List;


/**
 * a reporter to compute and print OGs in syntons.
 */

public class OGReporter extends BaseReporter {

	/*
	 * utilities
	 */

	/**
	 * return sign of argument
	 */
	protected static final int sign_(int x) {
		return (x >= 0 ? 1 : -1);
	}

	/**
	 * get direction of gene and return true if direct
	 */
	protected static final boolean isDirect_(Vertex v) {
		return (v.has(Registry.IsDirect) && v.get(Registry.IsDirect).toBoolean());
	}

	/**
	 * get sign of gene
	 */
	protected static final int getSign_(Vertex v) {
		return (isDirect_(v) ? 1 : -1);
	}
	
	/**
	 * check if the 4 vertices have the same orientation (that + croisement = worst case)
	 */
	protected static final boolean isWorstCase_(Vertex u1,Vertex u2,Vertex v1, Vertex v2) {
		int s = getSign_(u1);
		if ((s==getSign_(u2))&&(s==getSign_(v1))&&(s==getSign_(v2))) return true;
		return false;
	}
	
	/**
	 * get rank of gene
	 */
	protected static final int getPosition_(Vertex v) {
		return (v.has(Registry.Rank) ? v.get(Registry.Rank).toInt() : -1);
	}
	
	protected static boolean comparetoOGList_(Vertex [] og, HashSet<Vertex []> L) {
		if (L.isEmpty()) return true;
		
		for (Vertex [] og1 : L) {
			if (((og1[0] == og[0])&&(og1[1] == og[1])) || ((og1[0] == og[1])&&(og1[1] == og[0]))) return false;
		}
		return true;
	}
	
	
	/*
	 * API
	 */
	
	/**
	 * print POGs
	 */
	public static void printOGs(Graph synton, List<Integer> deltagap) {
		//System.err.println(deltagap.size());
		
		int nbNodes = synton.nbVertices();
		int nbColors =  RColorSet.nbColors(synton);
		
		Vertex[][] positions = new Vertex[nbNodes][nbColors]; // matrix of layered data graph nodes (ie we need labels)
		//positions[indexVertexMultigraphe][indexColor]
		
		// compute POGs
		//
		int inode = 0;
		for (Vertex V : synton.vertices()) {
			RVertexList tuple = C3PartUtil.getNodes(V);
			int color = 0;
			for (Vertex v : tuple) {
				positions[inode][color++] = v;
			}
			inode++;
		}
		
		int ncolors = positions[0].length;
		
		// In this part of the procedure, the goal is to collect pairs of genes from the same spine
		// that do not respect the majority orientation of the synton on one color or the other
		
		int [] signcolor = new int [ncolors]; // one int for each color : 
											  // 1 if the majority of genes from this color in the synton is direct
											  // -1 if it's indirect
											  // 0 if there's no majority
		for (int color = 0 ; color < ncolors ; color++) {
			int countdirect = 0;
			int nbnonstars = 0;
			for (int u = 0 ; u < nbNodes ; u++) {
				Vertex v = positions[u][color];
				if (!C3PartUtil.isStar(v)) {
					nbnonstars++;
					if (getSign_(v)>0) countdirect++;
				}
			}			
			if ((countdirect == nbnonstars/2) && (nbnonstars== (nbnonstars/2)*2)) signcolor[color]=0;
			else {
				if (countdirect > nbnonstars/2) signcolor[color]=1;
				else signcolor[color]=-1;
			}
		}
		
		
		HashSet<Vertex []> ListofDRs = new HashSet<Vertex []>();
		
		for (int color1 = 0 ; color1 < ncolors ; color1++) {
			for (int color2 = color1 +1 ; color2 < ncolors ; color2++) {
				for (int u = 0 ; u < nbNodes ; u++) {
					Vertex u1 = positions[u][color1];
					Vertex u2 = positions[u][color2];
					if (!((C3PartUtil.isStar(u1)) || (C3PartUtil.isStar(u2))) && (!((getSign_(u1) == signcolor[color1]) && (getSign_(u2) == signcolor[color2])))){
						Vertex [] uu = new Vertex[2];
						uu[0] = u1;
						uu[1] = u2;
						ListofDRs.add(uu);
					}
				}
			}
		}
		

		// In this part, the goal is to collect pairs of genes from the same spine
		// that contradict the majority of the other pairs on the same colors in the synton
		
		HashSet<Vertex []> ListofNOGs = new HashSet<Vertex []>();

		for (int color1 = 0 ; color1 < ncolors ; color1++) {
			for (int color2 = color1+1 ; color2 < ncolors ; color2++) {
				

				// on those two colors, for a given spine u, we count the number of worst cases and put it in a array
				// informally this number is the number of 'bad' crossings : we'll want to recover a spine with a big number of bad crossings
				// as a NOG, even if it has the same orientations as the other spines in the synton
				
				int [] worstcaseweight = new int [nbNodes];

				for (int u = 0 ; u < nbNodes ; u++) {
					worstcaseweight[u] = 0;
					for (int v = 0 ; v < nbNodes ; v++) {
						Vertex u1 = positions[u][color1];
						Vertex u2 = positions[u][color2];
						Vertex v1 = positions[v][color1];
						Vertex v2 = positions[v][color2];
						if((u != v) && (!C3PartUtil.isStar(u1)) && (!C3PartUtil.isStar(v1)) && (!C3PartUtil.isStar(u2)) && (!C3PartUtil.isStar(v2)))  {
							int pos1 = sign_(getPosition_(u1) - getPosition_(v1)); // +1 if u is after v on color1, -1 else
							int pos2 = sign_(getPosition_(u2) - getPosition_(v2)); // same on color2
							
							if ((getPosition_(u1) - getPosition_(v1)) > ((deltagap.get(color1)+1) *nbNodes)) pos1 = -1;// u1=4650, v1=2 : u1 avant v1
							if ((getPosition_(v1) - getPosition_(u1)) > ((deltagap.get(color1)+1) *nbNodes)) pos1 = 1;// v1=4650, u1=2 : u1 apres v1
							if ((getPosition_(u2) - getPosition_(v2)) > ((deltagap.get(color2)+1) *nbNodes)) pos2 = -1;
							if ((getPosition_(v2) - getPosition_(u2)) > ((deltagap.get(color2)+1) *nbNodes)) pos2 = 1;
							if ((pos1*pos2 < 0)&& (isWorstCase_(u1,u2,v1,v2))) { //croisement worst case
								worstcaseweight[u]++;
							}
						}
					}
				}
				for (int u = 0 ; u < nbNodes ; u++) {
					worstcaseweight[u] = nbNodes- worstcaseweight[u];
				}
				
				
				for (int u = 0 ; u < nbNodes ; u++) {
					int nbaccords = 0;
					int nbdesaccords = 0;
					for (int v = 0 ; v < nbNodes ; v++) {
						Vertex u1 = positions[u][color1];
						Vertex u2 = positions[u][color2];
						Vertex v1 = positions[v][color1];
						Vertex v2 = positions[v][color2];
						if((u != v) && (!C3PartUtil.isStar(u1)) && (!C3PartUtil.isStar(v1)) && (!C3PartUtil.isStar(u2)) && (!C3PartUtil.isStar(v2)))  {
							int signu = sign_(getSign_(u1) * getSign_(u2)); // first spine, u, check if its orientation changes
							int signv = sign_(getSign_(v1) * getSign_(v2)); // same for v
							int pos1 = sign_(getPosition_(u1) - getPosition_(v1)); // +1 if u is after v on color1, -1 else
							int pos2 = sign_(getPosition_(u2) - getPosition_(v2)); // same on color2
							
							if ((getPosition_(u1) - getPosition_(v1)) > ((deltagap.get(color1)+1) *nbNodes)) pos1 = -1;// u1=4650, v1=2 : u1 avant v1
							if ((getPosition_(v1) - getPosition_(u1)) > ((deltagap.get(color1)+1) *nbNodes)) pos1 = 1;// v1=4650, u1=2 : u1 apres v1
							if ((getPosition_(u2) - getPosition_(v2)) > ((deltagap.get(color2)+1) *nbNodes)) pos2 = -1;
							if ((getPosition_(v2) - getPosition_(u2)) > ((deltagap.get(color2)+1) *nbNodes)) pos2 = 1;
							
							if ((getPosition_(u1) - getPosition_(v1) == 0) || (getPosition_(u2) - getPosition_(v2) == 0)){
								if (signu==signv) nbaccords++;
								else nbdesaccords++ ;
							}
							
							else{
							
								if (pos1*pos2 > 0) { // ie pas de croisement -> orientation conservee sinon desaccord
									if ((signu==1) && (signv==1)) nbaccords++;
									else nbdesaccords++ ;
								}
								else {// ie croisement -> orientations opposees sinon desaccord
									if ((signu==-1) && (signv==-1)) nbaccords++;
									else {
										if (isWorstCase_(u1,u2,v1,v2)) nbdesaccords+=worstcaseweight[v];
										else nbdesaccords++ ;
									}
								}
							}
						}
					}
					
					if (nbdesaccords > nbaccords) { // we don't return the nodes when there's an equality
						// print u
						Vertex [] uu = new Vertex[2];
//						positions[u][color1].set(Registry.Comment, " des: "+ nbdesaccords + " acc: " + nbaccords);
//						positions[u][color2].set(Registry.Comment, " des: "+ nbdesaccords + " acc: " + nbaccords);
						uu[0] = positions[u][color1];
						uu[1] = positions[u][color2];
						ListofNOGs.add(uu);
//						String nameu1 = getRegistryValue_(positions[u][color1], Registry.Label, "ERROR");
//						String nameu2 = getRegistryValue_(positions[u][color2], Registry.Label, "ERROR");
//						System.err.println("OG : " + nameu1 + " - " + nameu2 + " INV " + nbdesaccords + " NOTINV " + nbaccords);
					}
				}
				
				
			}
		}

		if (!ListofDRs.isEmpty()){ 
			//System.err.println("A new synton with DRs");
			
			HashSet<Vertex []> ListofDRslim = new HashSet<Vertex []>();
			for (Vertex [] og : ListofDRs) {
				if (comparetoOGList_(og, ListofDRslim)) ListofDRslim.add(og);
			}
		
			for (Vertex [] og : ListofDRslim) {
				String nameu1 = getRegistryValue_(og[0], Registry.Label, "ERROR");
				String nameu2 = getRegistryValue_(og[1], Registry.Label, "ERROR");
				
				println("y " + nameu1 + " " + nameu2);// + og[0].get(Registry.Comment)); 
			}
		}
		

		if (!ListofNOGs.isEmpty()){ 
			//System.err.println("A new synton with OGs");
			
			HashSet<Vertex []> ListofNOGslim = new HashSet<Vertex []>();
			for (Vertex [] og : ListofNOGs) {
				if (comparetoOGList_(og, ListofNOGslim)) ListofNOGslim.add(og);
			}
		
			for (Vertex [] og : ListofNOGslim) {
				String nameu1 = getRegistryValue_(og[0], Registry.Label, "ERROR");
				String nameu2 = getRegistryValue_(og[1], Registry.Label, "ERROR");
				
				println("z " + nameu1 + " " + nameu2);// + og[0].get(Registry.Comment)); // note : pb si le vertex appartient a plusieurs ogs, faudrait que ca soit lie a la paire
			}
		}
	
	}
	
}
		


