package helix.graph.util;

/*
 * experimental section : no guarantee
 * 
 */

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantLock;

public class ConcurrentService {

    protected int poolSize_ = 0;
    protected int nbProcess_ = 0;
    
    protected boolean enabled_ = false;
    
	protected ExecutorService executor_;
	protected ReentrantLock lock_;
	protected CountDownLatch latch_;

	/*
     * internal wrapper
     */
	protected class Wrapper implements Runnable {
	
	    Runnable process_;

	    public Wrapper(Runnable process) {
	        process_ = process;
            lock_.lock();
            try {
                nbProcess_++;
            } finally {
                lock_.unlock();
            }
	    }

	    public void run() {
    		int procnum = nbProcess_;
	    	try {
	    		process_.run();
	    	}
	    	catch (Exception e) {
	    		System.err.println("*** process " + procnum + " aborted : "+ e.getMessage());
	    		e.printStackTrace();
	    	}
	    	
            lock_.lock();
            try {
                nbProcess_--;
                if (nbProcess_ == 0) {
                	latch_.countDown();
                }
            } finally {
                lock_.unlock();
            }
	    }
	}

    /*
     * constructor
     * disable with poolsize <= 1
     */
    
    public ConcurrentService(int poolSize) {
    	poolSize_ = (poolSize >= 1) ? poolSize : getAvailableProcessors();
	    executor_ = Executors.newFixedThreadPool(poolSize_);
	    lock_ = new ReentrantLock();
	    latch_ = new CountDownLatch(1);
	    
	    enabled_ = (poolSize_ > 1);
    }
    
    public ConcurrentService() {
      this(0);
    }
	
    /*
     * API
     */

    public static int getAvailableProcessors() {
        return Runtime.getRuntime().availableProcessors();
     }

    public boolean isEnabled() {
    	return enabled_;
    }

    public int getPoolSize() {
    	return poolSize_;
     }

    public int getNbRunningProcesses() {
        return nbProcess_; // approximate 
     }
    
    public void addProcess(Runnable process) {
    	if (enabled_) {
    		Wrapper w = new Wrapper(process);
    		executor_.execute(w);
    	}
    	else {
    		process.run();
    	}
    }

    public void addProcess(Iterable<Runnable> processes) {
    	for (Runnable p : processes) {
    		addProcess(p);
    	}
    }
    
    public void lock() {
    	if (enabled_)
    		lock_.lock();
    }
    
    public void unlock() {
    	if (enabled_)
    		lock_.unlock();
    }

    public void waitCompletion() {
    	if (! enabled_)
    		return;
    	
		try {
			latch_.await();
		} catch (InterruptedException ie) {}
    }
    
    public void shutdown() {
    	if (! enabled_)
    		return;
    	
    	executor_.shutdown();
		try {
			executor_.awaitTermination(Integer.MAX_VALUE, TimeUnit.SECONDS);
		} catch (InterruptedException ie) {}
    }

    
}
