package helix.graph.util;

import helix.graph.model.Edge;
import helix.graph.model.GSet;
import helix.graph.model.Graph;
import helix.graph.model.Registry;
import helix.graph.model.Vertex;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.logging.Handler;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;

public class GraphDebugger {

	/**
	 * logger to log messages 
	 */
	protected Logger logger_;
	
	/**
	 * number of spaces for message indentation 
	 */
	protected int indentSpaces_ = 4;

	/**
	 * current indentation string 
	 */
	protected String indentString_ = "";

	/**
	 * current list of vertex attributes to print
	 */
	protected ArrayList<String> attributes_ = new ArrayList<String>();
	
	/*
	 * internal utilities
	 */

	/**
	 * vertex -> string
	 */
	protected String strVertex_(Vertex v) {
		assertAttributes_();
		StringBuffer buffer = new StringBuffer("n ");
		for (String key : attributes_) {
			String value = v.has(key) ? v.get(key).toString() : "<none>";
			buffer.append(key + ": " + value + " ");
		}
		return buffer.toString().trim();
	}
	
	/*
	 * API
	 */
	
	/**
	 * full constructor
	 */
	public GraphDebugger(Logger logger) {
		logger_ = logger;
	}

	/**
	 * default constructor using LogUtil.getLogger()
	 */
	public GraphDebugger() {
		this(LogUtil.getLogger());
	}
	
	/*
	 * handy methods for logging message
	 */

	/**
	 * get current logger
	 */
	public Logger getLogger() {
		return logger_;
	}

	/**
	 * change current logger
	 */
	public Logger setLogger(Logger logger) {
		Logger previous = logger_;
		logger_ = logger;
		return previous;
	}
	
	/**
	 * set logger on/off
	 * return previous value
	 */
	public boolean setVerbose(boolean isOn) {
		boolean old = LogUtil.isVerbose(logger_);
		LogUtil.setVerbose(logger_, isOn);
		return old;
	}
	
	/**
	 * get Logging verbosity
	 */
	public boolean getVerbose() {
		return LogUtil.isVerbose(logger_);
	}
	
	/**
	 * add handler into logger
	 * same as this.setLogger().addHandler(handler)
	 */
	public void addHandler(Handler handler) {
		this.logger_.addHandler(handler);
	}

	/**
	 * add a print stream handler with given filename into logger
	 */
	public void addHandler(String filename) {
		try {
		File outfile = new File(filename);
		PrintStream printer = new PrintStream(outfile);
		StreamHandler handler = new StreamHandler(printer, new SimpleLogFormatter());
		this.addHandler(handler);
		} catch (FileNotFoundException fnfe) {
			System.err.println(fnfe.getMessage());
		}
	}

	/**
	 * set number of spaces for one indentation 
	 */
	public int setIndentSpaces(int indent) {
		int previous = indentSpaces_;
		indentSpaces_ = indent;
		return previous;
	}
	
	/**
	 * increment message indentation
	 */
	public String indentUp() {
		for (int i = 0 ; i < indentSpaces_ ; i++)
			indentString_ = indentString_ + " ";
		return indentString_;
	}
	
	/**
	 * decrement message indentation
	 */
	public String indentDown() {
		int end = indentString_.length() - indentSpaces_;
		indentString_ = indentString_.substring(0, (end >= 0 ? end : 0));
		return indentString_;
	}

	/*
	 * methods for logging graphs
	 */
	
	/**
	 * add a vertex attributes to print
	 */
	public void addVertexAttribute(String attribute) {
		if (! attributes_.contains(attribute))
			attributes_.add(attribute);
	}
	
	/**
	 * remove a vertex attributes from printing
	 */
	public void removeVertexAttribute(String attribute) {
		attributes_.remove(attribute);
	}

	/**
	 * return currently printed vertex attributes
	 */
	public Iterable<String> getVertexAttributes() {
		return new ArrayList<String>(attributes_);
	}

	/**
	 * assert at least one attribute is present
	 */
	protected void assertAttributes_() {
		if (attributes_.isEmpty())
			attributes_.add(Registry.Label);
	}

	/**
	 * print debug message with indentation
	 */
	
	public void log(String message) {
		logger_.info(indentString_ + message);
	}
	
	/**
	 * print one vertex
	 */
	public void log(Vertex v) {
		log(strVertex_(v));
	}
	
	/**
	 * print one edge
	 */
	public void log(Edge e) {
		log("e [" + strVertex_(e.v1()) + "] -> [" + strVertex_(e.v2()) + "]");
		// TODO temporary
//		if (e.has(Registry.Colors))
//			log("  colors= " + e.get(Registry.Colors).toString());
	}

	/**
	 * print graph
	 */
	public void log(Graph g, boolean withEdges, String header) {
		log("Graph: " + header);
		
		log("Vertices:");
		for (Vertex v : g.vertices()) {
			log(v);
		}
		
		if (withEdges) {
			log("Edges:");
			GSet<Edge> visit = g.edgeSetFactory().make();
			for (Vertex v : g.vertices()) {
				for (Edge e : v.edges()) {
					if (! visit.contains(e)) {
						log(e);
						visit.add(e);
					}
				}
			}
		}
		log("-------------");
	}

	/**
	 * simplified versions
	 */
	public void log(Graph g, String header) {
		log(g, false, header);
	}
	
	public void log(Graph g) {
		log(g, "");
	}

	/**
	 * print set of vertices
	 */
	public void log(GSet<Vertex> set, String header) {
		log("Set: " + header);
		log("N= " + set.size());
		log("Vertices:");
		for (Vertex v : set) {
			log(v);
		}
		log("-------------");
	}

	public void log(GSet<Vertex> set) {
		log(set, "");
	}
	
	/*
	 * undocumented functions for debugging
	 */
	
	public Vertex checkIntegrity(Graph g, String header) {
		for (Vertex v : g.vertices()) {
			for (Edge e : v.edges()) {
				Vertex u = e.getVertexTo(v);
				if (! g.hasVertex(u)) {
					System.err.println("* boojum detected -> " + header );
					for (String key : u.keys()) {
						System.err.println("*    " + key + " : " + u.get(key).toString());
					}
					return u;
				}
				
			}
		}
		return null;
	}
	
}
