package helix.graph.util;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

/**
 * a class implementing an index permutation
 * in range [from, to]
 * TODO ISelector instead
 */

public class IPermutation implements Iterable<Integer> {

	protected ArrayList<Integer> list_;          // hold the permutation
	protected Hashtable<Integer, Integer> hash_; // hold reverse lookup
	protected int from_, to_;					 // hold from and to

	/*
	 * init from Iterable
	 */
		
	private final void init_(Iterable<Integer> values) {
		
		int len = 0, min = -1, max = -1;

		for (int val : values) {
			if (val < 0)
				throw new IllegalArgumentException("invalid permutation element " + val);
			if ((min < 0) || (val < min))
				min = val;
			if (val > max)
				max = val;
			len++;
		}
			
		if (len != (max - min + 1))
			throw new IllegalArgumentException("invalid permutation size");

		HashSet<Integer> h = new HashSet<Integer>(len);
		for (int val : values) {
			if (h.contains(val))
				throw new IllegalArgumentException("duplicated permutation element " + val);
			h.add(val);
		}
		
		// ok : let's go
		
		list_ = new ArrayList<Integer>(len);
		hash_ = new Hashtable<Integer, Integer>(len);
		from_ = min;
		to_   = max;
		
		int n = 0;
		for (int val : values) {
			list_.add(val);
			hash_.put(val, n++);
		}
	}

	/*
	 * init from range
	 */
	private final void init_(int from, int to) {
		
		int len = to - from + 1;
		if (len <= 0)
			throw new IllegalArgumentException("invalid permutation size");
		
		ArrayList<Integer> tmp = new ArrayList<Integer>(len);
		
		for (int i = 0 ; i < len ; i++)
			tmp.add(from + i);
		
		this.init_(tmp);
	}
	
	/*
	 * parse String
	 */
	private static final Iterable<Integer> parse_(String s) throws NumberFormatException {
		
		ArrayList<Integer> lst = new ArrayList<Integer>();
		
		for (String tok : s.split("\\s+")) {
			int value = Integer.parseInt(tok);
			lst.add(value);
		}
		
		return lst;
	}

	/*
	 * constructors
	 */
	
	/**
	 * new IPermutation from values.
	 * integer values should be contiguous and positive or null
	 */
	public IPermutation(Iterable<Integer> values) {
		init_(values);
	}

	/**
	 * new IPermutation from parsed String.
	 */
	public IPermutation(String s) throws NumberFormatException {
		init_(parse_(s));
	}
	
	/**
	 * identity in [from, to]
	 */
	public IPermutation(int from, int to) {
		init_(from, to);
	}
	
	/**
	 * identity in [0, range-1]
	 */
	public IPermutation(int range) {
		this(0, range-1);
	}

	/*
	 * Iterable API
	 * @see java.lang.Iterable#iterator()
	 */
	
	public Iterator<Integer> iterator() {
		return list_.iterator();
	}

	/*
	 * equals / hashcode API
	 */
	
	/**
	 * override equals
	 */
	public boolean equals(Object o) {
		return (  (o instanceof IPermutation)
			    ? this.list_.equals(((IPermutation) o).list_)
				: false);
	}

	/**
	 * override hashCode
	 */
	public int hashCode() {
		return this.list_.hashCode();
	}
	
	/*
	 * API
	 */

	/**
	 * return value at given index
	 */
	public int get(int indx) {
		return list_.get(indx);
	}
	
	/**
	 * get index of value
	 * return -1 if value is not found
	 */
	public int lookup(int value) {
		return (hash_.containsKey(value) ? hash_.get(value) : -1);
	}

	/**
	 * size of permutation
	 */
	public int size() {
		return list_.size();
	}
	
	/**
	 * check if permutation is identity in [0-n]
	 */
	public boolean isIdentity() {
		int n = 0;
		for (int value : list_) {
			if (value != n)
				return false;
			n++;
		}
		return true;
	}
	
	/**
	 * check if permutation contains the set of integers
	 * given as argument
	 */
	public boolean containsAll(Iterable<Integer> set) {
		for (int i : set) {
			if (this.lookup(i) < 0)
				return false;
		}
		return true;
	}
	
	/**
	 * get permutation starting value (from)
	 */
	public int from() {
		return from_;
	}

	/**
	 * get permutation ending value (to)
	 */
	public int to() {
		return to_;
	}
	
	/**
	 * returns a new permutation with values shifted the specified offset.
	 * 'offset' may be negative but the starting value (from) of
	 * the resulting IPermutation should be >= 0
	 */
	public IPermutation shift(int offset) {
		if (from_ + offset < 0)
			throw new IllegalArgumentException("invalid permutation offset " + offset);
		ArrayList<Integer> lst = new ArrayList<Integer>(size());
		for (int i = 0 ; i < size() ; i++)
			lst.add(list_.get(i) + offset);
		return new IPermutation(lst);
	}

	/**
	 * same as <code>permutation.shift(-permutation.from())</code> i.e.
	 * return a new IPermutation starting at 0.
	 * this function is usually used with apply when the initial
	 * IPermutation is not in the proper range [0, size-1].
	 * <code>permutation.shift().apply(List)</code>
	 */
	public IPermutation shift() {
		return shift(-from_);
	}

	/**
	 * returns a new IPermutation, inverse of this permutation
	 */
	public IPermutation inverse() {
		ArrayList<Integer> values = new ArrayList<Integer>(this.size());
		for (int value = from_ ; value <= to_; value++) {
			values.add(this.lookup(value)); 
		}
		return new IPermutation(values);
	}
	
	/**
	 * apply permutation on list of objects.
	 * return a new list of same type retaining the elements
	 * of 'l' indexed by the current permutation [from, to].
	 * the size of 'l' should therefore be strictly greater than 'to'.
	 * example: if permutation is p=[1,3,2] (i.e in range [1,3]) and l=(a, b, c, d, e)
	 * then p.apply(l)=(b, d, c)
	 * @throws IllegalArgumentException if l.size() < permutation.to()
	 */
	public <T> List<T> apply(List<T> l) {
		if (l.size() < to_)
			throw new IllegalArgumentException("bad permutation range");
		ArrayList<T> tmp = new ArrayList<T>();
		for (int value : list_) {
			tmp.add(l.get(value));
		}
		return tmp;
	}

	/**
	 * override
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer("[");
		for (int value : list_) {
			buf.append(", " + value);
		}
		buf.delete(1,3);
		buf.append("]");
		return buf.toString();
	}
}
