package helix.graph.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An iterator over all possible permutations of 'size' elements.
 * Use in conjunction with the IPermutation class to apply permutation
 * on a list of objects.
 * example: <code>
 * ArrayList<String> lst = new ArrayList<String>();
 * lst.add("a"); lst.add("b"); lst.add("c");
 * IPermutationIterator iter = new IPermutationIterator(lst.size());
 * while (iter.hasNext()) {
 *   List<String> permut = iter.next().apply(lst);
 * }
 * </code>
 * 
 * The base algorithm is described by Kenneth H. Rosen, Discrete Mathematics and Its Applications,
 * 2nd edition (NY: McGraw-Hill, 1991), pp. 282-284.
 * see : http://www.merriampark.com/perm.htm
 * 
 *
 */
public class IPermutationIterator implements Iterator<IPermutation> {

	protected int size_;
	private boolean next_;
	private int [] permutation_;  // perm of index 1..size, perm[0]=0

	/*
	 * internal utility : form final permutation
	 */
	protected final IPermutation formPermutation_() {
		ArrayList<Integer> lst = new ArrayList<Integer>(size_);
		for (int i = 1 ; i <= size_ ; i++)
			lst.add(permutation_[i]-1);
		return new IPermutation(lst);
	}
	/*
	 * internal utility : swap permutation indexes
	 */
	private final void swap_(int i, int j) {
		int tmp = permutation_[i];
		permutation_[i] = permutation_[j];
		permutation_[j] = tmp;
	}

	/**
	 * constructor
	 * @param size > 0
	 */
	public IPermutationIterator(int size) {
		if (size <= 0)
			throw new IllegalArgumentException("size should be > 0");
		size_ = size;
		next_ = true;
		permutation_ = new int [size+1];
		for (int i = 0 ; i <= size ; i++) {
			permutation_[i] = i;
		}
	}

	/*
	 * Iterable API
	 */
	
	public boolean hasNext() {
		return next_;
	}

	public IPermutation next() {
		
		if (! hasNext())
			throw new NoSuchElementException();

		// get current permutation
		IPermutation current = formPermutation_();

		// prepare the next one
		int i = size_ - 1;
		while (permutation_[i] > permutation_[i+1])
			i--;
		
		next_ = (i > 0);
		
		if (next_) {
			int j = size_;
			while (permutation_[i] > permutation_[j])
				j--;

			swap_(i, j);
			int r = size_;
			int s = i + 1;
			while (r > s)
				swap_(r--, s++);
		}
		
		return current;
	}
			
	/**
	 * not implemented : throws UnsupportedOperationException
	 */
	public void remove() {
		throw new UnsupportedOperationException();
	}
	
}
