package helix.graph.util;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * a better stopwatch : you don't need to stop it all the time.<br>
 * here are valid sequences :<br>
 * <code>
 * start stop get start
 * start get ... get stop
 * start get ... reset ... get stop
 * </code>
 * This stopwatch also has some additional logging facilities.
 */

public class StopWatch {

	protected long fStart_ = 0;
	protected long fElapsed_ = 0;

	protected boolean fIsRunning_ = false;
	protected boolean fHasBeenUsedOnce_ = false;
	
	@SuppressWarnings("deprecation")
	// for Java < 1.6
	protected Logger logger_ = Logger.global;
	// for Java >= 1.6
	// protected Logger logger_ = Logger.getLogger(Logger.GLOBAL_LOGGER_NAME);
	
	/*
	 * internal utility : catch time
	 */
	private final long catchTime_() {
		return System.currentTimeMillis();
	}
	
	/*
	 * API
	 */
	
	/**
	 * Constructor : default / use Logger.global as logger
	 */
	public StopWatch() {
	}

	/**
	 * Constructor : default / use Logger.global as logger
	 */
	public StopWatch(Logger logger) {
		logger_ = logger;
	}

	/**
	 * Start the stopwatch.
	 * @param reset reset stopwatch
	 * @throws IllegalStateException if the stopwatch is already running or started for the
	 * first time with reset==false
	 */
	public void start(boolean reset) {
		long time = catchTime_();
		if ( fIsRunning_ ) {
			throw new IllegalStateException("Must stop before calling start again.");
		}
		if ( ! (reset || fHasBeenUsedOnce_) ) {
			throw new IllegalStateException("Must reset when calling start for the first time.");
		}
		if (reset) {
			fElapsed_ = 0;
		}
		fStart_ = time;
		fIsRunning_ = true;
		fHasBeenUsedOnce_ = true;
	}

	/**
	 * Start the stopwatch - with reset
	 * same as <code>start(true)</code>
	 */
	public void start() {
		start(true);
	}

	/**
	 * Stop the stopwatch.
	 *
	 * @throws IllegalStateException if the stopwatch is not already running.
	 */
	public void stop() {
		long time = catchTime_();
		if ( !fIsRunning_ ) {
			throw new IllegalStateException("Cannot stop if not currently running.");
		}
		fElapsed_ += (time - fStart_); // cumulated elapsed time
		fIsRunning_ = false;
	}
	
	/**
	 * Reset the stopwatch, whatever its state
	 */
	public void reset() {
		  fStart_ = catchTime_();
		  fElapsed_ = 0;
	}

	/**
	 * Express the "reading" on the stopwatch as a numeric type.
	 * note: you don't need to stop the stopwatch to read its value
	 */
	public long get() {
		if (fIsRunning_) {
			long time = catchTime_();
			return (time - fStart_) + fElapsed_;
		}
		return fElapsed_;
	}

	/**
	 * override: express the "reading" on the stopwatch as a String
	 */
	public String toString() {
		long elapsed = this.get();
		StringBuffer result = new StringBuffer();
		result.append(elapsed);
		result.append(" ms");
		return result.toString();
	}

	/**
	 * log current value on Logger
	 */
	public void log(Level level, String header) {
		logger_.log(level, header + " " + this.toString());
	}

	/**
	 * log current value on Logger at INFO level
	 */
	public void log(String header) {
		this.log(Level.INFO, header);
	}

	/**
	 * log current value on Logger at INFO level
	 */
	public void log() {
		this.log(Level.INFO, "stopwatch: ");
	}
	
}
